/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/


var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);


window.addEventListener("load", function(event) {

    var args = window.arguments[0];
    if (args.mode = "new" && cal.item.isEvent(args.calendarEvent)) {
        // we will create a new Event (not a task)
        // then we redefine the saveItem function, to add a flag
        // to indicate to the OBM connector that the new item is an internal event,
        // because it is created by the user himself
        var checkingItemChanged = false;
        var originalSaveItem = window.saveItem;
        window.saveItem = function() {
            var item = originalSaveItem();
            if (item.calendar.type == "obm" && !checkingItemChanged) {
                item.setProperty("X-OBMEXT-IS-INTERNAL", "YES");
            }

            return item;
        };

        // The function that checks if the item has changed needs to be
        // modified to ignore the internal flag set above since for new events
        // this flag will never be set before the window is opened.
        var originalItemChanged = window.isItemChanged;
        window.isItemChanged = function() {
            checkingItemChanged = true;
            let winCIC = window.compareItemContent;
            window.compareItemContent = obm_compareItemContent;
            let rv = originalItemChanged();
            window.compareItemContent = winCIC;
            checkingItemChanged = false;
            return rv;
        };
    }
}, false);

// Drop in function to work around Mozilla bug 687412 and the same for TB3
// and TB10 and the ORGANIZER CN property.
function obm_compareItemContent(aFirstItem, aSecondItem, aIgnoreProps, aIgnoreParams) {
    aIgnoreProps = aIgnoreProps || ["X-MOZ-GENERATION", "SEQUENCE", "DTSTAMP",
        "LAST-MODIFIED", "X-MOZ-SEND-INVITATIONS"
    ];
    let ignoreProps = arr2hash(aIgnoreProps);

    let ignoreParams = aIgnoreParams || { "ATTENDEE": ["CN"], "ORGANIZER": ["CN"] };
    for (let x in ignoreParams) {
        ignoreParams[x] = arr2hash(ignoreParams[x]);
    }

    function arr2hash(arr) {
        let hash = {};
        for (let x of arr) {
            hash[x] = true;
        }
        return hash;
    }

    // This doesn't have to be super correct rfc5545, it just needs to be
    // in the same order
    function normalizeComponent(comp) {

        let props = [];
        for (prop in cal.ical.propertyIterator(comp)) {
            if (!(prop.propertyName in ignoreProps)) {
                props.push(prop)
            }
        }
        props = props.sort();

        // let props = [
        //     normalizeProperty(prop)
        //     for (prop in cal.ical.propertyIterator(comp))
        //         if (!(prop.propertyName in ignoreProps))
        // ].sort();

        let comps = [];
        for (subcomp in cal.ical.subcomponentIterator(comp)) {
            comps.push(subcomp)
        }
        comps = comps.sort();

        return comp.componentType + props.join("\r\n") + comps.join("\r\n");
    }

    function normalizeProperty(prop) {
        let params = [k + "=" + v];

        for ([k, v] in cal.ical.paramIterator(prop)) {
            if (!(prop.propertyName in ignoreParams) || !(k in ignoreParams[prop.propertyName])) {
                params.push(v)
            }

        }
        params = params.sort();

        return prop.propertyName + ";" +
            params.join(";") + ":" +
            prop.valueAsIcalString;
    }

    return normalizeComponent(aFirstItem.icalComponent) == normalizeComponent(aSecondItem.icalComponent);
}