/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmRAddressBookSync"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter
var { ObmRAddressBookProxy, AddressBookChangesResponseShell } = ChromeUtils.import('chrome://obm-extension/content/components/obm-sync-rest/obmRAddressBookProxy.js');
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const abCardProperties = [
    "FirstName",
    "LastName",
    "DisplayName",
    "NickName",
    "PrimaryEmail",
    "PhoneticFirstName",
    "PhoneticLastName",
    "SpouseName",
    "FamilyName",
    "SecondEmail",
    "HomeAddress",
    "HomeAddress2",
    "HomeCity",
    "HomeState",
    "HomeZipCode",
    "HomeCountry",
    "WebPage2",
    "WorkAddress",
    "WorkAddress2",
    "WorkCity",
    "WorkState",
    "WorkZipCode",
    "WorkCountry",
    "WebPage1",
    "HomePhone",
    "WorkPhone",
    "FaxNumber",
    "PagerNumber",
    "CellularNumber",
    "JobTitle",
    "Department",
    "Company",
    "_AimScreenName",
    "Custom1",
    "Custom2",
    "Custom3",
    "Custom4",
    "Notes",
    "_GoogleTalk",
    "_AimScreenName",
    "_Yahoo",
    "_Skype",
    "_QQ",
    "_MSN",
    "_ICQ",
    "_JabberId",
    "_IRC",
    "AnniversaryYear",
    "AnniversaryMonth",
    "AnniversaryDay",
    "BirthYear",
    "BirthMonth",
    "BirthDay"
];

/*
Fonctionnement de la synchro addressbook

1) le lancement s'effectue par l'appel de mergeAddressBook
2) l'identification est lancée (ABProxy.doLogin), et appel _onLogged en retour
3) _onLogged() prépare la synchro
    lance ABProxy.getAddressBookSync et en retour _onGetAddressBookSync est appelé
4) _onGetAddressBookSync :
    - traite la liste recue des addressbooks à supprimer et à modifier
        - stocke cette liste dans _addressbooks
        - supprime les addressbooks à supprimer
    - traite la liste reçue des contacts -> _remoteDiff
    - recupère la liste locale des changements des contacts -> _localDiff
    - pour chaque addressbook, traite les conflits (_resolveConflicts)
        -> met à jour _remoteDiff, _localDiff et la liste des changements dans les prefs
    - _mergeRemoteDiff est ensuite appelé
5) _mergeRemoteDiff

    pour chaque addressbook
        met à jour les items de _remoteDiff[].updated
        supprime les items de _remoteDiff[].removed

    si c'est la première pass, lance _sendLocalDiff
    sinon stop
6) _sendLocalDiff:
    prepare les opérations à effectuer en asynchrone
    pour chaque modif, une operation consiste en:
        appel de removeContact, createContactInBook, modifyContact, unsubscribeBook du ABProxy
        appel de _onSentLocalDiff en retour
        suppression de l'item dans la liste des changements.
7) _onSentLocalDiff, appelé en asynchrone à chaque modif, lance l'operation de modif suivante
  ou stope tout.

*/

function ObmRAddressBookSync() {
    try {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
    } catch (e) {
        Components.utils.reportError(e);
    }
    // Init logger
    this.logger = new ObmLogger();

    this.stringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService).createBundle("chrome://obm-extension/locale/errors_warnings.properties");
    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("obmRAddressBookSync.js: ");
    this.logger.debug("new ObmRAddressBookSync()", false, false);

    this.criticalMethods = ["doLogin", "getAddressBookSync", "createContact", "createContactInBook", "modifyContact", "removeContact", "storeContact"];

    this._localDiff = {};
    this._remoteDiff = {};
    this._addressBooksToReload = [];

    this.displayNameDefaultFormat = UtilsOBM.getPrefSafe("extensions.obm.addressbooks.displayName.defaultFormat", "%FirstName% %LastName%");
};

function Context() {
    this.wrappedJSObject = this;
};

ObmRAddressBookSync.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRAddressBookSync]),
    classID: Components.ID("{68214b68-8421-480c-8cd3-bd0cf70a50c4}"),
    contractID: "@obm.org/obm-sync-rest/addressbooksync;1",
    classDescription: "OBM Address Book Sync",
    get wrappedJSObject() { return this; },

    /*
     * Private
     */
    _addListener: function(callback) {
        var self = this;
        let listener = function(aResCode, aMethodName, aResult, aData) {
            try {
                callback.apply(null, arguments);
            } catch (e) {
                if (self._isCriticalMethod(aMethodName)) {
                    self.callerListener.onStop(Cr.NS_ERROR_FAILURE, e);
                } else {
                    self.callerListener.onError(e);
                }
            }
        };
        this.ABProxy.addListener(listener);
    },

    /*
     * Helpers
     */
    _getLocalizedMessage: function(aString) {
        dump("" + aString + "\n");
        try {
            return this.stringBundle.GetStringFromName("obm.syncerror." + aString);
        } catch (e) {
            return aString;
        }
    },
    _isCriticalMethod: function(aMethodeName) {
        for (let method of this.criticalMethods) {
            if (aMethodeName == method) {
                return true;
            }
        }
        return false;
    },
    _printableError: function(aMethodName, aMsg) {
        var printError = "error";

        if (aMsg == "Invalid Certificate") {
            printError = "certError";
        }

        if (aMsg == "No token available") {
            printError = "authFault";
        }

        if (aMsg == "Connection failure") {
            printError = "connectFail";
        }

        if (aMsg == "Server internal error") {
            printError = "internalErr";
        }

        if (aMsg == "Indexing server is unavailable") {
            printError = "solrErr";
        }

        return this._getLocalizedMessage(printError);
    },
    _getOBMCredentials: function(aLogin, aPassword) {

        if (!UtilsOBM.getOBMCredentials(aLogin, aPassword) || !aLogin.value || aLogin.value == "" || !aPassword.value || aPassword.value == "") {
            var error = "NO_LOGIN_INFO";
            this.logger.error(error, true, false);
            this.callerListener.onError(error);
            throw (error);
        }
    },
    _getXPStringFromJSString: function(str) {
        var XPString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
        XPString.data = str;
        return XPString;
    },
    _getJSStringFromXPString: function(XPString) {
        XPString.QueryInterface(Ci.nsISupportsString);
        return XPString.data;
    },

    // DEPRECATED. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
    _getAddressBookNames: function() {
        let names = UtilsOBM.getPrefSafe("extensions.obm.addressbooks", null).toString();
        return names ? names.split(",") : [];
    },

    _addAddressBookNames: function(newName) {
        // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
        var names = this._getAddressBookNames();
        for (var i = 0; i < names.length; i++) {
            if (names[i] == newName) {
                return;
            }
        }
        var value = "";
        if (names.length == 0) {
            value = newName;
        } else {
            value = UtilsOBM.getPrefSafe("extensions.obm.addressbooks", null).toString() + "," + newName;
        }
        UtilsOBM.setPrefSafe("extensions.obm.addressbooks", value.toString());
    },

    _getDirectoryFromABName: function(abName) {
        let abURI = "moz-abmdbdirectory://alia-obm-" + abName + ".mab";
        let addressBookManager = Components.classes["@mozilla.org/abmanager;1"].getService(Components.interfaces.nsIAbManager);
        let dirPrefId = "ldap_2.servers." + abName;
        let iter = addressBookManager.directories;
        while (iter.hasMoreElements()) {
            var data = iter.getNext();
            if (data instanceof Ci.nsIAbDirectory) {
                if (data.dirPrefId == dirPrefId) {
                    var directory = data;
                }
            }
        }
        return directory;
    },
    _getAddressBookDisplayName: function(abName) {
        var abDisplayName = UtilsOBM.getAddressbookManager().getAddressBook(abName).directory.dirName;
        if (abDisplayName == "-1") {
            abDisplayName = UtilsOBM.getLocalizedPref("extensions.obm.addressbooks.users", "OBM - Utilisateurs");
        }
        return abDisplayName;
    },
    // DEPRECATED. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager
    _getLocalContact: function(abName, aId) {
        var directory = this._getDirectoryFromABName(abName);
        var aObjectItem = directory.getCardFromProperty("ObmId", aId, false);
        if (!aObjectItem) {
            throw ("Unexpected id (in addressbook " + abName + "): " + aId);
        }
        return aObjectItem;
    },

    _getLastSync: function() {
        this.logger.debug("_getLastSync()", false, false);
        let lastSync = UtilsOBM.getPrefSafe("extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", "1970-01-01T00:00:00Z");
        // TODO : need an isValidDate func ?
        if (lastSync == "1970-01-01T00:00:00Z") {
            lastSync = "0";
        }
        return lastSync;
    },
    _setLastSync: function(lastSync) {
        this.logger.debug("SETLASTSYNC (" + lastSync + ") : " + "extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", false, false);
        UtilsOBM.setPrefSafe("extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", lastSync);
    },
    _createContactObjFromId: function(abName, aId, aWillBeCreated) {
        // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
        var aObjectItem = this._getLocalContact(abName, aId);
        var contact = {
            QueryInterface: ChromeUtils.generateQI([Ci.obmIRContact]),

            firstName: aObjectItem.firstName,
            lastName: aObjectItem.lastName, // not nillable
            email: aObjectItem.primaryEmail,
            commonname: aObjectItem.displayName
        };

        //there are no getter setter for folowing properties in tb3
        contact.uid = aWillBeCreated ? null : aObjectItem.getProperty("ObmId", null);

        contact.aka = aObjectItem.getProperty("NickName", null);
        contact.email2 = aObjectItem.getProperty("SecondEmail", null);

        contact.workPhone = aObjectItem.getProperty("WorkPhone", null);
        contact.homePhone = aObjectItem.getProperty("HomePhone", null);
        contact.workFax = aObjectItem.getProperty("FaxNumber", null);
        contact.pagerPhone = aObjectItem.getProperty("PagerNumber", null);
        contact.mobilePhone = aObjectItem.getProperty("CellularNumber", null);

        // TODO smart merge
        contact.homeStreet = aObjectItem.getProperty("HomeAddress", "") + " " + aObjectItem.getProperty("HomeAddress2", "");
        contact.homeZipCode = aObjectItem.getProperty("HomeZipCode", null);
        contact.homeCity = aObjectItem.getProperty("HomeCity", null);
        contact.homeState = aObjectItem.getProperty("HomeState", null);
        contact.homeExpressPostal = null;
        contact.homeCountry = aObjectItem.getProperty("HomeCountry", null);
        contact.homeWebPage = aObjectItem.getProperty("WebPage2", null);

        contact.title = aObjectItem.getProperty("JobTitle", null);
        contact.company = aObjectItem.getProperty("Company", null);
        contact.service = aObjectItem.getProperty("Department", null);

        // TODO smart merge
        contact.workStreet = aObjectItem.getProperty("WorkAddress", "") + " " + aObjectItem.getProperty("WorkAddress2", "");
        contact.workZipCode = aObjectItem.getProperty("WorkZipCode", null);
        contact.workCity = aObjectItem.getProperty("WorkCity", null);
        contact.workState = aObjectItem.getProperty("WorkState", null);
        contact.workExpressPostal = null;
        contact.workCountry = aObjectItem.getProperty("WorkCountry", null);
        contact.workWebPage = aObjectItem.getProperty("WebPage1", null);

        contact.comment = aObjectItem.getProperty("Notes", null);
        var custom1 = aObjectItem.getProperty("Custom1", "");
        if ((custom1 != null) && (custom1.indexOf('X-OBM-CALURI:') == 0)) {
            contact.caluri = custom1.substr(13);
        } else {
            contact.caluri = '';
        }
        return contact;
    },
    _computeDisplayName: function(card) {
        let displayName = this.displayNameDefaultFormat;

        for (let property of abCardProperties) {
            try {
                displayName = displayName.replace('%' + property + '%', card.getProperty(property, ""));
            } catch (e) {
                this.logger.error("_computeDisplayName(): Failed to get property value '" + property + "', is the property valid?");
            }
        }

        return displayName.trim();
    },
    _updateLocalContact: function(localContact, aNewContact) {
        let previousContactHasDefaultDisplayName = localContact.displayName == this._computeDisplayName(localContact);

        localContact.firstName = aNewContact.firstName;
        localContact.lastName = aNewContact.lastName;
        localContact.primaryEmail = aNewContact.email;

        localContact.setProperty("ObmId", aNewContact.uid);
        localContact.setProperty("NickName", aNewContact.aka);

        localContact.setProperty("SecondEmail", aNewContact.email2);

        localContact.setProperty("WorkPhone", aNewContact.workPhone);
        localContact.setProperty("HomePhone", aNewContact.homePhone);
        localContact.setProperty("FaxNumber", aNewContact.workFax);
        localContact.setProperty("CellularNumber", aNewContact.mobilePhone);
        localContact.setProperty("PagerNumber", aNewContact.pagerPhone);

        // TODO smart split
        localContact.setProperty("HomeAddress", aNewContact.homeStreet);
        localContact.setProperty("WorkAddress2", null);
        localContact.setProperty("HomeZipCode", aNewContact.homeZipCode);
        localContact.setProperty("HomeCity", aNewContact.homeCity);
        localContact.setProperty("HomeState", aNewContact.homeState);
        localContact.setProperty("HomeCountry", aNewContact.homeCountry);
        localContact.setProperty("WebPage2", aNewContact.homeWebPage);

        localContact.setProperty("JobTitle", aNewContact.title);
        localContact.setProperty("Department", aNewContact.service);
        localContact.setProperty("Company", aNewContact.company);

        // TODO smart split
        localContact.setProperty("WorkAddress", aNewContact.workStreet);
        localContact.setProperty("WorkAddress2", null);
        localContact.setProperty("WorkZipCode", aNewContact.workZipCode);
        localContact.setProperty("WorkCity", aNewContact.workCity);
        localContact.setProperty("WorkState", aNewContact.workState);
        localContact.setProperty("WorkCountry", aNewContact.workCountry);
        localContact.setProperty("WebPage1", aNewContact.workWebPage);

        localContact.setProperty("Notes", aNewContact.comment);
        if (aNewContact.caluri) {
            localContact.setProperty('Custom1', 'X-OBM-CALURI:' + aNewContact.caluri);
        }

        //compute displayName
        if (aNewContact.commonname) {
            localContact.displayName = aNewContact.commonname;
        } else {
            if (!localContact.displayName || localContact.displayName == "" || previousContactHasDefaultDisplayName) {
                localContact.displayName = this._computeDisplayName(localContact);
            }
        }

        return localContact;
    },

    //
    // Resolve conflicts following override rule
    //
    _resolveConflicts: function(abName) {
        this.logger.debug("_resolveConflicts()", false, false);
        let serverPriority = (UtilsOBM.getPrefSafe("extensions.obm.override", "server") == "server");
        var remoteListHasChanged = false;
        for (var i = 0; i < this._remoteDiff[abName].removed.length; i++) {
            var contactId = "" + this._remoteDiff[abName].removed[i];
            var rLocalIdx = this._localDiff[abName].removed.indexOf(contactId);
            var uLocalIdx = this._localDiff[abName].updated.indexOf(contactId);
            if (rLocalIdx != -1) {
                // A conflict
                this._remoteDiff[abName].removed.splice(i--, 1);
                this._localDiff[abName].removed.splice(rLocalIdx, 1);
                UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "removed", contactId);
                remoteListHasChanged = true;
            }
            if (uLocalIdx != -1) {
                // A conflict
                if (serverPriority) {
                    this._localDiff[abName].updated.splice(uLocalIdx, 1);
                    UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "updated", contactId);
                } else {
                    if (!remoteListHasChanged) this._remoteDiff[abName].removed.splice(i--, 1);
                }
            }
        }
        for (var i = 0; i < this._remoteDiff[abName].updated.length; i++) {
            remoteListHasChanged = false;
            var contactId = "" + this._remoteDiff[abName].updated[i].uid;
            var rLocalIdx = this._localDiff[abName].removed.indexOf(contactId);
            var uLocalIdx = this._localDiff[abName].updated.indexOf(contactId);
            if (uLocalIdx != -1) {
                // A conflict
                if (serverPriority) {
                    this._localDiff[abName].updated.splice(uLocalIdx, 1);
                    UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "updated", contactId);
                } else {
                    this._remoteDiff[abName].updated.splice(i--, 1);
                    remoteListHasChanged = true;
                }
            }
            if (rLocalIdx != -1) {
                // A conflict
                if (serverPriority) {
                    this._localDiff[abName].removed.splice(uLocalIdx, 1);
                    UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "removed", contactId);
                } else {
                    if (!remoteListHasChanged) this._remoteDiff[abName].updated.splice(i--, 1);
                }
            }
        }

    },

    //
    // Query address book local modifications
    //
    _getLocalDiff: function(abName) {
        this.logger.debug("_getLocalDiff()", false, false);
        var changeTypes = ["added", "updated", "removed"];
        this._localDiff[abName] = {};
        for (var type of changeTypes) {
            let pref = UtilsOBM.getPrefSafe("extensions.obm.changes." + Ci.nsIAbDirectory + "." + abName + "." + Ci.nsIAbCard + "." + type, "");
            var changes = (pref != "" ? pref.split(",") : []);
            this._localDiff[abName][type] = [];
            for (var i = 0; i < changes.length; i++) {
                if (changes[i] == null || changes[i] == "") {
                    this.logger.error("_getLocalDiff id is null (" + "id:" + changes[i] + ", " + abName + ", " + type + ")", true, false);
                } else {
                    this._localDiff[abName][type].push(changes[i]);
                }
            }
        }
    },
    _ensureValidResponse: function(aResCode, aMethodName, aResult, aValidMethodName) {
        if (aResCode != Cr.NS_OK) {
            var msgError = "unknown";
            if (aResult instanceof Ci.nsISupportsString) msgError = aResult.data;

            if (msgError == "No token available" && UtilsOBM.getPrefSafe("extensions.obm.clearPasswordOnLoginFailure", true)) {
                var username = {};
                var password = {};
                if (UtilsOBM.getOBMCredentials(username, password)) {
                    UtilsOBM.setOBMCredentials(username.value, null);
                    UtilsOBM.setTemporaryPassword(null);
                }
            } else if (msgError == "NotAllowed") {
                return false;
            }

            var printError = this._printableError(aMethodName, msgError);

            // Trying to remove a contact that doesn't exist (probably deleted by another mean)
            // Consider this as a success as we want this contact removed anyway...
            if (msgError.indexOf("not found") > -1 && aMethodName == "removeContact") {
                return true;
            }

            msgError = aMethodName + ": Reported error: " + msgError + "(" + aResCode + ")";
            this.logger.error(msgError, true, false);

            if (this.callerListener) {
                if (this._isCriticalMethod(aMethodName)) {
                    this.callerListener.onStop(aResCode, printError);
                } else {
                    this.callerListener.onError(printError);
                }
            }
            return false;
        }

        if (aValidMethodName && aMethodName != aValidMethodName) {
            var msgError = "Method " + aMethodName + " unexpected. Expected: " + aValidMethodName;
            this.logger.error(msgError, true, false);
            this.callerListener.onError(msgError);
            return false;
        }
        return true;
    },

    //
    // Get address book id from its name
    //
    _getABId: function(abName) {
        return (abName == "users" ? "Ci.obmIRBook.USERS" : "Ci.obmIRBook.CONTACTS");
    },

    _onLogged: function(aResCode, aMethodName, aResult, aData) {
        this.callerListener.onOperationDone(true, null);
        this.logger.debug("_onLogged()", false, false);

        this.ABProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "doLogin")) {
            return;
        };
        this.callerListener.onNewOperation("getAddressBookSync", null);
        var self = this;
        this._getSyncOperationCount = 1;

        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetAddressBookSync(aResCode, aMethodName, aResult, aData)
        });

        this._localOperations = [];
        this._localOperations[0] = ["contacts", 0, 1,
            function(abName, aIdx, aTotal) {
                var lastSync = self._getLastSync();
                self.logger.debug("_sendNextLocalOperation()", false, false);
                self.callerListener.onProgress(aIdx, aTotal, "", "contacts");
                var data = new Context();
                data.isTheFirstPass = true;
                self.ABProxy.getAddressBookSync(lastSync, data);
            }
        ];

        this._localOperationsIdx = 0;
        this._sendNextLocalOperation();
    },
    _onGetAddressBookSync: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onGetAddressBookSync()", false, false);
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "getAddressBookSync")) {
            return;
        };

        var data = aData.wrappedJSObject;
        // var syncResponse = aResult.QueryInterface(Ci.obmIRAddressBookChangesResponse);
        var syncResponse = aResult;

        this._addressbooks = {};
        this._addressbooks["removed"] = syncResponse.getRemovedAddressBooksIds({});
        this._addressbooks["updated"] = syncResponse.getUpdatedAddressBooks({});


        var abMgr = UtilsOBM.getAddressbookManager();
        for (var addressbookId of this._addressbooks["removed"]) {
            try {
                abMgr.removeAddressBook(addressbookId);
            } catch (e) {
                //Probably already removed
                this.logger.debug("addressbook " + addressbookId + " has been asked to be removed, but does not exist.", false, false, false);
            }
        }

        for (var addressbook of this._addressbooks["updated"]) {
            // TODO. OBM_NEW_AD_MANAGER. use the new API  obmIAddressBooksManager
            this._ensureAddressBookIsPresent(addressbook);
        }
        // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
        this.abNames = this._getAddressBookNames();

        for (var abId of this.abNames) {
            this._remoteDiff[abId] = {};
            this._remoteDiff[abId]["updated"] = [];
            this._remoteDiff[abId]["removed"] = [];
            if (this._getLastSync() != "0") {
                this._remoteDiff[abId]["removed"] = UtilsOBM.capabilitySupported("ADDRESSBOOK_HANDLER_SUPPORTS_ADDRESSBOOKID_IN_REMOVALS") ?
                    syncResponse.getRemovedContactIdsForAddressBook(abId, {}) : syncResponse.getRemovedContactIds({});
            }
        }

        var localDiff = {};
        for (var i = 0; i < this.abNames.length; i++) {
            this._getLocalDiff(this.abNames[i]);
        }

        let locallyRemovedAddressBooksPref = UtilsOBM.getPrefSafe("extensions.obm.changes." + Ci.nsIAbDirectory + ".list." + Ci.nsIAbDirectory + ".removed", "");
        this._localDiff["unsubscribed"] = (locallyRemovedAddressBooksPref != "" ? locallyRemovedAddressBooksPref.split(",") : []);

        for (var contact of syncResponse.getUpdatedContacts({})) {
            if (!this._remoteDiff[contact.addressbookid]) {
                this._remoteDiff[contact.addressbookid] = { updated: [], removed: [] };
            }
            if (!this._remoteDiff[contact.addressbookid].updated) {
                this._remoteDiff[contact.addressbookid].updated = [];
            }
            if (!this._remoteDiff[contact.addressbookid].removed) {
                this._remoteDiff[contact.addressbookid].removed = [];
            }
            var length = this._remoteDiff[contact.addressbookid]["updated"].length;
            this._remoteDiff[contact.addressbookid]["updated"][length] = contact;

        }

        for (var addressbook of this._addressbooks["updated"]) {
            if (data.isTheFirstPass) {
                this._resolveConflicts(addressbook.uid);
            }
        }

        this._remoteDiff.lastSync = syncResponse.lastsync;

        this._getSyncOperationCount--;

        if (this._getSyncOperationCount <= 0) {
            var self = this;
            this.ABProxy.removeListener();
            this.callerListener.onOperationDone(true, null);
            this._mergeRemoteDiff(data.isTheFirstPass);
        } else {
            this._sendNextLocalOperation();
        }
    },
    _mergeRemoteDiff: function(isTheFirstPass) {
        this.logger.debug("_mergeRemoteDiff()", false, false);

        var mergeActionCount = 0;
        var mergeAction = [];
        for (var abName of this.abNames) {
            mergeAction[abName] = {};
            mergeAction[abName].updatedCount = this._remoteDiff[abName].updated.length;
            mergeAction[abName].removedCount = this._remoteDiff[abName].removed.length;
            mergeActionCount += this._remoteDiff[abName].updated.length;
            mergeActionCount += this._remoteDiff[abName].removed.length;
        }
        this.callerListener.onNewOperation("mergeRemoteDiff", mergeActionCount);
        this.logger.debug("Action count: " + mergeActionCount, false, false);

        var uIdx = 0;
        var rIdx = 0;
        // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to retrieve the list
        // of addressbooks to add/update/remove a local contact.
        for (var abName of this.abNames) {
            let addrDb = null;
            let directory = this._getDirectoryFromABName(abName);

            for (var updatedItem of this._remoteDiff[abName].updated) {
                try {
                    // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
                    var localContact = this._getLocalContact(abName, updatedItem.uid);
                    this.logger.debug(abName + "> WILL UPDATE: " + updatedItem.firstName, false, false);
                    this._updateLocalContact(localContact, updatedItem);
                    // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
                    directory.modifyCard(localContact);
                } catch (e) {
                    try {
                        let localContact = Cc["@mozilla.org/addressbook/cardproperty;1"].createInstance(Ci.nsIAbCard);
                        this._updateLocalContact(localContact, updatedItem);
                        this.logger.debug(abName + "> WILL CREATE: " + localContact.displayName, false, false);
                        // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to create a local contact
                        directory.addCard(localContact);
                    } catch (e) {
                        //this.logger.error("ERROR (ignored): " + e + "\n", true, false);
                    }
                }
                var abDisplayName = this._getAddressBookDisplayName(abName);
                this.callerListener.onProgress(uIdx++, mergeAction[abName].updatedCount, "updated", abDisplayName);
            }
            for (var removedItem of this._remoteDiff[abName].removed) {
                this.logger.debug(abName + "> WILL REMOVE: " + removedItem, false, false);
                try {
                    // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to remove a local contact
                    let contact = this._getLocalContact(abName, removedItem);
                    let cardsToDelete = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
                    cardsToDelete.appendElement(contact, false);
                    directory.deleteCards(cardsToDelete);
                } catch (e) {
                    //this.logger.error("ERROR (ignored): " + e + "\n", true, false);
                }
                var abDisplayName = this._getAddressBookDisplayName(abName);
                this.callerListener.onProgress(rIdx++, mergeAction[abName].removedCount, "removed", abDisplayName);
            }
            uIdx = 0;
            rIdx = 0;
        }


        this._setLastSync(this._remoteDiff.lastSync);

        this.callerListener.onOperationDone(true, null);
        if (isTheFirstPass) {
            this._sendLocalDiff();
        } else {
            if (this._addressBooksToReload.length > 0) {
                this._reloadNextAddressBook();
            } else {
                this.callerListener.onStop(Cr.NS_OK, "done");
            }
        }
    },
    _onSentLocalDiff: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onSentLocalDiff()", false, false);
        this._localDiffOperationCount--;
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, null)) {
            if ((aResult instanceof Ci.nsISupportsString) && aResult == "NotAllowed") {
                let found = false;
                let ctx = aData.wrappedJSObject;

                for (let abName of this._addressBooksToReload) {
                    if (abName == ctx.abName) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    this._addressBooksToReload.push(ctx.abName);
                }

                // Remove the change from the local list, we're handling this afterwards
                UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, ctx.abName, Ci.nsIAbCard, ctx.optype, ctx.id);
            }
            // Do not continue if this is a fatal error
            else if (this._isCriticalMethod(aMethodName)) {
                return;
            }
        } else {
            var ctx = null;
            if (aData && aData.wrappedJSObject)
                ctx = aData.wrappedJSObject;

            switch (aMethodName) {
                case "createContact":
                case "createContactInBook":
                case "storeContact":
                case "modifyContact":
                    let updatedContact = aResult;
                    // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
                    let localContact = this._getLocalContact(ctx.abName, ctx.id);
                    this._updateLocalContact(localContact, updatedContact);
                    // TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
                    let directory = this._getDirectoryFromABName(ctx.abName);
                    directory.modifyCard(localContact);
                    break;
                case "removeContact":
                case "unsubscribeBook":
                    break;
                default:
            };

            // we know now that the change has been applied, we remove it from the list
            if (ctx && ctx.optype) {
                if (aMethodName == "unsubscribeBook") {
                    UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, "list", Ci.nsIAbDirectory, ctx.optype, ctx.id);
                } else {
                    UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, ctx.abName, Ci.nsIAbCard, ctx.optype, ctx.id);
                }
            }
        }

        if (this._localDiffOperationCount <= 0) {
            this.logger.debug("Operations done", false, false);
            this.callerListener.onOperationDone(true, null);
            var self = this;
            this.ABProxy.removeListener();
            var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            this.logger.debug("timer init", false, false);
            this.callerListener.onNewOperation("wait-one-seconde", null);
            timer.initWithCallback({
                    notify: function(timer) {
                        self.logger.debug("ok, time is out", false, false);
                        self._fireSecondPass();
                    }
                },
                1100, timer.TYPE_ONE_SHOT);
            this.timer = timer;
        } else {
            this._sendNextLocalOperation();
        }
    },

    _reloadNextAddressBook: function(index, total) {
        if (!total) {
            index = 0;
            total = this._addressBooksToReload.length;
            this.callerListener.onNewOperation("cleaning", null);
        }

        if (index < this._addressBooksToReload.length) {
            let abId = this._addressBooksToReload[index++];

            this._reloadAddressBook(abId, index, total);
        } else {
            this.callerListener.onOperationDone(true, null);
            this._handleLostAddressBookRights();
            this.callerListener.onStop(Cr.NS_OK, "done");
        }
    },

    _reloadAddressBook: function(abId, index, total) {
        let self = this;
        let obmAbMgr = UtilsOBM.getAddressbookManager();
        let oldBook = obmAbMgr.getAddressBook(abId).directory;
        let oldBookInfo = { dirPrefId: oldBook.dirPrefId, dirName: oldBook.dirName, URI: oldBook.URI };

        // Shoot the addressbook...
        try {
            obmAbMgr.removeAddressBook(abId);
        } catch (e) {
            // Already removed ?
        }

        // ...Recreate it with the exact same attributes...
        obmAbMgr.addAddressBook(abId, oldBookInfo.dirName, oldBookInfo.URI, oldBookInfo.dirPrefId);

        // ...Query the server for updated...
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            let directory = UtilsOBM.getAddressbookManager().getAddressBook(abId);
            let response = aResult.QueryInterface(Ci.obmIRAddressBookSyncResponse);

            self.callerListener.onProgress(index, total, null, directory.directory.dirName);
            // ...Fill the addressbooks with updated contacts from the server
            for (let contact of response.getUpdatedContacts({})) {
                let localContact = Cc["@mozilla.org/addressbook/cardproperty;1"].createInstance(Ci.nsIAbCard);

                self._updateLocalContact(localContact, contact);

                self.logger.debug(directory.directory.dirName + " > WILL CREATE: " + localContact.displayName);
                directory.addCard(localContact);
            }

            self.ABProxy.removeListener();
            self._reloadNextAddressBook(index, total);
        });

        this.ABProxy.listAllChanges(0, abId, null);
    },

    _handleLostAddressBookRights: function() {
        if (UtilsOBM.getPrefSafe("extensions.obm.rights.showChangedRightsMessage", true)) {
            let details = "";

            if (this._addressBooksToReload.length > 0) {
                for (let abId of this._addressBooksToReload) {
                    details += "\n" + UtilsOBM.getAddressbookManager().getAddressBook(abId).directory.dirName;
                }

                let promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                    .getService(Components.interfaces.nsIPromptService);

                let title = UtilsOBM.getLocalizedMessage("obm.sync.rights.addressbooks.title");
                let message = UtilsOBM.getLocalizedMessage("obm.sync.rights.addressbooks.message") + "\n" + details;

                promptService.alert(null, title, message);
            }
        }

        this._addressBooksToReload = [];
    },

    _fireSecondPass: function() {
        this.logger.debug("_fireSecondPass()", false, false);
        this._getSyncOperationCount = 1;
        var self = this;
        this.callerListener.onOperationDone(true, null);
        this.callerListener.onNewOperation("getAddressBookSync", null);
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetAddressBookSync(aResCode, aMethodName, aResult, aData)
        });

        this._localOperations = [];
        this._localOperations[0] = ["contacts", 0, 1,
            function(abName, aIdx, aTotal) {
                var lastSync = self._getLastSync();
                self.logger.debug("_sendNextLocalOperation()", false, false);
                self.callerListener.onProgress(aIdx, aTotal, "", "contacts");
                var data = new Context();
                data.isTheFirstPass = false;
                self.ABProxy.getAddressBookSync(lastSync, data);
            }
        ];

        this._localOperationsIdx = 0;
        this._sendNextLocalOperation(false);
    },
    _sendLocalDiff: function() {
        this.logger.debug("_sendLocalDiff()", false, false);
        var queriesCount = 0;
        for (var i = 0; i < this.abNames.length; i++) {
            var abName = this.abNames[i];

            queriesCount += this._localDiff[abName].removed.length;
            queriesCount += this._localDiff[abName].updated.length;
            queriesCount += this._localDiff[abName].added.length;
        }
        queriesCount += this._localDiff["unsubscribed"].length;

        this.callerListener.onNewOperation("sendLocalDiff", queriesCount);

        this._localDiffOperationCount = queriesCount;
        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
        });

        if (queriesCount == 0) {
            this._onSentLocalDiff(Cr.NS_OK, null, null, null);
            return;
        }

        var idx = 0;
        this._localOperations = [];
        for (var i = 0; i < this.abNames.length; i++) {
            var abName = this.abNames[i];
            var abDisplayName = this._getAddressBookDisplayName(abName);
            for (var j = 0; j < this._localDiff[abName].removed.length; j++) {
                this._localOperations[idx++] = [abName, j, this._localDiff[abName].removed.length,
                    (function(abDisplayName) {
                        return function(abName, aIdx, aTotal) {
                            self.callerListener.onProgress(aIdx, aTotal, "removed", abDisplayName);
                            var ctx = new Context();
                            ctx.abName = abName;
                            ctx.id = self._localDiff[abName].removed[aIdx];
                            ctx.optype = "removed";
                            self.ABProxy.removeContact(self._getABId(abName), ctx.id, ctx);
                        }
                    })(abDisplayName)
                ];
            }
            for (var j = 0; j < this._localDiff[abName].added.length; j++) {
                this._localOperations[idx++] = [abName, j, this._localDiff[abName].added.length,
                    (function(abDisplayName) {
                        return function(abName, aIdx, aTotal) {
                            self.callerListener.onProgress(aIdx, aTotal, "added", abDisplayName);
                            let contextId = self._localDiff[abName].added[aIdx];
                            self._createContact(abName, contextId);
                        }
                    })(abDisplayName)

                ];
            }
            for (var j = 0; j < this._localDiff[abName].updated.length; j++) {
                this._localOperations[idx++] = [abName, j, this._localDiff[abName].updated.length,
                    (function(abDisplayName) {
                        return function(abName, aIdx, aTotal) {
                            self.callerListener.onProgress(aIdx, aTotal, "updated", abDisplayName);
                            let contextId = self._localDiff[abName].updated[aIdx];
                            self._modifyContact(abName, contextId);
                        }
                    })(abDisplayName)
                ];
            }
        }
        for (var j = 0; j < this._localDiff.unsubscribed.length; j++) {
            this._localOperations[idx++] = [this._localDiff.unsubscribed[j], j, this._localDiff.unsubscribed.length,
                (function(abDisplayName) {
                    return function(abName, aIdx, aTotal) {
                        self.callerListener.onProgress(aIdx, aTotal, "unsubscribed", abDisplayName);
                        var ctx = new Context();
                        ctx.abName = abName;
                        ctx.id = self._localDiff.unsubscribed[aIdx];
                        ctx.optype = "removed";
                        self.ABProxy.unsubscribeBook(ctx.id, ctx);
                    }
                })(abDisplayName)
            ];
        }
        this._localOperationsIdx = 0;
        this._sendNextLocalOperation();
    },
    _createContact: function(addressBookName, contextId) {
        let context = new Context();
        context.abName = addressBookName;
        context.id = contextId;
        context.optype = "added";
        UtilsOBM.capabilitySupported("ADDRESS_BOOK_HANDLER_SUPPORTS_STORECONTACT") ?
            this.ABProxy.storeContact(addressBookName, this._createContactObjFromId(addressBookName, context.id, true), context) :
            this.ABProxy.createContactInBook((addressBookName, this._createContactObjFromId(addressBookName, context.id, true), context));
    },
    _modifyContact: function(addressBookName, contextId) {
        let context = new Context();
        context.abName = addressBookName;
        context.id = contextId;
        context.optype = "updated";
        if (UtilsOBM.capabilitySupported("ADDRESS_BOOK_HANDLER_SUPPORTS_STORECONTACT")) {
            this.ABProxy.storeContact(addressBookName,
                this._createContactObjFromId(addressBookName, context.id, false), context);
        } else {
            this.ABProxy.modifyContact(this._getABId(addressBookName),
                this._createContactObjFromId(addressBookName, context.id, false), context);
        }
    },
    _sendNextLocalOperation: function() {
        var idx = this._localOperationsIdx++;
        var self = this;
        this._localOperations[idx][3](this._localOperations[idx][0], this._localOperations[idx][1], this._localOperations[idx][2]);
    },
    // DEPRECATED. OBM_NEW_AD_MANAGER. use the new API  obmIAddressBooksManager::ensureAddressBookExists
    //
    // param : addressbook should contain properties :
    //      uid ( the variable part of the file, ex : alia-obm-" + addressbook.uid + ".mab" )
    //  addressbook can contain properties :
    //      name ( don't set it if ya don't know )
    //      ownerDisplayName : self explainatory
    //
    _ensureAddressBookIsPresent: function(addressbook) {
        this.logger.debug("_ensureAddressBookIsPresent()", false, false);
        this.logger.debug("uid: " + addressbook.uid + ", name: " + addressbook.name, false, false);
        var addressBookName = addressbook.uid;
        var dirName = addressbook.name ? addressbook.name : "";

        /*
            get fake names configured in tbird properties @ extensions.obm.addressbooks.[dirName]
            the "name" reference is the one of the OBM SERVER.
        */
        if (dirName.length) {
            this.logger.debug("checking if addressbook " + dirName + " has a fake label from property extensions.obm.addressbooks." + dirName, false, false);
            this.logger.debug(UtilsOBM.getLocalizedPref("extensions.obm.addressbooks." + dirName, "nope"), false, false);
            dirName = UtilsOBM.getLocalizedPref("extensions.obm.addressbooks." + dirName, dirName);
            if (addressbook.ownerDisplayName != null && addressbook.ownerDisplayName != "") {
                dirName += " (" + addressbook.ownerDisplayName + ")";
            }
        }
        var addressBookManager = Components.classes["@mozilla.org/abmanager;1"].getService(Components.interfaces.nsIAbManager);
        var abURI = "moz-abmdbdirectory://" + addressBookName + ".mab";
        var dirPrefId = "ldap_2.servers." + addressBookName;
        var iter = addressBookManager.directories;
        var found = false;
        while (iter.hasMoreElements()) {
            var data = iter.getNext();
            if (data instanceof Ci.nsIAbDirectory) {
                if (data.dirPrefId == dirPrefId) {
                    found = true;
                    /* Update name if necessary */
                    if (dirName.length && data.dirName != dirName) {
                        data.dirName = dirName;
                    }
                }
            }
        }
        if (!found) {
            addressBookManager.newAddressBook(dirName.length ? dirName : addressBookName, abURI, 2, dirPrefId);
        }
        this._addAddressBookNames(addressBookName);
    },

    /*
     * obmIRAddressBookSync
     */

    mergeAddressBook: function(aListener) {
        this.logger.debug("mergeAddressBook()", false, false);
        this.callerListener = aListener;
        this.callerListener.onStart();

        //
        // Remote jobs
        //
        this.callerListener.onNewOperation("logUserIn", null);
        var login = {};
        var password = {};
        this._getOBMCredentials(login, password);

        try {
            var PortUrl = UtilsOBM.getPrefSafe("extensions.obm.server"); // +
            // "/AddressBook";
        } catch (e) {
            var error = "NO_SERVER_INFO";
            this.logger.error(error, true, false);
            this.callerListener.onError(error);
            throw (error);
        }

        this.ABProxy = new ObmRAddressBookProxy();
        this.ABProxy.PortUrl = PortUrl;

        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onLogged(aResCode, aMethodName, aResult, aData)
        });
        this.ABProxy.doLogin(login.value, password.value, UtilsOBM.getOrigin(), null);
    }
};

// module

var components = [ObmRAddressBookSync];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}