/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmRCalendarProxy"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js')
    // unixFileWriter
var { XPCOMUtils } = ChromeUtils.import("resource://gre/modules/XPCOMUtils.jsm");

ChromeUtils.defineModuleGetter(this, "ObmRESTClient", "chrome://obm-extension/content/components/obmRESTClient.jsm");


const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function _xS(value) {
    return value || "";
}

function _xL(value, def) {
    if (typeof value == "number") {
        return value;
    } else if (typeof value == "string") {
        return parseInt(value, 10);
    } else if (def) {
        return def;
    } else {
        return 0;
    }
}

function _xB(value) {
    return value + "" == "true";
}

function serializeREvent(aEvent /* obmIREvent */ , aNoId) {
    function convertEvent(aInnerEvent, aKind, aInnerNoId) {
        // aInnerEvent.QueryInterface(Ci.obmIREventRecurrence);
        let event = aInnerEvent;

        function sanitizeForXml(input) {
            return new String(input).replace(/[^\u0009\u000a\u000d\u0020-\ud7ff\ue000-\ufffd]/g, "\ufffd");
        }

        function _appendXMLElement(aName, aValue, aParent) {
            if (aValue) {
                let parent = aParent || docEl;
                let elem = doc.createElement(aName);
                parent.appendChild(elem);
                elem.textContent = sanitizeForXml(aValue);
                return elem;
            }
        }

        let doc, obj, docEl;
        if (aKind == "event") {
            // xmlns="http://www.obm.org/xsd/sync/event.xsd"/>;
            doc = OBMXMLCreateDoc("event");
            docEl = doc.documentElement;
            if (event.uid && !aInnerNoId) {
                docEl.setAttribute("id", event.uid);
            }
            docEl.setAttribute("isInternal", event.isInternal ? "true" : "false");
            docEl.setAttribute("sequence", event.generation || 0);

            _appendXMLElement("extId", event.extId);
        } else {
            doc = OBMXMLCreateDoc("eventException");
            _appendXMLElement("recurrenceId", event.recurrenceId);
        }

        docEl.setAttribute("allDay", event.allday);
        docEl.setAttribute("type", "VEVENT");

        _appendXMLElement("owner", event.owner);
        _appendXMLElement("priority", event.priority);
        _appendXMLElement("privacy", event.privacy);
        _appendXMLElement("date", event.date);
        _appendXMLElement("tz", event.tz);
        _appendXMLElement("duration", event.duration);
        _appendXMLElement("title", event.title);
        _appendXMLElement("description", event.description);
        _appendXMLElement("category", event.category);
        _appendXMLElement("location", event.location);
        _appendXMLElement("alert", event.alert);
        _appendXMLElement("opacity", event.opacity);
        _appendXMLElement("ownerEmail", event.ownerEmail);

        let atts = event.getAttendees({});
        if (atts.length) {
            let attElem = docEl.appendChild(doc.createElement("attendees"));
            for (let att of atts) {
                let xmlAtt = attElem.appendChild(doc.createElement("attendee"));
                xmlAtt.setAttribute("email", _xS(att.email));
                xmlAtt.setAttribute("displayName", _xS(att.displayName));
                xmlAtt.setAttribute("state", _xS(att.status));
                xmlAtt.setAttribute("required", _xS(att.type));
                xmlAtt.setAttribute("isOrganizer", att.isOrganizer ? "true" : "false");
            }
        }

        let recur = docEl.appendChild(doc.createElement("recurrence"));
        recur.setAttribute("kind", _xS(event.kind));
        recur.setAttribute("freq", _xL(event.frequence));
        recur.setAttribute("days", _xS(event.days));
        if (event.end) recur.setAttribute("end", event.end);


        let exs = event.getExceptions({});
        if (exs.length) {
            let excElem = recur.appendChild(doc.createElement("exceptions"));
            for (let ex of exs) {
                excElem.appendChild(doc.createElement("exception")).textContent = ex;
            }
        }

        let eventExs = event.getEventExceptions({});
        if (eventExs.length) {
            let excElem = recur.appendChild(doc.createElement("eventExceptions"));
            for (let ex of eventExs) {
                let exEventDoc = convertEvent(ex, "eventException");
                excElem.appendChild(exEventDoc.documentElement);
            }
        }

        return doc;
    }

    return '<?xml version="1.0" encoding="UTF-8"?>' +
        OBMSerializeDOM(convertEvent(aEvent, "event", aNoId));
}

function ObmRCalendarProxy() {
    try {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
        Components.utils.import("resource://obm-extension/modules/XMLUtilsOBM.jsm");
        Components.utils.import("resource://obm-extension/modules/OBMErrors.jsm");
    } catch (e) {
        Components.utils.reportError(e);
    }

    // Init logger
    this.logger = new ObmLogger();

    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("obmCalendarProxy: ");
    this.logger.debug("new obmCalendarProxy()", false, false);

    this.RESTClient = new ObmRESTClient().wrappedJSObject;

    this.mListeners = [];
    this.mUrl = null;
    this.mLoginUrl = null;
    this.mToken = {
        QueryInterface: ChromeUtils.generateQI([Ci.obmIRAuthToken]),
        session_id: null
    };
};

ObmRCalendarProxy.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCalendarProxy, Ci.obmIRProxyBase]),
    classID: Components.ID("{e2b34be1-b2b2-47de-bfe2-4f8126183ca5}"),
    contractID: "@obm.org/obm-sync-rest/calendarproxy;1",
    classDescription: "OBM Calendar Proxy",

    /*
     * obmIRProxyBase
     */
    addListener: function(listener) {
        for (var i = 0; i < this.mListeners.length; ++i) {
            if (this.mListeners[i] == listener)
                return;
        }
        this.mListeners.push(listener);
    },

    removeListener: function(listener) {
        this.mListeners = [];
    },
    _notifyListeners: function(aResultCode, aMethodName, aObj, aData) {
        // A copy is needed, a new listeners could has been added during the
        // callbacks
        var listeners = this.mListeners.slice(0, this.mListeners.length);
        for (var i = 0; i < listeners.length; ++i) {
            try {
                listeners[i].apply(null, [aResultCode, aMethodName, aObj, aData]);
            } catch (e) {
                // Continue to notify listeners
                this.logger.error(e + " (ignored)");
            }
        }
    },

    get token() {
        return this.mToken;
    },
    set token(aToken) {
        this.mToken = aToken;
    },

    get PortUrl() {
        return this.mUrl;
    },
    set PortUrl(aURL) {
        if (aURL.charAt(aURL.length - 1) == '/') {
            aURL = aURL.substring(0, aURL.length - 1);
        }
        this.mUrl = aURL + "/calendar";
        this.mLoginUrl = aURL + "/login";
    },

    /*
     * obmIRCalendarProxy
     */
    _callback: function(aMethodName, aObj, aDoc, aData) {
        this.logger.debug("_callback(aMethodName = " + aMethodName + ")", false, false);

        let docEl = aDoc && aDoc.documentElement;

        if (aObj && aObj.name == "Error") {
            var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);

            if (aMethodName == "doLogin" && aObj.fileName != "Connection failure" && aObj.fileName != "Server internal error") {
                if (UtilsOBM.capabilitySupported("LOGIN_HANDLER_SUPPORTS_AUTHFAULT")) {
                    let exception = OBMXPathFirst(docEl, "type/text()");

                    if (exception == "org.obm.sync.auth.AuthFault") {
                        msg.data = "No token available";
                    } else {
                        msg.data = aObj.fileName;
                    }
                } else {
                    msg.data = "No token available";
                }
            } else {
                let exception = OBMXPathFirst(docEl, "type/text()")

                if (exception == "org.obm.sync.NotAllowedException") {
                    msg.data = "NotAllowed";
                } else if (exception == "org.obm.sync.PermissionException") {
                    msg = new PermissionException(
                        docEl.getAttribute("calendar"),
                        docEl.getAttribute("right"),
                        docEl.getAttribute("eventTitle"));
                } else {
                    msg.data = aObj.fileName;
                }
            }
            if (aObj.fileName == "Invalid Certificate") {
                msg.data = "Invalid Certificate";
            }
            this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, msg, aData);
            return;
        }

        switch (aMethodName) {
            case "doLogin":
                var sid = OBMXPathFirst(docEl, "sid/text()");
                this.mToken.session_id = sid;
                this.logger.debug("token obtained with sid:" + sid, false, false);
                var syncVersion = null;
                let versionTag = OBMXPathFirst(docEl, "version");
                if (versionTag) {
                    syncVersion = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
                    syncVersion.data = versionTag.getAttribute("major") +
                        "." + versionTag.getAttribute("minor") +
                        "." + versionTag.getAttribute("release");
                    this.logger.debug("sync version:" + syncVersion.data, false, false);
                }
                var domain = OBMXPathFirst(docEl, "domain/text()");
                UtilsOBM.setPrefSafe("extensions.obm.server.domain", domain);
                UtilsOBM.domainUuid.getFromLoginResponse(docEl);
                this._notifyListeners(Cr.NS_OK, aMethodName, syncVersion, aData);

                this._storeUserSettings(docEl);
                this._storeServerCapabilities(docEl);

                break;
            case "getSync":
                this._notifyListeners(Cr.NS_OK, aMethodName, new GetSyncEventResponseShell(docEl), aData);
                break;
            case "getSyncWithSortedChanges":
                this._notifyListeners(Cr.NS_OK, aMethodName, new GetSyncEventWithSortedChangesResponseShell(docEl), aData);
                break;
            case "createEvent":
                this._notifyListeners(Cr.NS_OK, aMethodName, this._newNSISupportsString(OBMXPathFirst(aDoc, "value/text()")), aData);
                break;
            case "storeEvent":
                let event = new EventShell(docEl);

                this._notifyListeners(Cr.NS_OK, aMethodName, this._newNSISupportsString(event.uid), aData);
                break;
            case "modifyEvent":
                this._notifyListeners(Cr.NS_OK, aMethodName, new EventShell(docEl), aData);
                break;
            case "changeParticipationState":
                this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
                break;
            case "removeEvent":
                this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
                break;
            case "removeEventByExtId":
                this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
                break;
            case "listCalendars":
                var calendarInfos = this._makeCalendarInfoArray(docEl);
                this._notifyListeners(Cr.NS_OK, aMethodName, calendarInfos, aData);
                break;
            case "listResources":
                var calendarInfos = this._makeResourceInfoArray(docEl);
                this._notifyListeners(Cr.NS_OK, aMethodName, calendarInfos, aData);
                break;
            case "listCategories":
                this._notifyListeners(Cr.NS_OK, aMethodName, new GetCategoriesShell(docEl), aData);
                break;
            case "purge":
                this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
                break;
            case "getCalendarMetadata":
                var calendarInfos = this._makeCalendarInfoArray(docEl);
                this._notifyListeners(Cr.NS_OK, aMethodName, calendarInfos, aData);
                break;
            case "getResourceMetadata":
                var calendarInfos = this._makeResourceInfoArray(docEl);
                this._notifyListeners(Cr.NS_OK, aMethodName, calendarInfos, aData);
                break;
            default:
                this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, null, aData);
        };
    },

    _newNSISupportsString: function(str) {
        let nsiss = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);

        nsiss.data = str;

        return nsiss;
    },

    _storeUserSettings: function(docEl) {
        let settings = OBMXPath(docEl, "//settings/setting");

        // We first clear the preferences so that if the server returns an incomplete list
        // for the logged-in user we're sure to not keep outdated preferences
        UtilsOBM.deletePreferenceBranch("extensions.obm.user.preferences");
        UtilsOBM.resetPref("calendar.allday.defaultTransparency");

        if (settings) {
            for (var setting of settings) {
                let name = setting.getAttribute("name");
                let value = setting.firstChild.nodeValue;

                UtilsOBM.setPrefSafe("extensions.obm.user.preferences." + name, value);
                // Special case of default allday opacity
                if (name == "set_allday_opacity") {
                    UtilsOBM.setPrefSafe("calendar.allday.defaultTransparency", value);
                }
            }
        }
    },

    _storeServerCapabilities: function(docEl) {
        let capabilities = OBMXPath(docEl, "//server-capabilities/server-capability");

        UtilsOBM.deletePreferenceBranch("extensions.obm.server.capabilities");

        if (capabilities) {
            for (var capability of capabilities) {
                let name = capability.getAttribute("name");

                UtilsOBM.setPrefSafe("extensions.obm.server.capabilities." + name, capability.firstChild.nodeValue);
            }
        }
    },

    getSync: function(aCalName, aLastSync, aData) {
        this.logger.debug("getSync()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalName);
        p.add("lastSync", aLastSync);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "getSync", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    getSyncWithSortedChanges: function(aCalName, aLastSync, aData) {
        this.logger.debug("getSyncWithSortedChanges()", false, false);

        let p = this.RESTClient.newRESTClientParameters();

        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalName);
        p.add("lastSync", aLastSync);
        p.add("syncRangeAfter", UtilsOBM.getSyncAfterFromPref());

        // closure
        let self = this;
        this.RESTClient.invoke(this.mUrl, "getSyncWithSortedChanges", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    createEvent: function(aCalendarId, aEvent, aNotification, aData) {
        this.logger.debug("createEvent()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalendarId);
        p.add("event", serializeREvent(aEvent, true));
        p.add("notification", aNotification);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "createEvent", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    storeEvent: function(aCalendarId, aEvent, aNotification, aData) {
        this.logger.debug("storeEvent(calendar=" + aCalendarId + ", notification=" + aNotification + ")", false, false);

        let self = this;
        let p = this.RESTClient.newRESTClientParameters();

        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalendarId);
        p.add("event", serializeREvent(aEvent, true));
        p.add("notification", aNotification);

        this.RESTClient.invoke(this.mUrl, "storeEvent", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    modifyEvent: function(aCalendarId, aEvent, aUpdateAttendees, aNotification, aGeneration, aData) {
        this.logger.debug("modifyEvent()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalendarId);
        p.add("event", serializeREvent(aEvent));
        p.add("updateAttendees", aUpdateAttendees);
        p.add("notification", aNotification);
        p.add("sequence", aGeneration);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "modifyEvent", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    changeParticipationState: function(calendar, extId, participationState, aNotification, aGeneration, aReccurenceId, aData) {
        this.logger.debug("changeParticipationState()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", calendar);
        p.add("extId", extId);
        p.add("state", participationState);
        p.add("notification", aNotification);
        p.add("sequence", aGeneration);

        //recurrence Id support enabled
        p.add("recursive", aData.recursive ? "true" : "false");
        if (aReccurenceId) {
            p.add("recurrenceId", aReccurenceId);
        }
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "changeParticipationState", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    removeEvent: function(aCalendarId, aUid, aNotification, aGeneration, aData) {
        this.logger.debug("removeEvent()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalendarId);
        p.add("id", aUid);
        p.add("notification", aNotification);
        p.add("sequence", aGeneration);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "removeEvent", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    removeEventByExtId: function(aCalendarId, aExtId, aNotification, aGeneration, aData) {
        this.logger.debug("removeEventByExtId()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalendarId);
        p.add("extId", aExtId);
        p.add("notification", aNotification);
        p.add("sequence", aGeneration);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "removeEventByExtId", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    doLogin: function(aLogin, aPwd, aOrigin, aData) {
        this.logger.debug("doLogin()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("login", aLogin);
        p.add("password", aPwd);
        this.logger.debug("doLogin() SEL5" + aOrigin, false, false);
        p.add("origin", aOrigin);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mLoginUrl, "doLogin", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    listCalendars: function(aLimit, aOffset, aPattern, aData) {
        this.logger.debug("listCalendars()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        if (aLimit >= 0) {
            p.add("limit", aLimit);
            p.add("offset", aOffset);
            p.add("pattern", aPattern);
        }
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "listCalendars", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    listResources: function(aLimit, aOffset, aPattern, aData) {
        this.logger.debug("listResources()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        if (aLimit >= 0) {
            p.add("limit", aLimit);
            p.add("offset", aOffset);
            p.add("pattern", aPattern);
        }
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "listResources", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    listCategories: function(aData) {
        this.logger.debug("listCategories()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "listCategories", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml); }, aData);
    },
    getCalendarMetadata: function(calendarEmails, calendarEmailsLength, aData) {
        this.logger.debug("getCalendarMetadata()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);

        for (var i = 0; i < calendarEmails.length; i++)
            p.add("calendar", calendarEmails);

        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "getCalendarMetadata", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml); }, aData);
    },
    getResourceMetadata: function(calendarEmails, calendarEmailsLength, aData) {
        this.logger.debug("getResourceMetadata()", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);

        for (var i = 0; i < calendarEmails.length; i++)
            p.add("resource", calendarEmails);

        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "getResourceMetadata", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml); }, aData);
    },
    // void purge(in AString calendar, in nsISupports aData);
    purge: function(aCalName, aData) {
        this.logger.debug("purge(" + aCalName + ")", false, false);
        var p = this.RESTClient.newRESTClientParameters();
        p.add("sid", this.mToken.session_id);
        p.add("calendar", aCalName);
        // closure
        var self = this;
        this.RESTClient.invoke(this.mUrl, "purge", p, true, function(name, obj, xml, data) { self._callback(name, obj, xml, data); }, aData);
    },
    _makeResourceInfoArray: function(aDocEl) {
        // Makes an XPCOM array
        var infoArray = {
            list: OBMXPath(aDocEl, "resourceInfo"),
            QueryInterface: ChromeUtils.generateQI([Ci.nsIArray, Ci.nsIMutableArray]),
            get length() { return this.list ? this.list.length : 0; },
            enumerate: function() { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            indexOf: function(idx, elt) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            queryElementAt: function(idx, uuid) {
                var rescalshell = new ResourceShell(this.list[idx]);
                return rescalshell;
            },
            appendElement: function(el, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            insertElementAt: function(el, idx, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            replaceElementAt: function(el, idx, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            removeElementAt: function(idx) {
                if (this.list) {
                    this.list.splice(idx, 1);
                }
            }
        };
        return infoArray;
    },

    _makeCalendarInfoArray: function(aDocEl) {
        // Makes an XPCOM array
        var infoArray = {

            list: OBMXPath(aDocEl, "info"),
            QueryInterface: ChromeUtils.generateQI([Ci.nsIArray, Ci.nsIMutableArray]),
            get length() { return this.list ? this.list.length : 0; },
            enumerate: function() { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            indexOf: function(idx, elt) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            queryElementAt: function(idx, uuid) {
                let calshell = new CalendarShell(this.list[idx]);
                return calshell; //.QueryInterface(uuid);
            },
            appendElement: function(el, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            insertElementAt: function(el, idx, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            replaceElementAt: function(el, idx, weak) { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
            removeElementAt: function(idx) {
                if (this.list) {
                    this.list.splice(idx, 1);
                }
            }
        };
        return infoArray;
    }
};

function CalendarShell(aNode) {
    this.mNode = aNode;

    // obmIRCalendarInfo
    /*<info>
        <first>user3</first>
        <last>user3</last>
        <mail>user3@dom1.local</mail>
        <uid>user3</uid>
        <read>true</read>
        <write>true</write>
    </info>*/
}

CalendarShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCalendarInfo]),

    get uid() { return _xS(OBMXPathFirst(this.mNode, "uid/text()")); },
    get firstname() { return _xS(OBMXPathFirst(this.mNode, "first/text()")); },
    get lastname() { return _xS(OBMXPathFirst(this.mNode, "last/text()")); },
    get mail() { return _xS(OBMXPathFirst(this.mNode, "mail/text()")); },
    get read() { return _xB(OBMXPathFirst(this.mNode, "read/text()")); },
    get write() { return _xB(OBMXPathFirst(this.mNode, "write/text()")); }
};

function ResourceShell(aNode) {
    this.mNode = aNode;
}

ResourceShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRResourceInfo]),

    // obmIRResourcesInfo
    /* <info>
        <uid>res3</uid>
        <name>res3</name>
        <mail>res3@dom1.local</mail>
        <description>res3description</description>
        <read>true</read>
        <write>true</write>
    </info> */
    get id() { return _xS(OBMXPathFirst(this.mNode, "id/text()")); },
    get name() { return _xS(OBMXPathFirst(this.mNode, "name/text()")); },
    get mail() { return _xS(OBMXPathFirst(this.mNode, "mail/text()")); },
    get description() { return _xS(OBMXPathFirst(this.mNode, "description/text()")); },
    get read() { return _xB(OBMXPathFirst(this.mNode, "read/text()")); },
    get write() { return _xB(OBMXPathFirst(this.mNode, "write/text()")); }
};

function ObmRAttendee(email, displayName, status, type, isOrganizer) {
    this.mEmail = email;
    this.mDisplayName = displayName;
    this.mStatus = status;
    this.mType = type;
    this.mIsOrganizer = isOrganizer;
}
ObmRAttendee.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRAttendee]),

    get email() { return this.mEmail; },
    get displayName() { return this.mDisplayName; },
    get status() { return this.mStatus; },
    set status(val) { this.mStatus = val; },
    get type() { return this.mType; },
    get isOrganizer() { return this.mIsOrganizer; }
}

function ParticipationChangeShell(aNode) {
    this.mNode = aNode;
}

ParticipationChangeShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRParticipationChange]),

    // obmIRParticipationChange
    get uid() { return this.mNode.getAttribute("id"); },
    get extId() { return this.mNode.getAttribute("extId"); },
    get generation() { return _xL(this.mNode.getAttribute("sequence")); },
    get recurrenceId() { return this.mNode.getAttribute("recurrenceId"); },

    getAttendees: function(aCount) {
        let obmAttendees = [];
        let attendees = OBMXPath(this.mNode, "attendees/*");
        if (attendees) {
            for (let att of attendees) {
                let attendee = new ObmRAttendee(att.getAttribute("email"),
                    att.getAttribute("displayName"),
                    att.getAttribute("state"),
                    att.getAttribute("required"),
                    _xB(att.getAttribute("isOrganizer")));
                obmAttendees.push(attendee);
            }
        }

        function cachedGetAttendees(aInnerCount) {
            aInnerCount.value = obmAttendees.length;
            return obmAttendees;
        };

        this.getAttendees = cachedGetAttendees;
        return cachedGetAttendees(aCount || {});
    },

    getAttendeeByEmail: function(email) {
        let attendees = this.getAttendees({});
        let lowerMail = email.toLowerCase();
        for (let att of attendees) {
            if (att.email.toLowerCase() == lowerMail) {
                return att;
            }
        }
        return null;
    }
};

function EventShell(aNode) {
    this.mNode = aNode;
}

EventShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIREventRecurrence, Ci.obmIREvent]),

    // obmIREvent
    get uid() { return this.mNode.getAttribute("id"); },
    get hash() { return OBMXPathFirst(this.mNode, "hash/text()"); },
    get extId() { return OBMXPathFirst(this.mNode, "extId/text()"); },
    get generation() { return _xL(this.mNode.getAttribute("sequence"), 0); },
    get recurrenceId() { return _xS(OBMXPathFirst(this.mNode, "recurrenceId/text()")); },
    get owner() { return _xS(OBMXPathFirst(this.mNode, "owner/text()")); },
    get ownerEmail() { return _xS(OBMXPathFirst(this.mNode, "ownerEmail/text()")); },
    get priority() { return _xL(OBMXPathFirst(this.mNode, "priority/text()")); },
    get privacy() { return _xL(OBMXPathFirst(this.mNode, "privacy/text()")); },
    get allday() { return _xB(this.mNode.getAttribute("allDay")); },
    get anonymized() { return _xB(this.mNode.getAttribute("anonymized")); },
    get date() { return _xS(OBMXPathFirst(this.mNode, "date/text()")); },
    get tz() { return _xS(OBMXPathFirst(this.mNode, "tz/text()")); },
    get duration() { return _xL(OBMXPathFirst(this.mNode, "duration/text()")); },
    get timeupdate() {
        return _xS(OBMXPathFirst(this.mNode, "timeupdate/text()") ||
            OBMXPathFirst(this.mNode, "timecreate/text()"));
    },
    get title() { return _xS(OBMXPathFirst(this.mNode, "title/text()")); },
    get description() { return _xS(OBMXPathFirst(this.mNode, "description/text()")); },
    get category() { return _xS(OBMXPathFirst(this.mNode, "category/text()")); },
    get location() { return _xS(OBMXPathFirst(this.mNode, "location/text()")); },
    get opacity() { return _xS(OBMXPathFirst(this.mNode, "opacity/text()")); },
    get recurrence() { return this; },
    get wrappedJSObject() { return this; },
    getAttendees: function(aCount) {
        let obmAttendees = [];
        let attendees = OBMXPath(this.mNode, "attendees/*");
        if (attendees) {
            for (let att of attendees) {
                var attendee = new ObmRAttendee(att.getAttribute("email"),
                    att.getAttribute("displayName"),
                    att.getAttribute("state"),
                    att.getAttribute("required"),
                    _xB(att.getAttribute("isOrganizer")));
                obmAttendees.push(attendee);
            }
        }

        function cachedGetAttendees(aInnerCount) {
            aInnerCount.value = obmAttendees.length;
            return obmAttendees;
        };

        this.getAttendees = cachedGetAttendees;
        return cachedGetAttendees(aCount || {});
    },

    getAttendeeByEmail: function(email) {
        let attendees = this.getAttendees({});
        let lowerMail = email.toLowerCase();
        for (let att of attendees) {
            if (att.email.toLowerCase() == lowerMail) {
                return att;
            }
        }
        return null;
    },

    // obmIREventRecurrence
    get kind() { return _xS(OBMXPathFirst(this.mNode, "recurrence/@kind")); },
    get frequence() { return _xL(OBMXPathFirst(this.mNode, "recurrence/@freq")); },
    get days() { return _xS(OBMXPathFirst(this.mNode, "recurrence/@days")); },
    get end() { return _xS(OBMXPathFirst(this.mNode, "recurrence/@end")); },

    getExceptions: function(aCount) {
        let obmExceptions = OBMXPath(this.mNode, "recurrence/exceptions/exception/text()") || [];

        function cachedGetExceptions(aInnerCount) {
            aInnerCount.value = obmExceptions.length;
            return obmExceptions;
        }

        this.getExceptions = cachedGetExceptions;
        return cachedGetExceptions(aCount || {});
    },
    getEventExceptions: function(aCount) {
        let obmEventExceptions = [];
        let eventExs = OBMXPath(this.mNode, "recurrence/eventExceptions/eventException");
        for (let eventEx of eventExs) {
            obmEventExceptions.push(new EventShell(eventEx));
        }

        function cachedGetEventExceptions(aInnerCount) {
            aInnerCount.value = obmEventExceptions.length;
            return obmEventExceptions;
        }

        this.getEventExceptions = cachedGetEventExceptions;
        return cachedGetEventExceptions(aCount || {});
    },
    get alert() { return _xL(OBMXPathFirst(this.mNode, "alert/text()"), -1); },
    get isInternal() { return _xB(this.mNode.getAttribute("isInternal")); }
};

function GetSyncEventWithSortedChangesResponseShell(aNode) {
    this.mUpdated = [];
    this.mRemoved = [];
    this.mParticipationChanged = [];
    this.mRemovedExtId = [];
    this.mLastSync = aNode.getAttribute("lastSync");

    OBMXPath(aNode, "updated/*").forEach(updated => {
        if (!OBMXPathFirst(updated, "recurrenceId/text()")) {
            // Only add event if its not an exception, this is already taken
            // care of in the parent item.
            this.mUpdated.push(new EventShell(updated));
        }
    })

    OBMXPath(aNode, "removed/*").forEach(removed => {
        this.mRemoved.push(removed.getAttribute("id"));
        let extId = removed.getAttribute("extId");
        if (extId) {
            this.mRemovedExtId.push(extId);
        }
    })

    OBMXPath(aNode, "participationChanges/*").forEach(pchange => {
        let pcshell = new ParticipationChangeShell(pchange);
        this.mParticipationChanged.push(pcshell);
    })


    /*
    <calendar-changes lastSync="20081127230000000" xmlns="http://www.obm.org/xsd/sync/calendar-changes.xsd">
        <participation-changes>
            <participation extId="966c71d7-04d0-4551-85b7-c9641e20f1d6" id="26">
              <attendees>
                  <attendee email="blum@obm.matthieu.lng" state="ACCEPTED"/>
                  <attendee email="jaures@obm.matthieu.lng" state="ACCEPTED"/>
              </attendees>
            </participation>
        </participation-changes>
        <removed>
            <event id="user3-2" extId="85f148ae-2400-4860-b1dd-ae3fa062b07b"/>
            <event id="user3-3" extId="e1e253f9-b7a6-488b-a360-5ca294afb803"/>
            <event id="user3-4" extId="ad3f09e7-1a5e-40c2-ae1d-12906fbc3760"/>
            <event id="user3-5" extId="e19991ea-8131-4b61-88fe-e25b84311073"/>
        </removed>
        <updated>
            <event id="1">
            <title>eeeeee</title>
            <owner>user3</owner>
            <date>20081127110000000</date>
            <duration>3600</duration>
            <priority>2</priority>
            <privacy>0</privacy>
            <attendees>
                <attendee displayName="user3 user3" email="user3@dom1.local" required="CHAIR" state="ACCEPTED"/>
            </attendees>
            <recurrence kind="none"/>
            </event>
        </updated>
    </calendar-changes>
    */
}

GetSyncEventWithSortedChangesResponseShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCalendarSyncWithSortedChangesResponse]),

    getUpdatedEvents: function(count) {
        count.value = this.mUpdated.length;
        return this.mUpdated;
    },
    getRemovedEventIds: function(count) {
        count.value = this.mRemoved.length;
        return this.mRemoved;
    },
    getRemovedEventExtIds: function(count) {
        count.value = this.mRemovedExtId.length;
        return this.mRemovedExtId;
    },
    getParticipationChanges: function(count) {
        count.value = this.mParticipationChanged.length;
        return this.mParticipationChanged;
    },
    get lastsync() {
        return this.mLastSync;
    }
};

function GetSyncEventResponseShell(aNode) {
    this.mUpdated = [];
    this.mRemoved = [];
    this.mRemovedExtId = [];
    this.mLastSync = aNode.getAttribute("lastSync");

    for (var updated of OBMXPath(aNode, "updated/*")) {
        this.mUpdated.push(new EventShell(updated));
    }

    for (var removed of OBMXPath(aNode, "removed/*")) {
        this.mRemoved.push(removed.getAttribute("id"));
        let extId = removed.getAttribute("extId");
        if (extId) {
            this.mRemovedExtId.push(extId);
        }
    }

    /*
    <calendar-changes lastSync="20081127230000000" xmlns="http://www.obm.org/xsd/sync/calendar-changes.xsd">
        <removed>
            <event id="user3-2" extId="85f148ae-2400-4860-b1dd-ae3fa062b07b"/>
            <event id="user3-3" extId="e1e253f9-b7a6-488b-a360-5ca294afb803"/>
            <event id="user3-4" extId="ad3f09e7-1a5e-40c2-ae1d-12906fbc3760"/>
            <event id="user3-5" extId="e19991ea-8131-4b61-88fe-e25b84311073"/>
        </removed>
        <updated>
            <event id="1">
                <title>eeeeee</title>
                <owner>user3</owner>
                <date>20081127110000000</date>
                <duration>3600</duration>
                <priority>2</priority>
                <privacy>0</privacy>
                <attendees>
                    <attendee displayName="user3 user3" email="user3@dom1.local" required="CHAIR" state="ACCEPTED"/>
                </attendees>
                <recurrence kind="none"/>
            </event>
        </updated>
    </calendar-changes>
    */
}

GetSyncEventResponseShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCalendarSyncResponse]),

    getUpdatedEvents: function(count) {
        count.value = this.mUpdated.length;
        return this.mUpdated;
    },
    getRemovedEventIds: function(count) {
        count.value = this.mRemoved.length;
        return this.mRemoved;
    },
    getRemovedEventExtIds: function(count) {
        count.value = this.mRemovedExtId.length;
        return this.mRemovedExtId;
    },
    get lastsync() {
        return this.mLastSync;
    }
};

function GetCategoriesShell(aNode) {
    this.mCategories = (OBMXPath(aNode, "cat") || []).map(
        function createCatShell(xcat) {
            return {
                QueryInterface: ChromeUtils.generateQI([Ci.obmIRCategory]),
                id: _xL(xcat.getAttribute("id")),
                label: _xS(xcat.getAttribute("label"))
            };
        }
    );
}

GetCategoriesShell.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCategoriesResponse]),
    getCategories: function(count) {
        count.value = this.mCategories.length;
        return this.mCategories;
    }
};

// module
var components = [ObmRCalendarProxy];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}