/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmRCalendarSync"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter
var { ObmRCalendarProxy } = ChromeUtils.import('chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarProxy.js');

var { PermissionException } = ChromeUtils.import('resource://obm-extension/modules/OBMErrors.jsm')

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function getListener(options) {
    options = options || {};
    let opListener = {
        onGetResult: function() {},
        onOperationComplete: function(calendar, status, opType, id, detail) {
            if (Components.isSuccessCode(status)) {
                if ("success" in options) {
                    options.success(calendar);
                }
            } else {
                if ("error" in options) {
                    options.error(calendar);
                }
            }
            if ("complete" in options) {
                options.complete(calendar);
            }
        }
    };
    return opListener;
}

var self = this;

function nextTick(callback) {
    self.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    self.timer.initWithCallback({ notify: callback }, 10, timer.TYPE_ONE_SHOT);
}

/*
This object is responsible of two things:
 - [A] to merge one (mergeOneCalendar()) or all calendars (mergeAllCalendars())
 - [B] to retrieve the list of calendars (getSubscribleCalendars()).
        FIXME: I think this feature has been added here, instead of into the calendar
        proxy, because ObmRCalendarSync is implementing useful methods (even if
        some of these methods should be in the proxy)

For the three methods :

1) _doLogin: it calls CalProxy.doLogin
2) after the login (_onLogged),
    if ([B] or OBM calendar is not loaded)
        it retrieves the list of calendars (CalProxy.listCalendars)
        result passed to _onListCalendar
        if just [B], it is finished
        else, it imports the OBM calendar into the mozilla calendar manager
            and then it retrieve the list of categories (_listCategories)
    else
        it retrieve the list of categories (_listCategories)
3) on the categories list retrieval (_onListCategories)
    it stores the list of categories into the pref calendar.categories.names

4) The sync begin : _getLocalDiff
    - it fills _localDiff with the list of event to add/update/remove into OBM
    - it fills _localDiff with the list of exceptions (FIXME: what is it ?) to add/update/
    - it prepares the list of operation to do on each calendars : == list of
        callbacks which do a synchronisation with a specific calendar

    It then launch each item of this list. _onGetSync() asynchronously for item.
5) _onGetSync, called for calendar
    - it fills _remoteDiff with informations from the response
    - if it is the first pass, it resolve conflicts (_resolveConflicts)
        it changes _remoteDiff and _localDiff depending of conflicts found

6) when all sync data have been retrieved (all _onGetSync have been called), _mergeRemoteDiff is called
    it processes all _remoteDiff updated/removed (remove or updateOrAdopt the events)

7) then for the first pass, _sendLocalDiff is called. It calls OBM (CalProxy) for local modification
    the list is in _localDiff
    - calls CalProxy.removeEventByExtId/removeEvent for removed events
    - calls _createEventObjFromId for event to add or modify
        it converts the mozilla Event to an OBM event
        it calls CalProxy.createEvent or CalProxy.modifyEvent
        it update the change list into prefs

    _onSentLocalDiff is called for OBM server call (after the http response)
        it update the local calendar
8) the second pass (_fireSecondPass) begins after the last _onSentLocalDiff (after 1 sec ??)
    it calls getSync for calendar -> 5)
    FIXME: needs more documentation about this second pass (why there is a second pass?)
*/
function ObmRCalendarSync() {
    try {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
        var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
        Components.utils.import("resource://obm-extension/modules/OBMErrors.jsm");
        var scriptLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Ci.mozIJSSubScriptLoader);
        scriptLoader.loadSubScript("chrome://obm-extension/content/synchronizer/CalendarConflictResolver.js");
    } catch (e) {
        Components.utils.reportError(e);
    }
    // Init logger
    this.logger = new ObmLogger();

    this.stringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)
        .createBundle("chrome://obm-extension/locale/errors_warnings.properties");
    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("obmRCalendarSync.js: ");
    this.logger.debug("new ObmRCalendarSync()", false, false);

    this.criticalMethods = ["doLogin", "getSyncWithSortedChanges", "modifyEvent", "createEvent", "storeEvent"];

    this._localDiff = {};
    this._remoteDiff = {};

    // {calendar.id: { evtId: {action: "updated", ...} , ... }, ... }
    // the list of events that will be sent to the server
    this._evtsLocalDiff = {};

    this._calendarsToRemove = [];
    this._calendarsNowReadOnly = [];
}

function Context() {
    this.wrappedJSObject = this;
};

ObmRCalendarSync.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmIRCalendarSync]),
    classID: Components.ID("{b70b2123-cdd5-45d6-a439-7842c1b02b6d}"),
    contractID: "@obm.org/obm-sync-rest/calendarsync;1",
    classDescription: "OBM Calendar Sync",
    get wrappedJSObject() { return this; },

    /*
     * Private
     */

    _addListener: function(callback) {
        let self = this;
        let listener = function(aResCode, aMethodName, aResult, aData) {
            try {
                callback.apply(null, arguments);
            } catch (e) {
                if (self._isCriticalMethod(aMethodName)) {
                    self.callerListener.onStop(Cr.NS_ERROR_FAILURE, e);
                } else {
                    self.callerListener.onError(e);
                }
            }
        };
        this.CalProxy.addListener(listener);
    },

    //
    // Helpers
    //
    _getLocalizedMessage: function(aString) {
        //dump(""+aString+"\n");
        try {
            return this.stringBundle.GetStringFromName("obm.syncerror." + aString);
        } catch (e) {
            return aString;
        }
    },
    _isCriticalMethod: function(aMethodeName) {
        for (let method of this.criticalMethods) {
            if (aMethodeName == method) {
                return true;
            }
        }
        return false;
    },
    _printableError: function(aMethodName, aMsg, aData) {
        var printError = "error";

        if (aMsg == "No token available") {
            printError = "authFault";
        }

        if (aMsg == "Invalid Certificate") {
            printError = "certError";
        } else if (aMsg == "Connection failure") {
            printError = "connectFail";
        } else if (aMsg == "Server internal error") {
            printError = "internalErr";
        } else if (aMsg == "Indexing server is unavailable") {
            printError = "solrErr";
        } else if ((aMethodName == "createEvent" || aMethodName == "modifyEvent") &&
            aData != null && aData.wrappedJSObject != null &&
            aData.wrappedJSObject.localEvent != null &&
            aData.wrappedJSObject.localEvent.title != null) {
            return this._getLocalizedMessage("srvErrorExplained") + " " + aData.wrappedJSObject.localEvent.title;
        } else if (/^Read permission denied for /.test(aMsg)) {
            aMsg = aMsg + "";
            this.logger.debug("typeof aMsg: " + typeof aMsg, false, false);
            var lMsg = this._getLocalizedMessage("calendarReadPermissionDenied");
            var email = aMsg.split(" ").pop();
            return lMsg.replace("%s", email);
        }

        return this._getLocalizedMessage(printError);
    },
    _getOBMCredentials: function(aLogin, aPassword) {
        this.logger.debug("_getOBMCredentials()", false, false);

        if (!UtilsOBM.getOBMCredentials(aLogin, aPassword) || !aLogin.value || aLogin.value == "" || !aPassword.value || aPassword.value == "") {
            var error = "NO_LOGIN_INFO";
            this.logger.error(error, true, false);
            this.callerListener.onError(error);
            throw (error);
        }
    },
    _getXPStringFromJSString: function(str) {
        var XPString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
        XPString.data = str;
        return XPString;
    },
    _getJSStringFromXPString: function(XPString) {
        XPString.QueryInterface(Ci.nsISupportsString);
        return XPString.data;
    },

    _getLastSync: function(calendar) {
        var calMgr = cal.getCalendarManager();
        var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
        let lastSync = UtilsOBM.getPrefSafe("extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync", "0");
        // TODO : need an isValidDate func ?
        if (lastSync == "1970-01-01T00:00:00Z" || lastSync == "") {
            lastSync = "0";
        }
        return lastSync;
    },
    _setLastSync: function(calName, lastSync) {
        this.logger.debug("SETLASTSYNC (" + lastSync + ") for " + calName + ": " + "extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync=" + lastSync, false, false);
        try {
            UtilsOBM.setPrefSafe("extensions.obm.changes." + Ci.calICalendar + "." + calName + "." + "lastSync", lastSync);
        } catch (e) {
            // From time to time, the prefs.js is mysteriously locked and saving
            // the preferences fails with an NS_ERROR_FILE_ACCESS_DENIED error
            this.logger.error(e);
            // Complain to the user and ask her to try again
            var errorMsg = this._getLocalizedMessage("prefSaveFailed");
            this.callerListener.onStop(-1, errorMsg);
            throw e;
        }
    },
    _mozEventToOBMEvent: function(aCalendar, aEvent, aWillBeCreated) {
        var calMgr = cal.getCalendarManager();
        var calName = calMgr.getCalendarPref_(aCalendar, "X-OBM-NAME");
        var localEvent = aEvent;
        var eventToSend = {
            QueryInterface: ChromeUtils.generateQI([Ci.obmIREventRecurrence, Ci.obmIREvent]),
            recurrence: this
        };

        // Handle faked master items by sending their single occurrence as a
        // master item to the OBM server.
        if (aEvent && aEvent.hasProperty('X-MOZ-FAKED-MASTER') && aEvent.recurrenceInfo) {
            var exids = aEvent.recurrenceInfo.getExceptionIds({});
            localEvent = aEvent.recurrenceInfo.getExceptionFor(exids[0]);
        }

        if (!aWillBeCreated) {
            eventToSend.uid = UtilsOBM.getOBMEventId(true, localEvent);
        }
        eventToSend.extId = localEvent.id;
        if (localEvent.hasProperty('X-OBMEXT-IS-INTERNAL')) {
            eventToSend.isInternal = true;
        }

        if (localEvent.recurrenceId) {
            eventToSend.recurrenceId = UtilsOBM.jsDateToRestDate(cal.dtz.dateTimeToJsDate(localEvent.recurrenceId));
        }

        this.logger.debug("reading ganaration from localEvent : " + localEvent.generation, false, false);

        eventToSend.generation = UtilsOBM.getSequence(localEvent);
        eventToSend.owner = this._login.split("@")[0];
        eventToSend.ownerEmail = calMgr.getCalendarPref_(aCalendar, "X-OBM-EMAIL");

        var organizer = localEvent.organizer;
        //if (organizer != null && organizer.id != null && organizer.id != "") {
        //  eventToSend.ownerEmail = organizer.id.replace(/MAILTO:/i, "");
        //}
        var calPriority = localEvent.priority;

        if (calPriority >= 1 && calPriority <= 4) {
            eventToSend.priority = 3; // high priority in OBM
        } else if (calPriority == 5) {
            eventToSend.priority = 2; // medium priority in OBM
        } else if (calPriority >= 6 && calPriority <= 9) {
            eventToSend.priority = 1; // low priority in OBM
        } else {
            eventToSend.priority = 2; // medium priority in OBM, by default
        }

        switch (localEvent.privacy) {
            case "CONFIDENTIAL":
                eventToSend.privacy = (UtilsOBM.capabilitySupported("CONFIDENTIAL_EVENTS") ? "2" : "1");
                break;
            case "PRIVATE":
                eventToSend.privacy = "1";
                break;
            case "PUBLIC":
            default:
                eventToSend.privacy = "0";
        }

        eventToSend.allday = localEvent.startDate.isDate;
        eventToSend.date = UtilsOBM.jsDateToRestDate(cal.dtz.dateTimeToJsDate(localEvent.startDate));

        if (eventToSend.allday) {
            let startDate = localEvent.startDate.getInTimezone(cal.dtz.UTC);
            let endDate = localEvent.endDate.getInTimezone(cal.dtz.UTC);

            eventToSend.duration = endDate.subtractDate(startDate).inSeconds;
        } else {
            eventToSend.duration = localEvent.duration.inSeconds;
        }

        let tz = eventToSend.allday ? cal.dtz.defaultTimezone : localEvent.startDate.timezone;
        eventToSend.tz = tz ? tz.tzid : null;

        eventToSend.title = localEvent.title;
        eventToSend.description = localEvent.getProperty("DESCRIPTION");

        let categories = localEvent.getCategories({})
        for (let category of categories) {
            var matches = category.match(RegExp(/\[OBM\] (.*)/));
            if (matches) {
                eventToSend.category = matches[1];
                break;
            }
        }

        eventToSend.location = localEvent.getProperty("LOCATION");

        if (localEvent.getProperty("TRANSP") == "TRANSPARENT") {
            eventToSend.opacity = "TRANSPARENT";
        } else {
            eventToSend.opacity = "OPAQUE";
        }
        eventToSend.opacity = localEvent.getProperty("TRANSP");

        this._mozRecurrenceToOBMEvent(localEvent, eventToSend);

        var allAttendees = localEvent.getAttendees({});
        var goodAttendees = []; // list of attendees to send
        var subcribedEmail = calMgr.getCalendarPref_(aCalendar, "X-OBM-EMAIL");

        var organizerEmail = null;
        var organizerAdded = false; // indicate if the organizer has been found into the list of attendee
        var organizer = localEvent.organizer;
        if (organizer != null && organizer.id != null && organizer.id != "") {
            organizerEmail = organizer.id.replace(RegExp(/MAILTO:/i), "");
        }

        var calAttendee = null; // the calAttendee object corresponding to the user
        for (var i = 0; i < allAttendees.length; i++) {
            if (allAttendees[i].participationStatus) {
                if (allAttendees[i].id.match(RegExp("MAILTO:" + subcribedEmail, "i"))) {
                    calAttendee = allAttendees[i];
                }
                if (organizerEmail != null && allAttendees[i].id.match(RegExp("MAILTO:" + organizerEmail, "i"))) {
                    organizerAdded = true;
                }
                goodAttendees.push(allAttendees[i]);
            }
        }
        var subscribedAdded = false;
        if (calAttendee) {
            subscribedAdded = true;
        }

        if (organizerEmail != null && !organizerAdded) {
            goodAttendees.push(organizer.clone());
            if (organizerEmail == subcribedEmail) {
                subscribedAdded = true;
            }
        }

        eventToSend.attendees = [];
        for (var calAttendee of goodAttendees) {
            var attendee = {};
            attendee.email = calAttendee.id.replace(RegExp(/MAILTO:/i), "");
            attendee.displayName = calAttendee.commonName;
            attendee.status = this._getAcceptableParticipationState(calAttendee.participationStatus);
            switch (calAttendee.role) {
                case "REQ-PARTICIPANT":
                    attendee.type = "REQ";
                    break;
                case "OPT-PARTICIPANT":
                default:
                    attendee.type = "OPT";
                    break;
            }

            if (organizerEmail != null && calAttendee.id.match(RegExp("MAILTO:" + organizerEmail, 'i'))) {
                attendee.isOrganizer = true;
            } else
                attendee.isOrganizer = false;
            eventToSend.attendees.push(attendee);
        }
        if (!subscribedAdded) {
            var attendee = {};
            attendee.email = subcribedEmail;
            attendee.displayName = subcribedEmail;
            attendee.status = "ACCEPTED";
            attendee.type = "REQ";
            if (eventToSend.attendees.length == 0) {
                attendee.isOrganizer = true;
            }
            eventToSend.attendees.push(attendee);
        }

        eventToSend.getAttendees = function(count) {
            count.value = this.attendees.length;
            return this.attendees;
        };

        var calAlert;
        var alarms = localEvent.getAlarms({});
        for (var alarm of alarms) {
            let offset = cal.alarms.calculateAlarmOffset(localEvent, alarm);
            if (offset) {
                offset = offset.clone();
                let dur = localEvent.duration;
                if (alarm.related == alarm.ALARM_RELATED_ABSOLUTE) {
                    offset.isNegative = !offset.isNegative;
                } else if (alarm.related == alarm.ALARM_RELATED_END && dur) {
                    dur = dur.clone();
                    dur.isNegative = false;
                    offset.addDuration(dur);
                }
                calAlert = offset;
                break;
            }
        }

        var calAlertToXmlAlert = function(aCalAlert) {
            if (aCalAlert.inSeconds > 0) {
                return 0; //OBM do not support alert after events
            } else if (aCalAlert.inSeconds == 0) { // OBM does not support 0 seconds delay alerts, we set it to 1 minute (OBMFULL-2655)
                return 60;
            } else {
                return (-1 * aCalAlert.inSeconds);
            }
        };

        var obmAlert = -1;
        if (calAlert) {
            if (localEvent.alarmLastAck) {
                if (localEvent.recurrenceInfo) {
                    obmAlert = calAlertToXmlAlert(calAlert);
                }
            } else {
                obmAlert = calAlertToXmlAlert(calAlert);
            }
        }
        eventToSend.alert = obmAlert;

        //recc exception events
        eventToSend.eventExceptions = [];
        eventToSend.getEventExceptions = function(count) {
            count.value = this.eventExceptions.length;
            return this.eventExceptions;
        };

        if (localEvent.recurrenceInfo) {
            var eventExs = this._mozExceptionsToOBMEvents(localEvent.recurrenceInfo);
            for (var i = 0; i < eventExs.length; i++) {
                this.logger.debug("mozEventEx recId:" + eventExs[i].recurrenceId, false, false);
                eventToSend.eventExceptions[i] = this._mozEventToOBMEvent(aCalendar, eventExs[i], aWillBeCreated);
            }
        }

        return eventToSend;
    },
    _createEventObjFromId: function(calendar, aCalName, aId, aInfos) {
        this.logger.debug("_createEventObjFromId:()", false, false);
        this.logger.debug("  id:             " + aId, false, false);

        var listener = {};
        var self = this;

        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            // <TO_REMOVE>
            // Fix second onOperationComplete call (bug LT 0.8)
            // Fixed in LT 0.9
            try {
                if (aDetail && aDetail.toString() == "Can't deduce item type based on QI") {
                    return;
                }
            } catch (e) {}
            // </TO_REMOVE>
            if (Components.isSuccessCode(aStatus)) {
                var localEvent = this.localEvent;
                var eventToSend = self._mozEventToOBMEvent(calendar, localEvent, this.willBeCreated);
                var ctx = new Context();

                ctx.calendar = calendar;
                ctx.calName = aCalName;
                ctx.id = aId;
                ctx.localEvent = localEvent;
                ctx.eventToSend = eventToSend;
                ctx.notification = aInfos.notification ? aInfos.notification : false;
                ctx.generation = UtilsOBM.getSequence(localEvent);

                if (aInfos.action == "added") {
                    self._createEventOnOBM(ctx);
                } else {
                    self._modifyEventOnOBM(ctx);
                }
            }
        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };

        // WARNING: this method is supposed to be asynchronous !!
        calendar.getItem(aId, listener);
    },

    _createEventOnOBM: function(ctx) {
        let useStoreEvent = UtilsOBM.capabilitySupported("CALENDAR_HANDLER_SUPPORTS_STOREEVENT");

        ctx.methodName = useStoreEvent ? "storeEvent" : "createEvent";

        if (useStoreEvent) {
            this.CalProxy.storeEvent(ctx.calName, ctx.eventToSend, ctx.notification, ctx);
        } else {
            this.CalProxy.createEvent(ctx.calName, ctx.eventToSend, ctx.notification, ctx);
        }
    },

    _modifyEventOnOBM: function(ctx) {
        let useStoreEvent = UtilsOBM.capabilitySupported("CALENDAR_HANDLER_SUPPORTS_STOREEVENT");

        ctx.methodName = useStoreEvent ? "storeEvent" : "modifyEvent";

        if (useStoreEvent) {
            this.CalProxy.storeEvent(ctx.calName, ctx.eventToSend, ctx.notification, ctx);
        } else {
            this.CalProxy.modifyEvent(ctx.calName, ctx.eventToSend, true, ctx.notification, ctx.generation, ctx);
        }
    },

    _mozRecurrenceToOBMEvent: function(aCalEvent, aObmEvent) {
        var recurrenceInfo = aCalEvent.recurrenceInfo;
        var exceptionIDs = [];
        var invalidChoice;
        var endDate = new Date();

        if (recurrenceInfo) {
            var recurrenceCount = recurrenceInfo.countRecurrenceItems();
            for (var i = 0; i < recurrenceCount; i++) {
                var recurrenceItem = recurrenceInfo.getRecurrenceItemAt(i);
                if (recurrenceItem.isNegative && recurrenceItem instanceof Ci.calIRecurrenceDate) {
                    exceptionIDs.push(recurrenceItem.date);
                } else {
                    if (!recurrenceItem.isNegative && recurrenceItem instanceof Ci.calIRecurrenceRule) {

                        if (recurrenceItem.isFinite) {
                            if (recurrenceItem.isByCount) {
                                // end date ==> date of last occurrence
                                var previousOccurence;
                                var nextOccurrence = aCalEvent;

                                do {
                                    previousOccurence = nextOccurrence;
                                    var occurrenceTime = previousOccurence.startDate;
                                    nextOccurrence = recurrenceInfo.getNextOccurrence(occurrenceTime);
                                } while (nextOccurrence != null);

                                endDate = cal.dtz.dateTimeToJsDate(previousOccurence.startDate);
                            } else { // recurrence lasts up to a certain date
                                endDate = cal.dtz.dateTimeToJsDate(recurrenceItem.untilDate);
                            }
                        }

                        switch (recurrenceItem.type) {
                            case "DAILY":

                                var onDays = recurrenceItem.getComponent("BYDAY", {});
                                if (onDays.length > 0) {
                                    //on every weekday
                                    var days = ['0', '0', '0', '0', '0', '0', '0'];
                                    for (var j = 0; j < onDays.length; j++) {
                                        days[onDays[j] - 1] = '1';
                                    }
                                    days = days.join("");

                                    aObmEvent.kind = "weekly";
                                    aObmEvent.frequence = recurrenceItem.interval;
                                    aObmEvent.days = days;
                                    break;
                                }
                                aObmEvent.kind = "daily";
                                aObmEvent.frequence = recurrenceItem.interval;
                                aObmEvent.days = "";
                                break;

                            case "WEEKLY":
                                aObmEvent.kind = "weekly";
                                aObmEvent.frequence = recurrenceItem.interval;
                                var onDays = recurrenceItem.getComponent("BYDAY", {});
                                var days = ['0', '0', '0', '0', '0', '0', '0'];
                                if (!onDays || onDays === null || onDays.length === 0 || onDays == "") {
                                    // lightning bug : no day selected
                                    // -> no day stored but start day of event used in views
                                    days[aCalEvent.startDate.weekday] = '1';
                                } else {
                                    for (var j = 0; j < onDays.length; j++) {
                                        days[onDays[j] - 1] = '1';
                                    }
                                }
                                days = days.join("");
                                // make Sunday day 7 no more needed in obm 2.2
                                //days = days.replace(/(0|1)([0|1]{6})/, "$2$1");
                                aObmEvent.days = days;
                                break;

                            case "MONTHLY":
                                var onDays = recurrenceItem.getComponent("BYDAY", {});
                                //dump("recc MONTHLY, onDays" + onDays[0] + "\n");
                                if (onDays.length == 1) {
                                    aObmEvent.kind = "monthlybyday";
                                    aObmEvent.frequence = recurrenceItem.interval;
                                    aObmEvent.days = "";
                                    break;
                                }
                                onDays = recurrenceItem.getComponent("BYMONTHDAY", {});
                                //dump("recc, onDays" + onDays[0] + "\n");
                                if (onDays.length == 1) {
                                    aObmEvent.kind = "monthlybydate";
                                    aObmEvent.frequence = recurrenceItem.interval;
                                    aObmEvent.days = "";
                                    break;
                                }
                                // lightning bug : no day or date selected
                                // force recc by date
                                aObmEvent.kind = "monthlybydate";
                                aObmEvent.frequence = recurrenceItem.interval;
                                aObmEvent.days = "";
                                //invalidChoice = true;
                                break;
                            case "YEARLY":
                                aObmEvent.kind = "yearly";
                                aObmEvent.frequence = recurrenceItem.interval;
                                aObmEvent.days = "";
                                break;
                            default:
                                invalidChoice = true;
                                break;
                        }
                    }
                }
            }
        }
        aObmEvent.exceptions = [];
        aObmEvent.getExceptions = function(count) {
            count.value = this.exceptions.length;
            return this.exceptions;
        };

        if (!recurrenceInfo || invalidChoice) {
            aObmEvent.kind = "none";
            aObmEvent.frequence = 1;
            aObmEvent.days = "";
        } else {
            for (let exceptionID of exceptionIDs) {
                aObmEvent.exceptions.push(UtilsOBM.jsDateToRestDate(cal.dtz.dateTimeToJsDate(exceptionID)).toString());
            }
        }
        if (!endDate || (recurrenceItem && !recurrenceInfo.isFinite)) {
            aObmEvent.end = null;
        } else {
            aObmEvent.end = UtilsOBM.jsDateToRestDate(endDate);
        }
    },

    _calGetItem: function(calendar, remoteEvent, callback) {
        var self = this;

        var listener = {
            localEvent: null
        };

        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            //<TO_REMOVE>
            // Fix second onOperationComplete call (bug LT 0.8)
            // Fixed in LT 0.9
            try {
                if (aDetail && aDetail.toString() == "Can't deduce item type based on QI") {
                    dump("_calGetItem, aDetail = Can't deduce item type based on QI\n");
                    return;
                }
            } catch (e) {
                dump(e);
            }
            //</TO_REMOVE>
            //dump("--aStatus: " + aStatus + " aOperationType: " + aOperationType + " aId: " + aId + " aDetail: " + aDetail + "\n");
            // FIXME : bug lightning ? aStatus for getItem is allways 0
            if (!Components.isSuccessCode(aStatus) || !this.localEvent) {
                // nothing to do ?
                callback(null);
            } else {
                var localEvent;
                try {
                    localEvent = aDetail.QueryInterface(Ci.calIEvent)
                } catch (e) {
                    localEvent = this.localEvent;
                }

                callback(localEvent);
            }
        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };
        calendar.getItem(remoteEvent.extId, listener);
    },

    _resolveConflicts: function(calName, next) {
        this.logger.debug("_resolveConflicts()", false, false);
        this.logger.debug("  calName: " + calName, false, false);

        var calMgr = cal.getCalendarManager();
        var calendar;
        for (var i = 0; i < this.calendars.length; i++) {
            calendar = this.calendars[i];
            var name = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
            if (name == calName) {
                break;
            }
        }
        if (!this._localDiff[calendar.id]) { this._localDiff[calendar.id] = { event: {} }; }
        //var localDiff = this._localDiff[calendar.id];
        //var _remoteDiff = this._remoteDiff;
        this.logger.debug("  local events : " + JSON.stringify(this._localDiff[calendar.id]), false, false);
        //        var localToSend = {};
        //        var remoteToApply = {participationChanged: [], updated: [], removed: []};
        var self = this;
        var applyConflictsResolution = function(localToSend, remoteToApply) {
            var objLength = function(obj) {
                var i = 0;
                for (var foo in obj) {
                    i++;
                }
                return i;
            }

            this._remoteDiff[calName] = remoteToApply;
            this._localDiff[calendar.id].event = localToSend;
            this.logger.debug("_resolveConflicts: got " +
                objLength(localToSend) + " events to send," +
                remoteToApply.participationChanged.length + " participationChanged to apply, " +
                remoteToApply.updated.length + " updated to apply, " +
                remoteToApply.removed.length + "removed to apply", false, false);
            this.logger.debug("localToSend ? " + JSON.stringify(localToSend), false, false);
            this.logger.debug("end of _resolveConflicts()", false, false);
        };
        var onConflictsResolved = function(localToSend, remoteToApplly) {
            applyConflictsResolution.call(self, localToSend, remoteToApplly);
            next();
        };
        let serverPriority = (UtilsOBM.getPrefSafe("extensions.obm.override", "server") == "server");
        var conflictsResolver = new CalendarConflictResolver(calendar, this, serverPriority);
        conflictsResolver.resolveConflicts(this._localDiff[calendar.id], this._remoteDiff[calName], onConflictsResolved);

    },

    _getEligibleCalendarsForSync: function() {
        if (this._justOneCalendar) {
            return [UtilsOBM.getOBMCalendarByUid(this._selectedCalendarUid)];
        } else {
            return UtilsOBM.getSubscribedOBMCalendar();
        }
    },

    /**
     * the method which begins the real synchronisation, after the login process
     * and the retrieve of the categories list
     * (and the retrieve of list of OBM calendar if it is not already loaded)
     */
    _getLocalDiff: function() {
        this.logger.debug("_getLocalDiff()", false, false);
        var calMgr = cal.getCalendarManager();
        this.callerListener.onNewOperation("getLocalDiff", null);

        for (let calendar of this._getEligibleCalendarsForSync()) {
            let syncRangeChanged = calendar.getProperty("X-OBM-SYNCRANGE-CHANGED");

            if (syncRangeChanged) {
                UtilsOBM.applyRemoteCalendarRights(calendar, UtilsOBM.getCalInfoForLocalCalendar(calendar, calendar.readOnly), true);
            }
        }

        // So that we get an updated list
        this.calendars = this._getEligibleCalendarsForSync();

        for (var i = 0; i < this.calendars.length; i++) {

            var calName = calMgr.getCalendarPref_(this.calendars[i], "X-OBM-NAME");
            var changeTypes = ["added", "updated", "removed", "participationChanged"];
            this._localDiff[this.calendars[i].id] = {};
            this._localDiff[this.calendars[i].id]["event"] = {};
            this._localDiff[this.calendars[i].id]["task"] = {};

            this.logger.debug("_getLocalDiff : parsing " + calName, false, false);

            var allEventsChanges = UtilsOBM.getAllCalendarSyncItems(this.calendars[i]);

            this.logger.debug("_getLocalDiff : allEventsChanges : " + JSON.stringify(allEventsChanges), false, false);
            /** this thing produces events duplication on reccurence exceptions
                        for ( var j in allEventsChanges ) {
                            if ( allEventsChanges[j].action && ( allEventsChanges[j].action == "added" || allEventsChanges[j].action == "updated" ) ) {
                                //Does this thing work ???
                                // WARNING _exceptionToItem calls asynchronous method of calICalendar objects
                                    // FIXME we should implement callbacks here... ?
                                var addedIds = this._exceptionToItem(this.calendars[i], j);
                                for(var id in addedIds) {
                                    allEventsChanges[id] = {action: "added"};
                                }
                            }
                        }
            */
            if (this.logger.getLevelWriter(this.logger.DEBUG_LEVEL)) {
                this.logger.debug("_getLocalDiff: ", false, false);
                for (var j in allEventsChanges) {
                    this.logger.debug("   " + allEventsChanges[j].action + " " + j, false, false);
                }
            }
            this._localDiff[this.calendars[i].id]["event"] = allEventsChanges;
            /*
            this.logger.debug("_getLocalDiff updated: " + this._localDiff[calName].event["updated"], false, false);
            this.logger.debug("_getLocalDiff removed: " + this._localDiff[calName].event["removed"], false, false);
            this.logger.debug("_getLocalDiff added: " + this._localDiff[calName].event["added"], false, false);
            this.logger.debug("_getLocalDiff participationChanged: " + this._localDiff[calName].event["participationChanged"], false, false);
            */
        }

        this.callerListener.onOperationDone(true, null);

        this.callerListener.onNewOperation("getSyncWithSortedChanges", null);
        this._getSyncOperationCount = this.calendars.length;
        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetSync(aResCode, aMethodName, aResult, aData)
        });

        this._localOperations = [];
        var calMgr = cal.getCalendarManager();
        for (var i = 0; i < this.calendars.length; i++) {
            var calendar = this.calendars[i];
            this._localOperations[i] = (
                function(aCal, aIdx, aTotal) {
                    return function() {
                        var aCalName = calMgr.getCalendarPref_(aCal, "X-OBM-NAME");
                        self.callerListener.onProgress(aIdx, aTotal, "", aCalName);
                        var lastSync = self._getLastSync(aCal);
                        var data = new Context();
                        data.methodName = "getSyncWithSortedChanges";
                        data.isTheFirstPass = true;
                        data.calendar = aCal;
                        data.calName = aCalName;
                        data.lastSync = lastSync;
                        self.CalProxy.getSyncWithSortedChanges(aCalName, lastSync, data);
                    }
                }
            )(calendar, i, this.calendars.length);
        }

        this._localOperationsIdx = 0;
        this._sendNextLocalOperation();
    },
    _sendNextLocalOperation: function() {
        var idx = this._localOperationsIdx++;
        this.logger.debug("_sendNextLocalOperation: sizeof array: " + this._localOperations.length + ", index : " + idx, false, false);
        if (idx < this._localOperations.length) {
            this.logger.debug("_sendNextLocalOperation: element: " + JSON.stringify(this._localOperations[idx]), false, false);
        }
        this._localOperations[idx]();
    },
    _mozExceptionsToOBMEvents: function(aReccurenceInfo) {
        var aExceptionIDs = aReccurenceInfo.getExceptionIds({});
        var exceptionEvents = [];
        if (aExceptionIDs.length != 0) {

            for (var aExceptionID of aExceptionIDs) {
                var ocurrence = aReccurenceInfo.getOccurrenceFor(aExceptionID, false);
                this.logger.debug("exception ocurrence:" + aExceptionID, false, false);
                exceptionEvents.push(ocurrence);
            }
        }
        return exceptionEvents;
    },
    _exceptionToItem: function(calendar, aId) {

        this.logger.debug("_exceptionToItem()", false, false);
        this.logger.debug("  calendar: " + calendar, false, false);
        this.logger.debug("  aId: " + aId, false, false);
        var listener = {};
        this.addedIds = [];
        var self = this;
        listener.calendar = calendar;
        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            // <TO_REMOVE>
            // Fix second onOperationComplete call (bug LT 0.8)
            // Fixed in LT 0.9
            try {
                if (aDetail.toString() == "Can't deduce item type based on QI") {
                    return;
                }
            } catch (e) {}
            // </TO_REMOVE>
            if (Components.isSuccessCode(aStatus)) {
                var calMgr = cal.getCalendarManager();
                var calName = calMgr.getCalendarPref_(this.calendar, "X-OBM-NAME");
                var localEvent = this.localEvent;

                //get exception
                try {
                    if (localEvent.recurrenceInfo != null) {
                        var aExceptionIDs = localEvent.recurrenceInfo.getExceptionIds({});
                        // has Exception
                        if (aExceptionIDs.length != 0) {

                            for (var aExceptionID of aExceptionIDs) {
                                var ocurrence = localEvent.recurrenceInfo.getOccurrenceFor(aExceptionID, false);

                                var calEvent = cal.createEvent();
                                var uuidgen = Components.classes["@mozilla.org/uuid-generator;1"].getService(Components.interfaces.nsIUUIDGenerator);
                                var uuid = uuidgen.generateUUID().toString();

                                calEvent.calendar = this.calendar;
                                calEvent.id = uuid.substring(1, uuid.length - 1);
                                calEvent.title = ocurrence.title;
                                calEvent.setProperty("DESCRIPTION", ocurrence.getProperty("DESCRIPTION"));
                                calEvent.setProperty("LOCATION", ocurrence.getProperty("LOCATION"));
                                calEvent.startDate = ocurrence.startDate;
                                calEvent.endDate = ocurrence.endDate;
                                calEvent.setProperty("CATEGORIES", ocurrence.getProperty("CATEGORIES"));
                                calEvent.priority = ocurrence.priority;
                                calEvent.privacy = ocurrence.privacy;
                                calEvent.setProperty("TRANSP", localEvent.getProperty("TRANSP"));

                                var attendees = ocurrence.getAttendees({});
                                for (var attendee of attendees) {
                                    calEvent.addAttendee(attendee);
                                }

                                var listenerAdd = {};
                                var caller = this;
                                listenerAdd.pcaller = self;
                                listenerAdd.onOperationComplete = function(cal, res, op, id, item) {
                                    this.pcaller.addedIds.push(id);
                                };
                                this.calendar.adoptItem(calEvent, listenerAdd);
                            } //end for
                        }
                    }
                } catch (e) {
                    self.logger.debug("error getting exception items: " + e, false, false);
                }
            }
        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };

        // WARNING this method is asynchronous, and here, we assume that all things
        // append synchronously !!!!
        calendar.getItem(aId, listener);
        return this.addedIds;
    },
    _onReLogged: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onReLogged()", false, false);
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "doLogin")) {
            return;
        }

        var data = aData.wrappedJSObject;
        var self = this;
        // recall method with params in aData
        switch (data.methodName) {
            case "getSyncWithSortedChanges":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onGetSync(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.getSyncWithSortedChanges(data.calName, data.lastSync, aData);
                break;
            case "createEvent":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.createEvent(data.calName, data.eventToSend, data.notification, aData);
                break;
            case "storeEvent":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.storeEvent(data.calName, data.eventToSend, data.notification, aData);
                break;
            case "modifyEvent":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.modifyEvent(data.calName, data.eventToSend, true, data.notification, aData);
                break;
            case "changeParticipationState":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.changeParticipationState(data.calName, data.eventToSend, data.participationStatus, data.notification, data.generation, data.recurrenceId, aData);
                break;
            case "removeEvent":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.removeEvent(data.calName, data.id, data.notification, data.generation, aData);
                break;
            case "removeEventByExtId":
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
                });
                this.CalProxy.removeEventByExtId(data.calName, data.id, data.notification, data.generation, aData);
                break;
            default:
                //
        }
    },
    _ensureValidResponse: function(aResCode, aMethodName, aResult, aValidMethodName, aData) {
        this.logger.debug("_ensureValidResponse(" + aMethodName + ")", false, false);
        if (aResCode != Cr.NS_OK) {
            var msgError = "unknown";
            if (aResult instanceof Ci.nsISupportsString || aResult instanceof PermissionException) {
                msgError = aResult;
            }

            if (msgError == "Invalid access token") {
                this.logger.debug("Invalid access token -> re-login and retry", false, false);
                // try to re-login if ok re-call method
                var login = {};
                var password = {};
                this._getOBMCredentials(login, password);

                var self = this;
                this.CalProxy.removeListener();
                this._addListener(function(aResCode, aMethodName, aResult, aData) {
                    self._onReLogged(aResCode, aMethodName, aResult, aData);
                });

                this.CalProxy.doLogin(login.value, password.value, UtilsOBM.getOrigin(), aData);
            } else if (msgError == "NotAllowed") {
                let calendar = aData.wrappedJSObject.calendar;

                if (aMethodName == "getSyncWithSortedChanges") {
                    this._calendarsToRemove.push(calendar.name);
                } else {
                    this._calendarsNowReadOnly.push(calendar.name);
                }

                return false;
            } else if (msgError instanceof PermissionException) {
                var localizedRight = UtilsOBM.getLocalizedMessage("obm.sync.rights." + msgError.right.toLowerCase());
                var msgParams = [msgError.calendar, localizedRight, msgError.eventTitle];
                var message = UtilsOBM.getLocalizedMessageWithParams("obm.syncerror.permissionException", msgParams);
                if (this.callerListener) {
                    if (this._isCriticalMethod(aMethodName)) {
                        this.callerListener.onStop(aResCode, message);
                    } else {
                        this.callerListener.onError(message);
                    }
                }
            } else {
                var printError = this._printableError(aMethodName, msgError, aData);

                //bypass error on getCalendarMetadata
                if (aMethodName == "getCalendarMetadata") { return false; }
                if (aMethodName == "getResourceMetadata") { return false; }

                if (msgError == "Event did not exist." && aMethodName == "removeEventByExtId") {
                    //normal error: we try to remove an event that doesn't exist anymore on server
                    return true;
                }

                this.logger.error(aMethodName + ": Reported error: " + msgError + "(" + aResCode + ")", true, false);

                if (this.callerListener) {
                    if (this._isCriticalMethod(aMethodName)) {
                        this.callerListener.onStop(aResCode, printError);
                    } else {
                        this.callerListener.onError(printError);
                    }
                }
            }
            return false;
        }

        if (aValidMethodName && aMethodName != aValidMethodName) {
            var msgError = "Method " + aMethodName + " unexpected. Expected: " + aValidMethodName;
            this.logger.error(msgError, true, false);
            this.callerListener.onError(msgError);
            return false;
        }
        this.logger.debug("_ensureValidResponse() returning true", false, false);
        return true;
    },

    /**
     * callback called when the login process is done.
     *
     * @FIXME: it is called for web service calls ! No token ??
     */
    _onLogged: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onLogged()", false, false);
        var self = this;
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "doLogin")) {
            return;
        };
        this.callerListener.onOperationDone(true, null);

        UtilsOBM.setPrefSafe("extensions.obm.current.sync.version", this._getJSStringFromXPString(aResult));
        if (this._onlyListCalendars) {
            this.callerListener.onNewOperation("listCalendars", null);
            var self = this;
            this._addListener(function(aResCode, aMethodName, aResult, aData) {
                self._onListCalendar(aResCode, aMethodName, aResult, aData);
            });
            this.CalProxy.listCalendars(this._paginer.limit, this._paginer.offset, this._paginer.pattern, null);
        } else if (this._onlyListResources) {
            this.callerListener.onNewOperation("listResources", null);
            var self = this;
            this._addListener(function(aResCode, aMethodName, aResult, aData) {
                self._onListResources(aResCode, aMethodName, aResult, aData);
            });
            this.CalProxy.listResources(this._paginer.limit, this._paginer.offset, this._paginer.pattern, null);
        } else {
            this._callGetResourceMetadata();
        }
    },
    _getResourceEmails: function() {
        this.logger.debug("_getResourceEmails()", false, false);
        var resMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
        var resEmails = [];
        var resources = UtilsOBM.getSubscribedOBMResource();
        for (var i = 0; i < resources.length; i++) {
            var res = resources[i];
            var resEmail = resMgr.getCalendarPref_(res, "X-OBM-RESOURCE-EMAIL");
            resEmails.push(resEmail);
        }
        return resEmails;
    },
    _getCalendarEmails: function() {
        // Returns the names of the current OBM calendars
        var calMgr = cal.getCalendarManager();
        var calEmails = [];
        var calendars = UtilsOBM.getSubscribedOBMCalendar();
        for (var i = 0; i < calendars.length; i++) {
            var calendar = calendars[i];
            var calEmail = calMgr.getCalendarPref_(calendar, "X-OBM-EMAIL");
            calEmails.push(calEmail);
        }
        return calEmails;
    },
    _getResourceMetadata: function(calendarEmails) {
        var resourceCalendars = UtilsOBM.getSubscribedOBMResource();
        if (resourceCalendars.length == 0) {
            this._onResourceMetadataEnd();
            return;
        }
        this.callerListener.onNewOperation("getResourceMetadata", {});
        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetResourceMetadata(aResCode, aMethodName, aResult, aData);
        });

        this.CalProxy.getResourceMetadata(calendarEmails, calendarEmails.length, {});
    },
    _getCalendarMetadata: function(calendarEmails) {
        // Move directly to listCategories() if getCalendarMetadata() is disabled by configuration...
        // We also need to ensure that the server supports the new NotAllowedException. If it isn't,
        // we must force the use of getCalendarMetadata() every time
        if (!UtilsOBM.getPrefSafe("extensions.obm.rights.useGetCalendarMetadata", false) &&
            UtilsOBM.capabilitySupported("CALENDAR_HANDLER_SUPPORTS_NOTALLOWEDEXCEPTION")) {

            this._listCategories();
            return;
        }

        // Retrieves the calendar metadata (mostly permissions) in order to
        // compare them with the permissions on the current calendars.
        // Permissions on the calendars in Lightning are adjusted accordingly,
        // and calendars may be removed if the read right has been removed.
        this.callerListener.onNewOperation("getCalendarMetadata", {});
        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetCalendarMetadata(aResCode, aMethodName, aResult, aData);
        });

        this.CalProxy.getCalendarMetadata(calendarEmails, calendarEmails.length, {});
    },
    _onGetCalendarMetadata: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onGetCalendarMetadata", false, false);
        this.callerListener.onOperationDone(true, aData);
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "getCalendarMetadata")) {
            // Keep compatibility with older versions of obm-sync which don't
            // implement getCalendarMetadata
            this._listCategories();
            return;
        };
        var calendarInfos = aResult.QueryInterface(Ci.nsIArray);
        var calendarManager = cal.getCalendarManager();
        var localCalendars = UtilsOBM.getSubscribedOBMCalendar();
        var remoteCalendarsHash = this._getRemoteCalendarsAsHash(calendarInfos);

        for (var localCalendar of localCalendars) {
            var localCalendarEmail = calendarManager.getCalendarPref_(localCalendar, "X-OBM-EMAIL");
            var remoteCalInfo = remoteCalendarsHash[localCalendarEmail];
            if (remoteCalInfo != null) {
                var remoteCalHasReadRights = remoteCalInfo.read;

                // The read right has been removed in OBM UI, remove the calendar
                // altogether
                if (!remoteCalHasReadRights) {
                    calendarManager.unregisterCalendar(localCalendar);
                    calendarManager.removeCalendar(localCalendar);
                } else {
                    UtilsOBM.applyRemoteCalendarRights(localCalendar, remoteCalInfo);
                }
            } else {
                calendarManager.unregisterCalendar(localCalendar);
                calendarManager.removeCalendar(localCalendar);
            }
        }
        this._listCategories();
    },

    _callGetResourceMetadata: function() {
        this.logger.debug("_callGetResourceMetadata()", false, false);
        UtilsOBM.log(`[OBM Calendar Sync] : _callGetResourceMetadata`)
        var ResourceEmails = this._getResourceEmails();
        this.callerListener.onNewOperation("getResourceMetadata", null);
        this._getResourceMetadata(ResourceEmails);
    },
    _onGetResourceMetadata: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onGetResourceMetadata", false, false);
        this.callerListener.onOperationDone(true, aData);
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "getResourceMetadata")) {
            this._onResourceMetadataEnd();
            return;
        };
        var resourceInfos = aResult.QueryInterface(Ci.nsIArray);
        var calendarManager = cal.getCalendarManager();
        var localResources = UtilsOBM.getSubscribedOBMResource();
        var remoteResourcesHash = this._getRemoteResourcesAsHash(resourceInfos);
        for (var localResource of localResources) {
            var localResourceEmail = calendarManager.getCalendarPref_(localResource, "X-OBM-RESOURCE-EMAIL");
            var remoteResInfo = remoteResourcesHash[localResourceEmail];
            if (remoteResInfo != null) {
                var remoteCalHasReadRights = remoteResInfo.read;
                if (!remoteCalHasReadRights) {
                    calendarManager.unregisterCalendar(localResource);
                    calendarManager.removeCalendar(localResource);
                } else {
                    this._setRemoteCalendarName(localResource, remoteResInfo);
                }
            } else {
                calendarManager.unregisterCalendar(localResource);
                calendarManager.removeCalendar(localResource);
            }
        }
        this._onResourceMetadataEnd();
    },
    _onResourceMetadataEnd: function() {
        if (this._OBMCalendarsArePresent()) {
            var calendarEmails = this._getCalendarEmails();
            this.callerListener.onNewOperation("getCalendarMetadata", null);
            this._getCalendarMetadata(calendarEmails);
        } else {
            this.callerListener.onNewOperation("listCalendars", null);
            var self = this;
            this._addListener(function(aResCode, aMethodName, aResult, aData) {
                self._onListCalendar(aResCode, aMethodName, aResult, aData);
            });
            this.CalProxy.listCalendars(-1, null, null, null);
        }
    },
    _setRemoteCalendarName: function(localCalendar, remoteCalendar) {
        var currentName = localCalendar.name;
        var defaultName = localCalendar.getProperty("X-OBM-DEFAULT-NAME");
        // Check if current name it's not set manually by user
        if (currentName == defaultName) {
            var newname = "[" + UtilsOBM.getLocalizedMessage("obm.resource.label") + "] " + remoteCalendar.name;
            localCalendar.name = newname;
            localCalendar.setProperty("X-OBM-DEFAULT-NAME", newname);
        }
    },
    /**
     * callback called for calendar, after the CalProxy.getSync call
     */
    _onGetSync: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onGetSync()", false, false);

        var data = aData.wrappedJSObject;
        var calendar = data.calendar;
        var calName = data.calName;

        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "getSyncWithSortedChanges", aData)) {
            // If this isn't a "lost read rights" error, we simply break here
            if (!(aResult instanceof Ci.nsISupportsString) || aResult != "NotAllowed") {
                return;
            }

            this._getSyncOperationCount--;

            let calMgr = Cc["@mozilla.org/calendar/manager;1"]
                .getService(Ci.calICalendarManager);

            calMgr.unregisterCalendar(calendar);
            calMgr.remove(calendar);

            // We need to amend this.calendars as the calendar was "manually" removed
            for (let i in this.calendars) {
                if (this.calendars[i] == calendar) {
                    this.calendars.splice(i, 1);
                    break;
                }
            }

            if (this._getSyncOperationCount <= 0) {
                this.CalProxy.removeListener();
                this.callerListener.onOperationDone(true, null);
                this._mergeRemoteDiff(data.isTheFirstPass);
            } else {
                this._sendNextLocalOperation();
            }
        } else {
            var syncResponse = aResult;

            this._remoteDiff[calName] = {};
            this._remoteDiff[calName]["updated"] = syncResponse.getUpdatedEvents({}); // list of EventShell
            this._remoteDiff[calName]["participationChanged"] = syncResponse.getParticipationChanges({}); // list of ParticipationChangeShell

            var lastSync = this._getLastSync(calendar);

            if (lastSync != "0") {
                this._remoteDiff[calName]["removed"] = syncResponse.getRemovedEventExtIds({}); // list of ids
            } else {
                this._remoteDiff[calName]["removed"] = [];
            }

            var self = this;
            var continueGetSync = function() {
                self._getSyncOperationCount--;
                self._remoteDiff[calName].lastSync = syncResponse.lastsync;

                if (self._getSyncOperationCount <= 0) {
                    self.CalProxy.removeListener();
                    self.callerListener.onOperationDone(true, null);
                    self._mergeRemoteDiff(data.isTheFirstPass);
                } else {
                    self._sendNextLocalOperation();
                }
            };

            if (data.isTheFirstPass) {
                this._resolveConflicts(calName, continueGetSync);
            } else {
                continueGetSync();
            }
        };
    },
    _setEventExceptions: function(aRecurrence, aCalEvent, aRecurrenceInfo, lastSync) {
        var eventExceptions = aRecurrence.getEventExceptions({});
        for (var eventException of eventExceptions) {
            try {
                let item = this._createLocalEvent(eventException, aCalEvent.calendar, aCalEvent, lastSync);
                aRecurrenceInfo.modifyException(item, false);
            } catch (e) {
                this.logger.error(e, true, false);
            }
        }
    },
    /**
     * Traduit une récurrence du format OBM au format Calendar.
     *
     * @param {obmIREventRecurrence}
     *            aRecurrence
     * @pram {calIEvent} aCalEvent L'évènement auquel la récurrence est associée
     * @type calIRecurrence
     */
    _convertRecurrence: function(aRecurrence, aCalEvent, lastSync) {
        this.logger.debug("  _convertRecurrence()");
        this.logger.debug("    - kind: " + aRecurrence.kind);
        this.logger.debug("    - freq: " + aRecurrence.frequence);
        this.logger.debug("    - days: " + aRecurrence.days);
        this.logger.debug("    - end: " + aRecurrence.end);

        var recurrenceInfo = Cc["@mozilla.org/calendar/recurrence-info;1"].createInstance(Ci.calIRecurrenceInfo);
        recurrenceInfo.item = aCalEvent;

        var recRule = this._convertRecurrenceRule(aRecurrence, aCalEvent);
        if (recRule == null) {
            return null;
        }
        if (!aRecurrence.end || aRecurrence.end == "") {
            // recurrenceInfo.isFinite = false;
        } else {
            this._setEndOfRecurrenceRule(aRecurrence, aCalEvent, recRule);
        }
        recurrenceInfo.insertRecurrenceItemAt(recRule, 0);

        this._setNegativeExceptions(aRecurrence, aCalEvent, recurrenceInfo);

        // add eventExceptions
        this._setEventExceptions(aRecurrence, aCalEvent, recurrenceInfo, lastSync);

        return recurrenceInfo;
    },
    _convertRecurrenceRule: function(aRecurrence, aCalEvent) {
        var recRule = Cc["@mozilla.org/calendar/recurrence-rule;1"].createInstance(Ci.calIRecurrenceRule);
        var frequency = aRecurrence.frequence ? aRecurrence.frequence : 1;
        switch (aRecurrence.kind) {
            case "daily":
                recRule.type = "DAILY";
                recRule.interval = frequency;
                break;

            case "weekly":
                recRule.type = "WEEKLY";
                recRule.interval = frequency;

                var onDays = [];

                var onDaysOBM = aRecurrence.days.split("");
                //onDaysOBM.unshift(onDaysOBM.pop());
                for (var i = 0; i < onDaysOBM.length; i++) {
                    if (onDaysOBM[i] == "1") {
                        onDays.push((i + 1).toString());
                    }
                }

                if (onDays.length > 0) {
                    recRule.setComponent("BYDAY", onDays.length, onDays);
                }
                break;

            case "monthlybyday":
                var sign = 1,
                    ordWeek = Math.ceil(aCalEvent.startDate.day / 7);

                // OBM only supports "the last" and not "the fifth" recurrence style
                // See OLC-79 for details
                if (ordWeek === 5) {
                    ordWeek = 1;
                    sign = -1;
                }

                this.logger.debug("      _convertRecurrenceRule()");
                this.logger.debug("        - monthlybyday: " + sign * (ordWeek * 8 + aCalEvent.startDate.weekday + 1));

                recRule.type = "MONTHLY";
                recRule.interval = frequency;
                recRule.setComponent("BYDAY", 1, [sign * (ordWeek * 8 + aCalEvent.startDate.weekday + 1)]);
                break;

            case "monthlybydate":
                recRule.type = "MONTHLY";
                recRule.interval = frequency;
                recRule.setComponent("BYMONTHDAY", 1, [aCalEvent.startDate.day]);
                break;

            case "yearly":
                recRule.type = "YEARLY";
                recRule.interval = frequency;
                break;

            case "none":
            default:
                return null;
                break;
        }
        return recRule;
    },
    _setEndOfRecurrenceRule: function(aRecurrence, aCalEvent, aRecRule) {
        let defaultZone = cal.dtz.defaultTimezone;
        let endDate = cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(aRecurrence.end)).getInTimezone(defaultZone);
        endDate.timezone = defaultZone;
        if (!aCalEvent.startDate.isDate) {
            endDate.hour = aCalEvent.startDate.hour;
            endDate.minute = aCalEvent.startDate.minute;
            endDate.second = aCalEvent.startDate.second;
        } else {
            endDate.isDate = true;
        }
        dump("endDate.day:" + endDate + "\n");
        try {
            aRecRule.endDate = endDate;
        } catch (e) {
            aRecRule.untilDate = endDate;
        }
    },
    _setNegativeExceptions: function(aRecurrence, aCalEvent, aRecurrenceInfo) {
        // first, remove all current exception IDs
        var exceptionIDs = aRecurrenceInfo.getExceptionIds({});
        for (var exceptionID of exceptionIDs) {
            aRecurrenceInfo.removeExceptionFor(exceptionID);
        }

        // now, add all "new" exception IDs
        let exceptions = aRecurrence.getExceptions({});
        let defaultZone = cal.dtz.defaultTimezone;
        for (var exception of exceptions) {
            var exceptionID = cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(exception)).getInTimezone(defaultZone);
            exceptionID.timezone = defaultZone;

            if (aCalEvent.startDate.isDate) { // event is all day
                exceptionID.isDate = true;
            }

            try {
                aRecurrenceInfo.removeOccurrenceAt(exceptionID); // getExceptionFor(exceptionID,
                // true /* aCreate */);
            } catch (e) {
                // in case exception date isn't a valid occurrence date of the recurrence
                // date set
            }
        }
    },
    _setAnonymizedEventExceptions: function(aRecurrence, aCalEvent, aRecurrenceInfo, lastSync) {
        var eventExceptions = aRecurrence.getEventExceptions({});
        this.logger.debug("eventExceptions:" + eventExceptions.length, false, false);
        for (var eventException of eventExceptions) {
            try {
                aRecurrenceInfo.modifyException(this._createLocalPrivateEvent(eventException,
                    aCalEvent.calendar, aCalEvent, lastSync), false);
            } catch (e) {
                this.logger.error(e, true, false);
            }
        }
    },
    _convertAnonymizedRecurrence: function(aRecurrence, aCalEvent, lastSync) {
        var recurrenceInfo = Cc["@mozilla.org/calendar/recurrence-info;1"].createInstance(Ci.calIRecurrenceInfo);
        recurrenceInfo.item = aCalEvent;

        var recRule = this._convertRecurrenceRule(aRecurrence, aCalEvent);
        if (recRule == null) {
            return null;
        }
        if (!aRecurrence.end || aRecurrence.end == "") {
            // recurrenceInfo.isFinite = false;
            console.log(aRecurrence);
        } else {
            this._setEndOfRecurrenceRule(aRecurrence, aCalEvent, recRule);
        }
        recurrenceInfo.insertRecurrenceItemAt(recRule, 0);

        this._setNegativeExceptions(aRecurrence, aCalEvent, recurrenceInfo);

        // add eventExceptions
        this._setAnonymizedEventExceptions(aRecurrence, aCalEvent, recurrenceInfo, lastSync);

        return recurrenceInfo;
    },
    _convertStartDate: function(aEvent) {
        let start = cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(aEvent.date));

        start = start.getInTimezone(UtilsOBM.getTimezoneOrUTC(aEvent.tz));
        start.isDate = aEvent.allday;

        return start;
    },
    _convertEndDate: function(aEvent) {
        let end = this._convertStartDate(aEvent);
        let duration = cal.createDuration();

        duration.inSeconds = aEvent.duration;
        end = end.getInTimezone(cal.dtz.UTC);
        end.addDuration(duration);

        return end.getInTimezone(UtilsOBM.getTimezoneOrUTC(aEvent.tz));
    },
    _convertRecurrenceId: function(aEvent) {
        let defaultZone = cal.dtz.defaultTimezone;
        let jsRecIdDateTime = cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(aEvent.recurrenceId)).getInTimezone(defaultZone);
        jsRecIdDateTime.timezone = defaultZone;
        return jsRecIdDateTime;
    },
    _convertPriority: function(aEvent) {
        // The priority scale used in OBM's internal storage is different (and
        // inverted) from Lightning's own
        let defaultLightningPriority = 5;
        let obmToLightningPriorities = {
            1: 9,
            2: 5,
            3: 1,
        };
        let convertedPriority = obmToLightningPriorities[aEvent.priority];
        return convertedPriority != null ? convertedPriority : defaultLightningPriority;
    },
    _convertAttendee: function(aAttendee) {
        let calAttendee = cal.createAttendee();
        calAttendee.id = "MAILTO:" + aAttendee.email;
        calAttendee.commonName = aAttendee.displayName;
        calAttendee.participationStatus = this._getAcceptableParticipationState(aAttendee.status);
        switch (aAttendee.type) {
            case "REQ":
                calAttendee.role = "REQ-PARTICIPANT";
                break;
            default:
                calAttendee.role = "OPT-PARTICIPANT";
        }
        calAttendee.isOrganizer = aAttendee.isOrganizer;
        return calAttendee;
    },
    _createLocalPrivateEvent: function(aEvent, calendar, aParentItem, lastSync) {
        this.logger.debug("_createLocalPrivateEvent()", false, false);
        this.logger.debug("  event:");
        this.logger.debug("    - title: " + aEvent.title);
        this.logger.debug("    - generation: " + aEvent.generation);
        this.logger.debug("    - hash: " + aEvent.hash);

        let calEvent = aParentItem ? aParentItem.clone() : cal.createEvent();
        if (aParentItem) {
            calEvent.parentItem = aParentItem;
        } else {
            if (aEvent.uid) UtilsOBM.setOBMEventId(true, calEvent, aEvent.uid);
            if (aEvent.extId) calEvent.id = aEvent.extId;
        }

        calEvent.startDate = this._convertStartDate(aEvent);
        calEvent.endDate = this._convertEndDate(aEvent);

        calEvent.calendar = calendar;

        calEvent.title = this.stringBundle.GetStringFromName("obm.private_event_title");
        calEvent.setProperty("DESCRIPTION", this.stringBundle.GetStringFromName("obm.private_event_description"));
        calEvent.setProperty("LOCATION", this.stringBundle.GetStringFromName("obm.private_event_location"));

        UtilsOBM.setSequence(calEvent, aEvent.generation);

        this._setChangeOrigin(calEvent);

        // This is a private event, it can only be private or confidential
        const privacyMap = { 1: "PRIVATE", 2: "CONFIDENTIAL" };
        calEvent.privacy = (aEvent.privacy in privacyMap ? privacyMap[aEvent.privacy] : "PRIVATE");

        calEvent.priority = this._convertPriority(aEvent);

        calEvent.setProperty("X-MOZ-SEND-INVITATIONS", "TRUE");

        if (aEvent.opacity != "") {
            calEvent.setProperty("TRANSP", aEvent.opacity);
        }
        if (aEvent.isInternal) {
            calEvent.setProperty("X-OBMEXT-IS-INTERNAL", "YES");
        }
        let calRecurrence = this._convertAnonymizedRecurrence(aEvent.recurrence, calEvent, calendar, lastSync);
        if (calRecurrence) {
            calEvent.recurrenceInfo = calRecurrence;
        }

        calEvent.removeAllAttendees();
        let attendees = aEvent.getAttendees({});
        let nowDateTime = UtilsOBM.getDateTime(lastSync).getInTimezone(cal.dtz.UTC);

        for (let attendeeElem of attendees) {
            let calAttendee = this._convertAttendee(attendeeElem);
            calAttendee.setProperty("RECEIVED-DTSTAMP", nowDateTime.icalString);
            if (calAttendee.isOrganizer) {
                calEvent.organizer = calAttendee;
            }
        }

        if (aParentItem) {
            calEvent.recurrenceId = this._convertRecurrenceId(aEvent);
        }
        this._setLastModifiedAndDtstamp(aEvent, calEvent);
        this._setEventHash(aEvent, calEvent);

        return calEvent;
    },
    _createLocalEvent: function(aEvent /* obmIREvent */ , calendar, aParentItem, lastSync) {
        this.logger.debug("_createLocalEvent()", false, false);
        this.logger.debug("  event:", false, false);
        this.logger.debug("    - title: " + aEvent.title, false, false);
        this.logger.debug("    - generation: " + aEvent.generation, false, false);
        this.logger.debug("    - hash: " + aEvent.hash);

        if (aParentItem) {
            var calEvent = aParentItem.clone();
        } else {
            var calEvent = cal.createEvent();
        }
        calEvent.calendar = calendar;
        if (aParentItem) {
            calEvent.parentItem = aParentItem;
        } else {
            if (aEvent.uid) UtilsOBM.setOBMEventId(true, calEvent, aEvent.uid);
            if (aEvent.extId) calEvent.id = aEvent.extId;
        }

        calEvent.title = aEvent.title;
        UtilsOBM.setSequence(calEvent, aEvent.generation);
        if (aEvent.description != "") {
            calEvent.setProperty("DESCRIPTION", aEvent.description);
        }
        if (aEvent.location != "") {
            calEvent.setProperty("LOCATION", aEvent.location);
        }
        if (aEvent.opacity != "") {
            calEvent.setProperty("TRANSP", aEvent.opacity);
        }
        calEvent.startDate = this._convertStartDate(aEvent);
        calEvent.endDate = this._convertEndDate(aEvent);

        if (aParentItem) {
            calEvent.recurrenceId = this._convertRecurrenceId(aEvent);
        }

        if (aEvent.category && aEvent.category != "") {
            if (calEvent.setCategories) {
                // lightning 0.9 support
                calEvent.setCategories(1, ["[OBM] " + aEvent.category]);
            } else {
                calEvent.setProperty("CATEGORIES", "[OBM] " + aEvent.category);
            }
        }

        calEvent.priority = this._convertPriority(aEvent);

        const privacyMap = { 1: "PRIVATE", 2: "CONFIDENTIAL", 0: "PUBLIC" };
        calEvent.privacy = (aEvent.privacy in privacyMap ? privacyMap[aEvent.privacy] : "PUBLIC");

        calEvent.removeAllAttendees();
        let attendees = aEvent.getAttendees({});
        let nowDateTime = UtilsOBM.getDateTime(lastSync).getInTimezone(cal.dtz.UTC);

        for (let attendeeElem of attendees) {
            this.logger.debug("_createLocalEvent: new attendee found: " + attendeeElem.email, false, false);
            let calAttendee = this._convertAttendee(attendeeElem);
            calAttendee.setProperty("RECEIVED-DTSTAMP", nowDateTime.icalString);
            if (calAttendee.isOrganizer) {
                calEvent.organizer = calAttendee;
            } else {
                calEvent.addAttendee(calAttendee);
            }
        }

        let calRecurrence = this._convertRecurrence(aEvent.recurrence, calEvent, calendar, lastSync);
        if (calRecurrence) {
            calEvent.recurrenceInfo = calRecurrence;
        }

        if (calEvent.alarmOffset === undefined) {
            calEvent.clearAlarms();
        }

        // here calendar.readOnly is false, even if the calendar is read-only.
        // if it wasn't, we couldn't add events to readonly calendars from the synchro
        let calendarIsReadOnly = UtilsOBM.getPrefSafe("extensions.obm.changes.calICalendar." +
            UtilsOBM.getCalendarOBMName(calendar) +
            ".readOnly", true);
        this.logger.debug("_createLocalEvent: Calendar " + calendar.uri.spec + " read-only ? " + calendarIsReadOnly, false, false);

        if (!calendarIsReadOnly && aEvent.alert != null && aEvent.alert != -1 && aEvent.alert != 0) {
            let dur = cal.createDuration();
            dur.inSeconds = -1 * aEvent.alert; // Alarm before event starts -> negative duration | OBM manage positive durations
            if (calEvent.alarmOffset === undefined) {
                let alarm = cal.createAlarm();
                alarm.related = Ci.calIAlarm.ALARM_RELATED_START;
                alarm.offset = dur;
                calEvent.addAlarm(alarm);
            } else {
                calEvent.alarmOffset = dur;
            }
            if (calEvent.recurrenceInfo) {
                calEvent.alarmLastAck = cal.now();
            }
        }

        if (aEvent.isInternal)
            calEvent.setProperty("X-OBMEXT-IS-INTERNAL", "YES");

        // FIX M#2628
        calEvent.setProperty("X-MOZ-SEND-INVITATIONS", "TRUE");
        this._setChangeOrigin(calEvent);
        this._setLastModifiedAndDtstamp(aEvent, calEvent);
        this._setEventHash(aEvent, calEvent);
        this.logger.debug("_createLocalEvent: Event created.", false, false);
        this.logger.debug("_createLocalEvent: event has " + aEvent.getAttendees({}).length + " attendees", false, false);

        return calEvent;
    },

    _setEventHash: function(aEvent, calEvent) {
        if (aEvent.hash) {
            calEvent.setProperty("X-OBM-HASH", aEvent.hash);
        }
    },

    _setLastModifiedAndDtstamp: function(aEvent, calEvent) {
        // The DTSTAMP and LAST-MODIFIED stamps MUST be in UTC per rfc5545, please keep it that way.
        let jsLastModifiedDate = cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(aEvent.timeupdate)).getInTimezone(cal.dtz.UTC);
        calEvent.setProperty("DTSTAMP", jsLastModifiedDate);
        calEvent.setProperty("LAST-MODIFIED", jsLastModifiedDate); // needs to be the last modified property !!!
    },
    _setChangeOrigin: function(calEvent) {
        calEvent.setProperty("X-OBM-ORIGIN", "obm-sync");
        if (calEvent.recurrenceInfo != null) {
            for (var exceptionId of calEvent.recurrenceInfo.getExceptionIds({})) {
                this._setChangeOrigin(calEvent.recurrenceInfo.getOccurrenceFor(exceptionId));
            }
        }
    },

    _updateOrAdoptEvent: function(calendar, event, isTheFirstPass, callback) {
        this.logger.debug("_updateOrAdoptEvent()", false, false);
        this.logger.debug("  calendar: " + calendar.type, false, false);
        this.logger.debug("  event:", false, false);
        this.logger.debug("    - title: " + event.title, false, false);
        this.logger.debug("    - id:   " + event.id, false, false);
        this.logger.debug("    - generation: " + event.generation, false, false);
        var atts = event.getAttendees({});
        for (var att of atts) {
            this.logger.debug("    - attendee: " + att.id + ": " + att.participationStatus, false, false);
        }

        var listener = {};
        var self = this;
        //if (notify) listener.isTheFirstPass = isTheFirstPass; // Closure should be enough
        //listener.newItem = event;
        //listener.calendar = calendar;
        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            var modifyListener = getListener({
                complete: function() {
                    if (callback) { callback(); }
                }
            });
            //<TO_REMOVE>
            // Fix second onOperationComplete call (bug LT 0.8)
            // Fixed in LT 0.9
            try {
                if (aDetail && aDetail.toString() == "Can't deduce item type based on QI") {
                    dump("_updateOrAdoptEvent, Can't deduce item type based on QI\n");
                    return;
                }
            } catch (e) {
                dump(e);
            }
            //</TO_REMOVE>
            //dump("--aStatus: " + aStatus + " aOperationType: " + aOperationType + " aId: " + aId + " aDetail: " + aDetail + "\n");
            // FIXME : bug lightning ? aStatus for getItem is allways 0.
            if (!Components.isSuccessCode(aStatus) || !this.localEvent) {
                if (event.status != "CANCELLED") {
                    self.logger.debug("Adopt item : " + event.title, false, false);
                    calendar.adoptItem(event, modifyListener);
                }
            } else {
                // Will work with LT 0.9 ... not !
                var localEvent;
                try {
                    localEvent = aDetail.QueryInterface(Ci.calIEvent)
                } catch (e) {
                    localEvent = this.localEvent;
                }
                if (calendar.name != localEvent.calendar.name) {
                    if (event.status != "CANCELLED") {
                        self.logger.debug("Adopt item : " + event.title, false, false);
                        calendar.adoptItem(event, modifyListener);
                    }
                } else {

                    if (event.status != "CANCELLED") {
                        self.logger.debug("Update item !", false, false);
                        UtilsOBM.retainStamps(event, function(event) {
                            var sendInvitations = localEvent.getProperty("X-MOZ-SEND-INVITATIONS");
                            if (sendInvitations == "TRUE") {
                                event.setProperty("X-MOZ-SEND-INVITATIONS", "TRUE");
                            }
                        });

                        calendar.modifyItem(event
                            /* aNewItem */
                            , localEvent
                            /* aOldItem */
                            , modifyListener
                            /* listener */
                        );
                    } else {
                        self.logger.debug("Remove item !", false, false);
                        calendar.deleteItem(localEvent, modifyListener
                            /* listener */
                        );
                    }
                }
            }

        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };

        calendar.getItem(event.id, listener);
    },
    _updateParticipationInOccurrence: function(participation, occurrence, occurrenceRO, lastSync) {
        let nowDateTime = UtilsOBM.getDateTime(lastSync).getInTimezone(cal.dtz.UTC);

        for (var attendee of participation.getAttendees({})) {
            // Do we really need to guard against the server sending a wrong participation status?
            this.logger.debug("_updateParticipation attendee.email:" + attendee.email, false, false);
            var newParticipationStatus = this._getAcceptableParticipationState(attendee.status);
            var localAttendee = occurrenceRO.getAttendeeById("MAILTO:" + attendee.email);
            if (localAttendee == null) {
                this.logger.debug("_updateParticipation: localAttendee " + attendee.email + " not found", false, false);
                continue;
            }
            // Can't modifiy localAttendee because it is immutable
            var newLocalAttendee = localAttendee.clone();
            this.logger.debug("_updateParticipation actual participationStatus:" + localAttendee.participationStatus, false, false);
            newLocalAttendee.participationStatus = newParticipationStatus;
            this.logger.debug("_updateParticipation new status:" + newLocalAttendee.participationStatus, false, false);
            if (localAttendee.participationStatus != newLocalAttendee.participationStatus) {
                newLocalAttendee.setProperty("RECEIVED-DTSTAMP", nowDateTime.icalString);
                // Can't modify directly localEvent because it is immutable too...
                occurrence.removeAttendee(localAttendee);
                occurrence.addAttendee(newLocalAttendee);
                this.logger.debug("_updateParticipation: modification done.", false, false);
            }
        }
        return occurrence;
    },
    _updateParticipation: function(calendar, participation, isTheFirstPass, lastSync, callback) {
        this.logger.debug("_updateParticipation:()", false, false);
        this.logger.debug("_updateParticipation: id = " + participation.id, false, false);
        this.logger.debug("_updateParticipation: extId = " + participation.extId, false, false);
        this.logger.debug("_updateParticipation: recurrenceId = " + participation.recurrenceId, false, false);
        var self = this;
        var updateParticipationCallback = function() {
            if (callback) { callback(); }
        };
        this._calGetItem(calendar, participation, function(localEvent) {
            if (!localEvent) {
                // this case should not appear
                self.logger.debug("_updateParticipation: localEvent not found " + participation.id, false, false);
                updateParticipationCallback();
            } else {
                // OBMFULL-4522
                // We need to clone here, otherwise the _setChangeOrigin method would fail
                let newLocalEvent = localEvent.clone();

                self._setChangeOrigin(newLocalEvent);
                if (!participation.recurrenceId) {
                    self.logger.debug("_updateParticipation: localEvent.id= " + localEvent.id, false, false);

                    self._updateParticipationInOccurrence(participation, newLocalEvent, localEvent, lastSync);
                    calendar.modifyItem(newLocalEvent, localEvent, getListener({ complete: updateParticipationCallback }));
                } else {
                    let recurrenceInfo = localEvent.recurrenceInfo;

                    if (recurrenceInfo) {
                        let dt = cal.createDateTime(participation.recurrenceId);
                        let exception = recurrenceInfo.getExceptionFor(dt, true);

                        if (exception) {
                            let newException = exception.clone();

                            self._updateParticipationInOccurrence(participation, newException, exception, lastSync);
                            newLocalEvent.recurrenceInfo.modifyException(newException, false);
                            calendar.modifyItem(newLocalEvent, localEvent, getListener({ complete: updateParticipationCallback }));
                        } else {
                            self.logger.debug("unable to find exception for occurrence recurrenceId " + participation.recurrenceId, false, false);
                            updateParticipationCallback();
                        }
                    } else {
                        updateParticipationCallback();
                    }
                }
            }
        });
    },

    _getAcceptableParticipationState: function(participationState) {
        var acceptableParticipationStates = ["ACCEPTED", "DECLINED", "NEEDS-ACTION"];
        var defaultParticipationState = "NEEDS-ACTION";
        var newParticipationState = defaultParticipationState;

        for (var state of acceptableParticipationStates) {
            if (state == participationState) {
                newParticipationState = state;
                break;
            }
        }
        return newParticipationState;
    },
    _removeEvent: function(calendar, eventId, isTheFirstPass, callback) {
        this.logger.debug("_removeEvent:()", false, false);
        this.logger.debug("  eventId: " + eventId, false, false);

        var listener = {};
        var self = this;
        var onComplete = function() {
            if (callback) { callback(); }
        };
        listener.calendar = calendar;
        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            try {
                if (aDetail.toString() == "Can't deduce item type based on QI") {
                    return;
                }
            } catch (e) {}

            if (aStatus == Cr.NS_OK) {
                // Found, so must be deleted
                self.logger.debug("Remove item !", false, false);
                if (this.localItem) {
                    try {
                        this.calendar.deleteItem(this.localItem, getListener({ complete: onComplete })
                            /* listener */
                        )
                    } catch (e) {};
                } else {
                    onComplete();
                }
            }


        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localItem = aItems[0];
            }
        };
        calendar.getItem(eventId, listener);
    },
    _aMergeActionCompleted: function(isTheFirstPass) {
        if (this.mergeActionCount > 0) {
            this.callerListener.onProgress(this.mergeActionIdx, this.mergeActionCount, "", "all");
            this.logger.debug("a action completed:" + this.mergeActionIdx, false, false);
        }
        this.mergeActionIdx++;
    },
    _isEventPrivateFor: function(event, calName) {
        if (UtilsOBM.capabilitySupported("SERVER_SIDE_ANONYMIZATION")) {
            return event.anonymized;
        }

        var attendees = event.getAttendees({});
        let calendarIsReadOnly = UtilsOBM.getPrefSafe("extensions.obm.changes.calICalendar." + calName + ".readOnly", true);

        if (!calendarIsReadOnly) {
            return false;
        }
        //dump("event.attendees: " + attendees + "\n");
        if (event.privacy !== 1 && event.privacy !== "PRIVATE") {
            return false;
        }
        if (event.owner == this._login.split("@")[0]) {
            return false;
        }
        if (!attendees) {
            return true;
        }
        for (attendee of attendees) {
            //dump("attendee.email: " + attendee.email + "\n");
            var email = attendee.email ? attendee.email : attendee.id;
            if (email == this._login.split("@")[0]) {
                return false;
                break;
            }
        }
        return true;
    },
    _mergeRemoteDiff: function(isTheFirstPass) {
        this.logger.debug("_mergeRemoteDiff()", false, false);
        var calMgr = cal.getCalendarManager();
        var self = this;
        var mergedCalendarsCount = 0;
        var endOfOneCalendarMerge = function() {
            mergedCalendarsCount++;
            if (mergedCalendarsCount == self.calendars.length) {
                return endOfMerge();
            }
        };
        var endOfMerge = function() {
            self.logger.debug("Merge completed", false, false);
            self.callerListener.onOperationDone(true, null);
            if (isTheFirstPass) {
                self._sendLocalDiff();
            } else {
                //end sync
                self.callerListener.onStop(Cr.NS_OK, "done");
                var resourceCalendars = UtilsOBM.getSubscribedOBMResource();
                for (var resourceCalendar of resourceCalendars) {
                    resourceCalendar.refresh();
                }

                self._handleLostCalendarRights();

                Components
                    .classes["@mozilla.org/observer-service;1"]
                    .getService(Components.interfaces.nsIObserverService)
                    .notifyObservers(null, "obmCalendarSyncCompleted", 0);
            }
        };

        this.mergeActionCount = 0;
        this.mergeActionIdx = 0;
        for (var i = 0; i < this.calendars.length; i++) {
            var calendar = this.calendars[i];
            var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
            this.mergeActionCount += this._remoteDiff[calName].updated.length;
            this.mergeActionCount += this._remoteDiff[calName].removed.length;
            this.mergeActionCount += this._remoteDiff[calName].participationChanged.length;
        }

        this.callerListener.onNewOperation("mergeRemoteDiff", this.mergeActionCount);
        this.logger.debug("Action count: " + this.mergeActionCount, false, false);
        if (this.mergeActionCount == 0) {
            this.logger.debug("No remote event have been updated/added/removed", false, false);
            this._aMergeActionCompleted(isTheFirstPass);
            return endOfMerge();
        }
        for (var i = 0; i < this.calendars.length; i++) {
            this._applyChanges(isTheFirstPass, this.calendars[i], endOfOneCalendarMerge);
        }
    },

    _handleLostCalendarRights: function() {
        if (UtilsOBM.getPrefSafe("extensions.obm.rights.showChangedRightsMessage", true)) {
            let changes = 0;
            let details = "";

            if (this._calendarsToRemove && this._calendarsToRemove.length > 0) {
                for (let calendar of this._calendarsToRemove) {
                    details += "\n" + UtilsOBM.getLocalizedMessageWithParams("obm.sync.rights.calendar.removed", [calendar], 1);
                    changes++;
                }
            }
            if (this._calendarsNowReadOnly && this._calendarsNowReadOnly.length > 0) {
                for (let calendar of this._calendarsNowReadOnly) {
                    details += "\n" + UtilsOBM.getLocalizedMessageWithParams("obm.sync.rights.calendar.now.readonly", [calendar], 1);
                    changes++;
                }
            }

            if (changes > 0) {
                let promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                    .getService(Components.interfaces.nsIPromptService);

                let title = UtilsOBM.getLocalizedMessage("obm.sync.rights.changed.title");
                let message = UtilsOBM.getLocalizedMessage("obm.sync.rights.changed.message") + "\n" + details;

                promptService.alert(null, title, message);
            }
        }

        this._calendarsToRemove = [];
        this._calendarsNowReadOnly = [];
    },

    _applyChanges: function(isTheFirstPass, calendar, callback) {
        var self = this;
        var logger = this.logger;
        var calMgr = cal.getCalendarManager();
        var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
        var lastSync = parseInt(this._remoteDiff[calName].lastSync, 10);
        var jobs = [];
        var applyUpdate = function(event, then) {
            logger.debug("applyUpdate", false, false);
            var calEvent;
            if (event instanceof Components.interfaces.calIEvent) {
                self.logger.debug("event " + event.uid + " is a Lightning event", false, false);
                calEvent = event;
            } else if (self._isEventPrivateFor(event, calName)) {
                calEvent = self._createLocalPrivateEvent(event, calendar, null, lastSync);
            } else {
                self.logger.debug("event " + event.uid + " is an OBM event", false, false);
                calEvent = self._createLocalEvent(event, calendar, null, lastSync);
            }

            if (calEvent.status == "CANCELLED") {
                self.logger.debug("event is private or refused (remove if exist)", false, false);
                logger.debug("calling self._removeEvent", false, false);
                self._removeEvent(calendar, calEvent.id, isTheFirstPass, then);
            } else {
                if (!calEvent.isMutable) {
                    calEvent = calEvent.clone();
                }
                self._setChangeOrigin(calEvent);
                logger.debug("calling self._updateOrAdoptEvent", false, false);
                self._updateOrAdoptEvent(calendar, calEvent, isTheFirstPass, then);
            }
        };

        var applyRemoval = function(event, then) {
            logger.debug("applyRemoval", false, false);
            self._removeEvent(calendar, event, isTheFirstPass, then);
        };

        var applyParticipationChange = function(participation, then) {
            logger.debug("applyParticipationChange", false, false);
            self._updateParticipation(calendar, participation, isTheFirstPass, lastSync, then);
        };

        calendar.startBatch();

        for (var event of this._remoteDiff[calName].updated) {
            jobs.push({ action: "update", target: event });
        }
        for (var event of this._remoteDiff[calName].removed) {
            jobs.push({ action: "remove", target: event });
        }
        for (var participation of this._remoteDiff[calName].participationChanged) {
            jobs.push({ action: "participationChange", target: participation });
        }

        var nextJob = function() {
            self._aMergeActionCompleted(isTheFirstPass);
            logger.debug("nextJob: " + jobs.length, false, false);
            if (!jobs.length) {
                // allow the observers to be fired before ending batch mode on the calendar
                nextTick(finish);
                return;
            }
            var job = jobs.shift();
            if (job.action == "update") {
                nextTick(function() { applyUpdate(job.target, nextJob) });
            } else if (job.action == "remove") {
                nextTick(function() { applyRemoval(job.target, nextJob) });
            } else if (job.action == "participationChange") {
                nextTick(function() { applyParticipationChange(job.target, nextJob) });
            } else {
                Components.utils.reportError("Unexpected job: " + JSON.stringify(job));
                nextJob();
            }
        };

        var finish = function() {
            calendar.endBatch();
            self._setLastSync(calName, self._remoteDiff[calName].lastSync);
            // we remove all local changes related to the current calendar
            // and we re-save local changes. They may be changes after the conflicts resolving
            // In case where an exception appear during the send of local change
            // we want to keep the list of changes to do after the merge of remote change
            // to replay it after.
            if (isTheFirstPass) {
                UtilsOBM.removeAllCalendarSyncItems(calendar);
                for (var j in self._localDiff[calendar.id].event) {
                    var ev = self._localDiff[calendar.id].event[j];
                    if (!"extId" in ev) {
                        UtilsOBM.setCalendarSyncItem(calendar, j, ev);
                    }
                }
            }
            callback();
        };


        nextJob();
    },
    /**
     * @param nsISupports aData  properties accessible via wrappedJSObject: calendar, calName, id
     */
    _onSentLocalDiff: function(aResCode, aMethodName, aResult, aData) {
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, aMethodName, aData)) {
            if ((aResult instanceof Ci.nsISupportsString) && (aResult == "Invalid access token")) {
                // do not continue -> must wait reLogin
                return;
                // We'll get here if a 'write' operation failed because we lost writing rights on the target calendar.
            } else if ((aResult instanceof Ci.nsISupportsString) && (aResult == "NotAllowed")) {
                let calendar = aData.wrappedJSObject.calendar;

                this._localDiffOperationCount--;
                UtilsOBM.applyRemoteCalendarRights(calendar, UtilsOBM.getCalInfoForReadOnlyLocalCalendar(calendar));

                // So that this.calendars contains the new, updated calendar
                this.calendars = UtilsOBM.getSubscribedOBMCalendar();
            } else if (this._isCriticalMethod(aMethodName)) {
                // do not continue: show a useful error message for the user
                for (var i = 0; i < this.calendars.length; i++) {
                    var calendar = this.calendars[i];
                    calendar.endBatch();
                }
                return;
            } else {
                // continue -> sync algo permit local operation to fail
                this._localDiffOperationCount--;
            }
            this._removeSyncItemFromHistory(aData);
            this._onSentLocalDiffEndOfOperation();
        } else {
            this._localDiffOperationCount--;
            switch (aMethodName) {
                case "createEvent":
                case "storeEvent":
                    var newId = this._getJSStringFromXPString(aResult);
                    var ctx = aData.wrappedJSObject;
                    var newLocalEvent = ctx.localEvent.clone();
                    var self = this;
                    this._setChangeOrigin(newLocalEvent);
                    UtilsOBM.setOBMEventId(true, newLocalEvent, newId);
                    ctx.calendar.modifyItem(newLocalEvent, ctx.localEvent, getListener({
                        complete: function() {
                            self._removeSyncItemFromHistory(aData);
                            self._onSentLocalDiffEndOfOperation();
                        }
                    }));

                    break;
                default:
                    this._removeSyncItemFromHistory(aData);
                    this._onSentLocalDiffEndOfOperation();
                    break;
            };
        }

    },
    _removeSyncItemFromHistory: function(aData) {
        if (aData && aData.wrappedJSObject) {
            // we should remove the corresponding change
            // we don't remove all changes in a single step after
            // the sending of all changes, to have the opportunity
            // to replay remaining changes after an exception appeared
            // during the sending (connection lost...)
            /*for ( var i in aData.wrappedJSObject ) {
                this.logger.debug("...."+i,false,false);
            }*/
            UtilsOBM.removeCalendarSyncItem(aData.wrappedJSObject.calendar, aData.wrappedJSObject.localEvent && aData.wrappedJSObject.localEvent.id ? aData.wrappedJSObject.localEvent.id : aData.wrappedJSObject.id);
            this.logger.debug("_onSentLocalDiff : removeCalendarSyncItem ok", false, false);
        }
    },
    _onSentLocalDiffEndOfOperation: function() {
        this.logger.debug("_onSentLocalDiffEndOfOperation: " + this._localDiffOperationCount, false, false);
        if (this._localDiffOperationCount <= 0) {
            this.callerListener.onOperationDone(true, null);
            for (var i = 0; i < this.calendars.length; i++) {
                var calendar = this.calendars[i];
                calendar.endBatch();
            }
            var self = this;
            this.CalProxy.removeListener();
            var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            this.logger.debug("timer init", false, false);
            this.callerListener.onNewOperation("wait-one-seconde", null);
            timer.initWithCallback({
                    notify: function(timer) {
                        self.logger.debug("ok, time is out", false, false);
                        return (self._fireSecondPass());
                    }
                },
                1100, timer.TYPE_ONE_SHOT);
            this.timer = timer;
        } else {
            this._sendNextLocalOperation();
        }
    },
    _fireSecondPass: function() {
        this.logger.debug("_fireSecondPass()", false, false);

        for (var i = 0; i < this.calendars.length; i++) {
            this.calendars[i].endBatch();
        }

        this._getSyncOperationCount = this.calendars.length;
        var self = this;
        this.callerListener.onOperationDone(true, null);
        this.callerListener.onNewOperation("getSyncWithSortedChanges", null);
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onGetSync(aResCode, aMethodName, aResult, aData)
        });

        this._localOperations = [];
        var calMgr = cal.getCalendarManager();
        for (var i = 0; i < this.calendars.length; i++) {
            var calendar = this.calendars[i];
            var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
            this._localOperations[i] =
                (function(aCal, aCalName, aIdx, aTotal) {
                    return function() {
                        self.callerListener.onProgress(aIdx, aTotal, "", aCalName);
                        var lastSync = self._getLastSync(aCal);
                        var data = new Context();
                        data.methodName = "getSyncWithSortedChanges";
                        data.isTheFirstPass = false;
                        data.calendar = aCal;
                        data.lastSync = lastSync;
                        data.calName = aCalName;
                        self.CalProxy.getSyncWithSortedChanges(aCalName, lastSync, data);
                    }
                })(calendar, calName, i, this.calendars.length);
        }

        this._localOperationsIdx = 0;
        return (this._sendNextLocalOperation());
    },
    _tellServerParticipationChanged: function(calendar, aCalName, aId, aInfos) {
        this.logger.debug("_tellServerParticipationChanged()", false, false);
        this.logger.debug("   aId: " + aId, false, false);
        this.logger.debug("   aInfos: " + JSON.stringify(aInfos), false, false);
        // Create listener that will instruct what to do when we will have retreived the event
        var listener = {};
        var self = this;
        listener.calendar = calendar;
        listener.calendarName = aCalName;
        // When calendar.getItem returns...
        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            var getInUtc = function(str) {
                if (str[(str.length - 1)] == "Z") {
                    return str;
                }
                var dt = cal.createDateTime(str);
                dt.timezone = cal.dtz.defaultTimezone;
                return dt.getInTimezone(cal.dtz.UTC).icalString;
            };
            // localEvent is the result of calendar.getItem. It's the event the user modified
            var localEvent = this.localEvent;
            if (!localEvent) {
                return self._onSentLocalDiff(Cr.NS_OK, null, null, null);
            }
            var calMgr = cal.getCalendarManager();
            //var eventToSend = self._mozEventToOBMEvent(this.calendar, localEvent, this.willBeCreated);
            //var notification = UtilsOBM.getNotificationStatus(this.calendar, localEvent);
            var notification = aInfos.notification || false;
            //var notification = true;

            var ctx = new Context();
            ctx.calendar = this.calendar;
            ctx.calName = this.calendarName;
            ctx.id = aId;
            ctx.localEvent = localEvent;
            ctx.notification = notification;
            ctx.recurrenceId = aInfos.recurrenceId ? getInUtc(aInfos.recurrenceId) : null;
            ctx.recursive = ctx.recurrenceId ? false : true;
            ctx.generation = aInfos.sequence ? aInfos.sequence : 0;
            ctx.methodName = "changeParticipationState";
            var localAttendeeEmail = calMgr.getCalendarPref_(this.calendar, "X-OBM-EMAIL");
            var localAttendeeId = "MAILTO:" + localAttendeeEmail;
            var event = null;
            self.logger.debug("_tellServerParticipationChanged: localAttendeeId=" + localAttendeeId, false, false);
            if (ctx.recurrenceId) {
                event = localEvent.recurrenceInfo ? localEvent.recurrenceInfo.getExceptionFor(cal.createDateTime(ctx.recurrenceId)) : null;
                if (!event) {
                    self.logger.debug("Warning: using event instead of occurrence: recurrenceId " + ctx.recurrenceId + " not found", false, false);
                }
            }
            if (!event) event = localEvent;
            var localAttendee = null;
            self.logger.debug(event.icalString, false, false);
            for (var att of event.getAttendees({})) {
                self.logger.debug("Parsing attendee: " + att.id, false, false);
                if (att.id.toLowerCase() == localAttendeeId.toLowerCase()) {
                    localAttendee = att;
                }
            }

            if (localAttendee != null) {
                ctx.participationStatus = localAttendee.participationStatus;
                self.CalProxy.changeParticipationState(this.calendarName, aId, localAttendee.participationStatus, notification, ctx.generation, ctx.recurrenceId, ctx);
            } else {
                self.logger.error("_tellServerParticipationChanged: Could not find attendee '" + localAttendeeId + "' in order to tell server that his participation status was updated!", false, false);
                return self._onSentLocalDiff(Cr.NS_OK, null, null, null);
            }
        }

        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };

        calendar.getItem(aId, listener);
    },
    _sendLocalDiff: function() {
        var calMgr = cal.getCalendarManager();
        for (var i = 0; i < this.calendars.length; i++) {
            this.calendars[i].startBatch();
        }

        this._localDiffOperationCount = 0;

        for (var i in this._localDiff) {
            if (this._localDiff[i].event) {
                for (j in this._localDiff[i].event) {
                    if (this._localDiff[i].event[j].action) {
                        this._localDiffOperationCount++;
                    }
                }
            }
        }
        var _localDiffOperationCount;
        _localDiffOperationCount = this._localDiffOperationCount;
        // for (var i = 0; i < this.calendars.length; i++) {
        //     var calendar = this.calendars[i];
        //     var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
        //     this._localDiffOperationCount += this._localDiff[calName].event.removed.length;
        //     this._localDiffOperationCount += this._localDiff[calName].event.updated.length;
        //     this._localDiffOperationCount += this._localDiff[calName].event.added.length;
        //     this._localDiffOperationCount += this._localDiff[calName].event.participationChanged.length;
        // }

        this.logger.debug("_sendLocalDiff(" + this._localDiffOperationCount + ")", false, false);
        this.callerListener.onNewOperation("sendLocalDiff", this._localDiffOperationCount);

        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
        });

        if (this._localDiffOperationCount == 0) {
            return (this._onSentLocalDiff(Cr.NS_OK, null, null, null));
        }
        this.logger.debug("this._localDiff contents: " + JSON.stringify(this._localDiff), false, false);
        var idx = 0;
        this._localOperations = [];
        for (var i = 0; i < this.calendars.length; i++) {
            var calendar = this.calendars[i];
            var calName = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");

            for (var j in this._localDiff[calendar.id].event) {
                var evChange = this._localDiff[calendar.id].event[j];

                if (evChange.action == "removed") {
                    this._localOperations[idx++] =
                        (function(aCal, aCalName, aId, aOpId, aInfos) {
                            return function() {
                                self.callerListener.onProgress(aOpId - 1, _localDiffOperationCount, "removed", aCalName);
                                //var id = self._localDiff[aCalName].event.removed[aIdx];
                                //var infos = self._localDiff[aCalName].event[aId];
                                //var notification = UtilsOBM.getNotificationStatus(calendar, id);
                                var notification = aInfos.notification ? aInfos.notification : false;
                                var data = new Context();
                                var generation = aInfos.generation ? aInfos.generation : 0;
                                data.calendar = aCal;
                                data.calName = aCalName;
                                data.id = aId;
                                data.notification = notification;
                                data.methodName = "removeEventByExtId";
                                self.CalProxy.removeEventByExtId(aCalName, aId, notification, generation, data);
                            }
                        })
                        (calendar, calName, evChange.extId, idx, evChange);

                } else if (evChange.action == "added" ||
                    evChange.action == "updated") {
                    this._localOperations[idx++] =
                        (function(aCal, aCalName, aId, aOpId, aInfos) {
                            return function() {
                                self.logger.debug("update, count = " + _localDiffOperationCount + ", opID= " + aOpId, false, false);
                                self.callerListener.onProgress(aOpId - 1, _localDiffOperationCount, aInfos.action, aCalName);
                                self._createEventObjFromId(aCal, aCalName, aId, aInfos);
                            }
                        })(calendar, calName, evChange.extId, idx, evChange);

                } else if (evChange.action == "participationChanged") {
                    this._localOperations[idx++] = (function(aCal, aCalName, aId, aOpId, aInfos) {
                        return function() {
                            self.logger.debug("id : " + aId + ", infos : " + JSON.stringify(aInfos), false, false);
                            self.callerListener.onProgress(aOpId - 1, _localDiffOperationCount, "participationChanged", aCalName);
                            self._tellServerParticipationChanged(aCal, aCalName, aId, aInfos);
                        }
                    })(calendar, calName, evChange.extId, idx, evChange);
                }
            }

        }
        this._localOperationsIdx = 0;
        this._sendNextLocalOperation();
    },
    _OBMCalendarsArePresent: function() {
        return UtilsOBM.getCalendarManager().getOBMCalendars({}).length > 0;
    },

    // /////////////////////////////

    /*
     * obmIRCalendarSync
     */

    _doLogin: function() {
        this.logger.debug("_doLogin()", false, false);
        this.callerListener.onNewOperation("logUserIn", null);
        var login = {};
        var password = {};
        this._getOBMCredentials(login, password);
        this._login = login.value;

        var self = this;

        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onLogged(aResCode, aMethodName, aResult, aData)
        });
        this.logger.debug("_doLogin() SEL4" + UtilsOBM.getOrigin(), false, false);
        this.CalProxy.doLogin(login.value, password.value, UtilsOBM.getOrigin(), null);
    },

    _createAndRegisterCalendar: function(uid, name, email, readOnly, color) {
        var calMgr = cal.getCalendarManager();
        UtilsOBM.setPrefSafe("extensions.obm.changes.calICalendar." + uid + ".readOnly", readOnly);
        UtilsOBM.setPrefSafe("extensions.obm.changes.calICalendar." + uid + ".firstRefresh", true);
        var uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
        uri.spec = UtilsOBM.makeCalendarUri(UtilsOBM.getNextLocalCalendarID(), uid);
        this.logger.debug("uri.spec: " + uri.spec, false, false);
        var newCalendar = calMgr.createCalendar("obm", uri); //obm
        // Components.utils.import("chrome://obm-extension/content/components/calOBMCalendar.js")
        // var newCalendar = new calOBMCalendar();
        // newCalendar.uri = uri;
        calMgr.registerCalendar(newCalendar);
        this.logger.debug("newCalendar.id:" + newCalendar.id, false, false);
        newCalendar.name = name;
        calMgr.setCalendarPref_(newCalendar, "X-OBM-NAME", uid.toString());
        calMgr.setCalendarPref_(newCalendar, 'X-OBM-EMAIL', email.toString());
        newCalendar.setProperty("color", color.toString());
        this._setLastSync(uid, "0");
        return newCalendar;
    },
    _getRemoteCalendarsAsHash: function(remoteCalendarInfos) {
        // Generates a hash of calendars with the OBM email as key
        var calendarsHash = {};
        for (var i = 0; i < remoteCalendarInfos.length; i++) {
            var remoteCalInfo = remoteCalendarInfos.queryElementAt(i, Ci.obmIRCalendarInfo);
            calendarsHash[remoteCalInfo.mail] = remoteCalInfo;
        }
        return calendarsHash;
    },
    _getRemoteResourcesAsHash: function(remoteCalendarInfos) {
        // Generates a hash of calendars with the OBM email as key
        var calendarsHash = {};
        for (var i = 0; i < remoteCalendarInfos.length; i++) {
            var remoteCalInfo = remoteCalendarInfos.queryElementAt(i, Ci.obmIRResourceInfo);
            calendarsHash[remoteCalInfo.mail] = remoteCalInfo;
        }
        return calendarsHash;
    },
    _onListCalendar: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onListCalendar()", false, false);
        this.callerListener.onOperationDone(true, null);
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "listCalendars")) {
            return;
        };
        var self = this;
        this.CalProxy.removeListener();
        let obmCalMgr = UtilsOBM.getCalendarManager();

        if (this._onlyListCalendars) {
            this.callerListener.onOperationDone(true, aResult);
            this.callerListener.onStop(Cr.NS_OK, "done");
            return;
        }

        var calendars = aResult.QueryInterface(Ci.nsIArray);
        var count = calendars.length;
        var login = {};
        this._getOBMCredentials(login, {});
        login = login.value.split('@')[0];
        for (var i = 0; i < count; i++) {
            var calInfo = calendars.queryElementAt(i, Ci.obmIRCalendarInfo);
            if (calInfo.uid == login) {
                // We consider that the user can always read/write on
                let name = UtilsOBM.getPrefSafe("extensions.obm.autoconf.calname", null) ||
                    calInfo.firstname + " " + calInfo.lastname;
                obmCalMgr.subscribeObmCalendar(calInfo.uid, name, calInfo.mail, false);
                this.callerListener.onOperationDone(true, null);
                break;
            }
        }
        this._listCategories();
    },
    _onListResources: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onListResources()", false, false);
        this.callerListener.onOperationDone(true, null);
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "listResources")) {
            return;
        };
        var self = this;
        this.CalProxy.removeListener();

        if (this._onlyListResources) {
            this.callerListener.onOperationDone(true, aResult);
            this.callerListener.onStop(Cr.NS_OK, "done");
            return;
        }
    },

    _listCategories: function() {
        this.callerListener.onNewOperation("listCategories", {});
        var self = this;
        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onListCategories(aResCode, aMethodName, aResult)
        });
        this.CalProxy.listCategories({});
    },

    _onListCategories: function(aResCode, aMethodName, aResult) {
        this.logger.debug("_onListCategories()", false, false);
        this.callerListener.onOperationDone(true, null);
        this.logger.debug("_onListCategories", false, false);
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "listCategories")) {
            return;
        };

        let currentCategories = UtilsOBM.getPrefSafe("calendar.categories.names");
        if (currentCategories && currentCategories != "") {
            currentCategories = currentCategories.split(",");
        } else {
            currentCategories = [];
        }

        var newCategories = [];
        for (var category of currentCategories) {
            if (!category.match(RegExp(/\[OBM\]/))) {
                newCategories.push(category);
            }
        }

        var serverCategories = aResult.getCategories({});
        for (var category of serverCategories) {
            var label = "[OBM] " + category.label;
            newCategories.push(label);
        }
        UtilsOBM.setPrefSafe("calendar.categories.names", newCategories.join(","));

        this._getLocalDiff();
    },

    mergeOneCalendar: function(aCalendar, aListener) {
        this.logger.debug("mergeOneCalendar()", false, false);
        this.logger.debug("  aCalendar: " + aCalendar, false, false);
        this._justOneCalendar = true;
        this._selectedCalendarUid = UtilsOBM.getCalendarOBMName(aCalendar);
        this.merge(aListener);
    },

    mergeAllCalendars: function(aListener) {
        this.logger.debug("mergeAllCalendars()", false, false);
        this._justOneCalendar = false;
        this.merge(aListener);
    },
    setCalendarProxy: function() {
        try {
            this.CalProxy = new ObmRCalendarProxy();
            var PortUrl = UtilsOBM.getPrefSafe("extensions.obm.server");
            this.CalProxy.PortUrl = PortUrl; // +
            // "/AddressBook";
        } catch (e) {
            var error = "NO_SERVER_INFO";
            this.logger.error(error, true, false);
            this.callerListener.onError(error);
            throw (error);
        }
    },
    getSubscribleCalendars: function(aListener, aLimit, aOffset, aPattern) {
        this.logger.debug("getSubscribleCalendars()", false, false);
        this.callerListener = aListener;
        this._paginer = { 'limit': aLimit, 'offset': aOffset, 'pattern': aPattern };
        this.callerListener.onStart();
        this._onlyListCalendars = true;

        this.setCalendarProxy();

        if (UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) == false) {
            this.callerListener.onStop(Cr.NS_OK, "done - Events not sync-able");
            return;
        }

        this._doLogin();
    },

    getSubscribleResources: function(aListener, aLimit, aOffset, aPattern) {
        this.logger.debug("getSubscribleResources()", false, false);
        this.callerListener = aListener;
        this.logger.debug("getSubscribleResources(): " + aLimit + ', ' + aOffset + ', ' + aPattern, false, false);
        this._paginer = { 'limit': aLimit, 'offset': aOffset, 'pattern': aPattern };
        this.callerListener.onStart();
        this._onlyListResources = true;

        this.setCalendarProxy();

        if (UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) == false) {
            this.callerListener.onStop(Cr.NS_OK, "done - Events not sync-able");
            return;
        }

        this._doLogin();
    },

    merge: function(aListener) {
        this.logger.debug("mergeCalendars", false, false);
        this.callerListener = aListener;
        this.callerListener.onStart();

        this.setCalendarProxy();

        if (UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) == false) {
            this.callerListener.onStop(Cr.NS_OK, "done - Events not sync-able");
            return;
        }

        Components
            .classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService)
            .notifyObservers(null, "obmCalendarSyncStarted", 0);

        this._doLogin();
    },

    purge: function(calName, aListener) {
        this.logger.debug("purge(" + calName + ")", false, false);

        this._calNameToPurge = calName;
        this.callerListener = aListener;
        this.callerListener.onStart();

        this.setCalendarProxy();

        // Logging in
        this.callerListener.onNewOperation("logUserIn", null);
        var login = {};
        var password = {};
        this._getOBMCredentials(login, password);
        this._login = login.value;

        var self = this;

        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onLoggedForPurge(aResCode, aMethodName, aResult, aData)
        });
        this.CalProxy.doLogin(login.value, password.value, UtilsOBM.getOrigin(), null);
    },
    _onLoggedForPurge: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onLoggedForPurge(" + this._calNameToPurge + ", " + aResCode + ", " + aResult + ")", false, false);
        var self = this;
        this.CalProxy.removeListener();
        if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "doLogin")) {
            this.logger.debug("_onLoggedForPurge(): Response not valid", false, false);
            return;
        };
        this.callerListener.onOperationDone(true, null);

        this._addListener(function(aResCode, aMethodName, aResult, aData) {
            self._onPurge(aResCode, aMethodName, aResult, aData)
        });
        this.CalProxy.purge(this._calNameToPurge, aData);
    },
    _onPurge: function(aResCode, aMethodName, aResult, aData) {
        this.logger.debug("_onPurge(aResCode=" + aResCode + ", aResult=" + aResult + ", aData=" + aData + ")", false, false);
        if (aResCode != Cr.NS_OK) {
            this.logger.notice("Purge: error \"" + aResult + "\"", false, false);
            var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                .getService(Components.interfaces.nsIPromptService);
            var title = UtilsOBM.getLocalizedMessage("obm.syncerror.srvError");
            var message = UtilsOBM.getLocalizedMessage("obm.syncerror.srvError");
            promptService.alert(null, title, message);
        } else {
            this.logger.notice("Purge: obm-sync performed purged. Resetting connector", false, false);
            // remember purged calendars to get'em back

            var
                calendarManager = cal.getCalendarManager(),
                calendars = calendarManager.getCalendars({}),
                calendarInfos = [],
                self = this;
            for (var calendar of calendars) {
                if (calendar.type == "obm") {
                    var infos = {
                        uid: UtilsOBM.getCalendarOBMName(calendar),
                        email: calendarManager.getCalendarPref_(calendar, "X-OBM-EMAIL"),
                        name: calendar.name,
                        color: calendar.getProperty("color"),
                        readOnly: calendar.readOnly
                    };
                    calendarInfos.push(infos);
                }
            }
            this.logger.debug("Calendar infos : " + JSON.stringify(calendarInfos), false, false);

            UtilsOBM.reset(function() {

                //recreate previous calendars
                for (var info of calendarInfos) {
                    self._createAndRegisterCalendar(info.uid, info.name, info.email, info.readOnly, info.color);
                }

                var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);

                var check = { value: false };
                var flags = promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_0 +
                    promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_1;

                var title = UtilsOBM.getLocalizedMessage("obmpurgetitle");
                var message = UtilsOBM.getLocalizedMessage("obmpurgemessage");
                var rightnow = UtilsOBM.getLocalizedMessage("obmreinitrightnow");
                var later = UtilsOBM.getLocalizedMessage("obmreinitlater");
                var button = promptService.confirmEx(null, title, message, flags, rightnow, later, "", null, check);

                if (button == 0) {
                    // restart now
                    var nsIAppStartup = Components.interfaces.nsIAppStartup;
                    Components.classes["@mozilla.org/toolkit/app-startup;1"].getService(nsIAppStartup).quit(nsIAppStartup.eForceQuit | nsIAppStartup.eRestart);
                }
            });
        }
    }
};

// module

var components = [ObmRCalendarSync];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}