/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmCalendarsManager"];

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
const Cc = Components.classes;
const Ci = Components.interfaces;

/**
 * Main object to access to OBM calendars
 */
function ObmCalendarsManager() {
    Components.utils.import('chrome://obm-extension/content/components/logger/logger.js')
    Components.utils.import("resource://gre/modules/Services.jsm");
    Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
    Components.utils.import("resource://calendar/modules/calUtils.jsm");
};

ObmCalendarsManager.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmICalendarsManager, Ci.calIFreeBusyProvider, Ci.nsIObserver, Ci.calICalendarManagerObserver]),
    classID: Components.ID("{a55551b2-2bea-406c-9b6b-f9be098ed178}"),
    contractID: "@obm.org/calendarsmanager;1",
    classDescription: "OBM Calendars Manager",

    // {{{ START_COMPAT_TB_3
    _xpcom_categories: [{
        category: "app-startup",
        service: true
    }],
    // END_COMPAT_TB_3 }}}

    initLogger: function() {
        this.logger = new ObmLogger();

        this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
        this.logger.setHeader("ObmCalendarsManager.js: ");
        this.logger.debug("new ObmCalendarsManager()", false, false);
    },

    _getNextLocalCalendarID: function() {
        let calendars = cal.getCalendarManager().getCalendars({});

        // FIXME n'importe quoi !!!
        let minID = 20;

        for (let calendar of calendars) {
            if (calendar.uri) {
                let calendarID = calendar.uri.spec.match(/id=(\d+)/);
                if (calendarID != null) {
                    if (Number(calendarID[1]) >= minID) {
                        minID = Number(calendarID[1]) + 1;
                    }
                }
            }
        }

        return minID;
    },

    getObmCalendar: function(aCalInfoUid) {
        let calMgr = cal.getCalendarManager();
        let calendars = calMgr.getCalendars({});
        for (let calendar of calendars) {
            if (calendar.type == "obm" && calMgr.getCalendarPref_(calendar, "X-OBM-NAME") == aCalInfoUid) {
                return calendar;
            }
        }
        return null;
    },

    subscribeObmCalendar: function(aUid, aName, aEmail, aReadOnly) {

        let name = aName.indexOf("[OBM]") > -1 ? aName : "[OBM] " + aName;
        let color = UtilsOBM.getRandomColor();
        //dump("_subscribeTo: " + aUid + ", " + name + ", " +  aEmail + ", " + color + ", " + aReadOnly + "\n");

        UtilsOBM.setPrefSafe(
            "extensions.obm.changes.calICalendar." +
            aUid +
            ".readOnly", aReadOnly
        );
        UtilsOBM.setPrefSafe(
            "extensions.obm.changes.calICalendar." +
            aUid +
            ".firstRefresh", true
        );

        //let uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
        //uri.spec = UtilsOBM.makeCalendarUri(this._getNextLocalCalendarID(), aUid);
        let uri = Services.io.newURI(UtilsOBM.makeCalendarUri(this._getNextLocalCalendarID(), aUid));
        //      uri.spec = "moz-profile-calendar://?id=" + this._getNextLocalCalendarID() + ";" + aUid;

        let calMgr = cal.getCalendarManager();
        var newCalendar = calMgr.createCalendar("obm", uri); //obm
        // Components.utils.import("chrome://obm-extension/content/components/calOBMCalendar.js")
        // var newCalendar = new calOBMCalendar();
        // newCalendar.uri = uri;
        calMgr.registerCalendar(newCalendar);

        newCalendar.name = name;
        newCalendar.setProperty("color", color.toString());
        newCalendar.setProperty("calendar-main-in-composite", true)

        calMgr.setCalendarPref_(newCalendar, "X-OBM-NAME", aUid.toString());
        calMgr.setCalendarPref_(newCalendar, 'X-OBM-EMAIL', aEmail.toString());

        UtilsOBM.setPrefSafe("extensions.obm.changes." +
            Ci.calICalendar + "." +
            aUid + "." +
            "lastSync",
            "1970-01-01T00:00:00Z");
        return newCalendar;
    },

    updateObmCalendar: function(aCalendar, aUid, aName, aEmail, aReadOnly) {

        if (aName) {
            let name = "[OBM] " + aName;
            aCalendar.name = name;
        }

        //dump("_updateCalendar: " + aUid + ", " + name + ", " +  aEmail + ", " + aReadOnly + "\n");

        UtilsOBM.setPrefSafe(
            "extensions.obm.changes.calICalendar." +
            aUid +
            ".readOnly", aReadOnly
        );

        let calMgr = cal.getCalendarManager();
        calMgr.setCalendarPref_(aCalendar, "X-OBM-NAME", aUid.toString());
        calMgr.setCalendarPref_(aCalendar, 'X-OBM-EMAIL', aEmail.toString());
    },

    getOBMCalendars: function(aCount) {
        let calMgr = cal.getCalendarManager();

        function obmCalendars(x) { return x.type == "obm" }
        let cals = calMgr.getCalendars({}).filter(obmCalendars);
        aCount.value = cals.length;
        return cals;
    },

    observe: function(aSubject, aTopic, aData) {
        switch (aTopic) {
            case "profile-after-change":
                // These files need to be loaded here due to startup order issues with Thunderbird 3.
                // The calendar and obm-extension resource is not available at app-startup or at 
                // top-level when the file is loaded through xpcom. These lines can be moved to the 
                // top-level without the try/catch block when removing Thunderbird 3 compatibility
                try {
                    var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
                    Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
                } catch (e) {
                    Components.utils.reportError(e);
                }

                cal.getCalendarManager().addObserver(this);

                Cc["@mozilla.org/calendar/freebusy-service;1"]
                    .getService(Ci.calIFreeBusyService)
                    .addProvider(this);

                this.initLogger();

                break;
                // {{{ START_COMPAT_TB_3 This could be removed
            case "app-startup":
                Cc["@mozilla.org/observer-service;1"]
                    .getService(Ci.nsIObserverService)
                    .addObserver(this, "profile-after-change", false);
                // END_COMPAT_TB_3 }}}
        }
    },

    getFreeBusyIntervals: function(aCalId, aRangeStart, aRangeEnd, aBusyTypes, aListener) {
        this.logger.debug("getFreeBusyIntervals(" + aCalId + ", " + aRangeStart + ", " + aRangeEnd + ", " + aBusyTypes + ")");

        if (!aCalId || !/^mailto:/i.test(aCalId)) {
            return;
        }

        let operation = {
            id: Date.now(),
            isPending: true,
            status: Components.results.NS_OK,
            success: false,
            cancel: function(aStatus) {
                this.status = aStatus;
            }
        };

        let url = UtilsOBM.getPrefSafe("extensions.obm.server"),
            aOrganizer = {},
            targetCalendar = aCalId.substr('mailto:'.length);

        UtilsOBM.getOBMCredentials(aOrganizer, {});

        url = url.replace("services", "freebusy");
        url += "/" + encodeURIComponent(targetCalendar) + "?organizer=" + encodeURIComponent(aOrganizer.value);

        let xmlHttpRequest = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);

        this.logger.debug("Sending FreeBusy request to '" + url + "'.", false, false);

        xmlHttpRequest.open("GET", url, true);
        xmlHttpRequest.overrideMimeType("text/plain");
        xmlHttpRequest.onreadystatechange = function(event) {
            if (xmlHttpRequest.readyState == 4) {
                if (xmlHttpRequest.status == 200) {
                    try {
                        let freeBusyEntries = xmlHttpRequest.responseText.match(/FREEBUSY.+/g);
                        var freeBusyIntervals = [];

                        if (freeBusyEntries) {
                            for (let i = 0; i < freeBusyEntries.length; i++) {
                                let dates = freeBusyEntries[i].split(/FREEBUSY(;FBTYPE=BUSY)?:/)[2].split("/");

                                if (dates.length != 2) {
                                    continue;
                                }

                                let startDate = cal.createDateTime(dates[0]);
                                let endDate = cal.createDateTime(dates[1]);
                                let timeInterval = Components
                                    .classes["@mozilla.org/calendar/period;1"]
                                    .createInstance(Components.interfaces.calIPeriod);

                                timeInterval.start = startDate;
                                timeInterval.end = endDate;

                                freeBusyIntervals.push({
                                    calId: aCalId,
                                    interval: timeInterval,
                                    freeBusyType: Ci.calIFreeBusyInterval.BUSY
                                });
                            }
                        }
                        operation.success = true;
                    } catch (e) {
                        operation.success = false;
                        operation.status = Components.results.NS_ERROR_FAILED;
                        freeBusyIntervals = null;
                    }

                    operation.isPending = false;
                    aListener.onResult(operation, freeBusyIntervals);
                }
            }
        };

        xmlHttpRequest.send(null);
        return operation;
    },

    onCalendarRegistered: function(aCalendar) {
        // Force OBM resource calendars readonly
        let isOBMCalendar = (aCalendar.uri.spec.indexOf("/obm-sync/resources") > -1);
        if (aCalendar.type == "ics" && isOBMCalendar) {
            aCalendar.readOnly = true;
        }
    },

    onCalendarUnregistering: function(aCalendar) {},

    onCalendarDeleting: function(aCalendar) {
        if (aCalendar.type == "obm") {
            aCalendar.QueryInterface(Components.interfaces.calICalendarProvider)
                .deleteCalendar(aCalendar, null);
        }
    },

    onCalendarPrefSet: function(aCalendar, aName, aValue) {
        this.onCalendarPrefChanged(aCalendar, aName, aValue, null)
    },

    onCalendarPrefDeleting: function(aCalendar, aName) {
        this.onCalendarPrefChanged(aCalendar, aName, null, null)
    },

    onCalendarPrefChanged: function(aCalendar, aName, aValue, aOldValue) {
        // Make sure organizerId is set correctly on OBM resource calendars
        if (aCalendar.type == "ics") {
            let orgId = aCalendar.getProperty("X-OBM-EMAIL");
            if (aName == "X-OBM-EMAIL") {
                aCalendar.setProperty("organizerId", "mailto:" + aValue);
            } else if (aName == "organizerId" && aValue != orgId) {
                aCalendar.setProperty("organizerId", "mailto:" + orgId);
            }
        }
    }
};

var components = [ObmCalendarsManager];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}