/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmRESTClient"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js')
    // unixFileWriter

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://gre/modules/Services.jsm");
var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
Components.utils.importGlobalProperties(["XMLHttpRequest"]);

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const NS_ERROR_MODULE_BASE_OFFSET = 0x45;
const NS_ERROR_MODULE_NETWORK = 6;
const NS_ERROR_MODULE_SECURITY = 21;

// Global logger :-(
var gLogger;

function RESTClientParameters() {
    //this.parameters = [];
    this.parameters = new Map();
}

RESTClientParameters.prototype = {

    parameters: null,

    add: function add(aName, aValue) {
        //        if (!(aName in this.parameters)) {
        //            this.parameters[aName] = [];
        //        }
        //	this.parameters[aName].push(aValue);
        this.parameters.set(aName, aValue);
        return this;
    },

    /**
     * Returns the properties of the object as a string of (URL-escaped) query parameters.
     */
    toQueryParameters: function toQueryParamters() {
        var result = "";

        gLogger.debug("toQueryParameters SEL12 parameters['origin']" + this.parameters.get('origin'), false, false);
        gLogger.debug("toQueryParameters SEL14 parameters.length" + this.parameters.length, false, false);
        //        for (let paramName of this.parameters) {
        //		gLogger.debug("toQueryParameters SEL13 paramName:", false, false);
        //            for (let paramValue of this.parameters[paramName]) {
        //                var elem = paramName + "=" + encodeURIComponent(paramValue);
        //                result += elem + "&";
        //            }
        //        }
        for (var [key, value] of this.parameters) {
            gLogger.debug("toQueryParameters SEL15 parameters.length" + key, false, false);
            if (Array.isArray(value)) {
                value.forEach(function(e) {
                    var elem = key + "=" + encodeURIComponent(e);
                    result += elem + "&";
                })
            } else {
                var elem = key + "=" + encodeURIComponent(value);
            }
            result += elem + "&";
        }

        result = result.replace(/&$/, "");

        var debugResult = result.replace(/password=[^&]*/, "password=******");
        gLogger.debug(debugResult, false, false);

        return result;
    }
}

function ObmRESTClient() {
    gLogger = new ObmLogger();

    gLogger.setLevelWriter(gLogger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    gLogger.setHeader("RESTClient.js: ");
}

ObmRESTClient.prototype = {
    // nsISupports interface
    QueryInterface: ChromeUtils.generateQI([Ci.nsIObserver]),
    classID: Components.ID("{00c18d10-bae3-11dd-ad8b-0800200c9a66}"),
    contractID: "@obm.org/restclient;1",
    classDescription: "REST Client",

    // {{{ START_COMPAT_TB_3
    _xpcom_categories: [{
        category: "app-startup",
        service: true
    }],
    // END_COMPAT_TB_3 }}}
    get wrappedJSObject() {
        return this;
    },
    observe: function(aSubject, aTopic, aData) {
        switch (aTopic) {
            case "profile-after-change":
                try {
                    Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
                } catch (e) {
                    Components.utils.reportError(e);
                }

                break;
                // {{{ START_COMPAT_TB_3 This could be removed
            case "app-startup":
                Cc["@mozilla.org/observer-service;1"]
                    .getService(Ci.nsIObserverService)
                    .addObserver(this, "profile-after-change", false);

                break;
                // END_COMPAT_TB_3 }}}
        }
    },

    username: null,
    password: null,

    newRESTClientParameters: function() {
        return new RESTClientParameters();
    },
    invoke: function(url, method, parameters, async, callback, aData) {
        let xmlHttp = this._getXmlHttp();
        let fullUri = url + "/" + method;
        gLogger.debug("invoke: " + fullUri, false, false);
        if (this.userName && this.password) {
            gLogger.debug("userName: " + this.userName, false, false);
            // gLogger.debug("password: " + this.password , false, false);
            xmlHttp.open("POST", fullUri, async, this.userName, this.password);

            // Some WS implementations (i.e. BEA WebLogic Server 10.0 JAX-WS) don't support Challenge/Response HTTP BASIC, so we send authorization headers in the first request
            xmlHttp.setRequestHeader("Authorization", "Basic " + this._toBase64(this.userName + ":" + this.password));
        } else {
            xmlHttp.open("POST", fullUri, async);
        }

        xmlHttp.setRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=utf-8");

        var RESTClient = this;
        let restargs = arguments;
        let restartingRequest = false;

        if (async && callback) {
            xmlHttp.onerror = function() {
                if (!restartingRequest) {
                    callback(method, new Error(500, "Connection failure"), null, aData);
                }
            };
            xmlHttp.onreadystatechange = function() {
                function isErrorModule(status, module) {
                    let targetModule = (((status >> 16) - NS_ERROR_MODULE_BASE_OFFSET) & 0x1fff);
                    return (!Components.isSuccessCode(status) && targetModule == module);
                }

                if (xmlHttp.readyState == 4) {
                    let errCode = xmlHttp.status || (xmlHttp.channel ? xmlHttp.channel.status : 0);
                    let errString = xmlHttp.statusText;

                    if (isErrorModule(xmlHttp.channel.status, NS_ERROR_MODULE_SECURITY)) {
                        let params = {
                            exceptionAdded: false,
                            prefetchCert: true,
                            location: fullUri
                        };

                        let window = Services.wm.getMostRecentWindow("obm:sync-dialog") ||
                            cal.getCalendarWindow() ||
                            Services.appShell.hiddenDOMWindow;
                        window.openDialog("chrome://pippki/content/exceptionDialog.xul",
                            "", "chrome,centerscreen,modal", params);
                        if (params.exceptionAdded) {
                            restartingRequest = true;
                            return RESTClient.invoke.apply(RESTClient, restargs);
                        } else {
                            callback(method, new Error(403, "Invalid Certificate"), null, aData);
                        }
                    } else if (isErrorModule(xmlHttp.channel.status, NS_ERROR_MODULE_NETWORK)) {
                        callback(method, new Error(errCode, "Connection failure"), null, aData);
                    } else if (xmlHttp.status == 200) {
                        RESTClient._onSendRESTRequest(method, async, callback, xmlHttp, aData);
                    } else {
                        callback(method, new Error(errCode, errString), null, aData);
                    }
                }
            }
        }

        xmlHttp.send(parameters.toQueryParameters());
        if (!async) {
            return this._onSendRESTRequest(method, async, callback, xmlHttp, aData);
        }
    },

    _onSendRESTRequest: function(method, async, callback, req, aData) {
        var result = null;
        var responseXML = req.responseXML;

        gLogger.debug("_onSendRESTRequest: " + method + " => " + req.responseText, false, false);
        if (responseXML.getElementsByTagName("error").length > 0) {
            var errorStr;
            var messageError = req.responseXML.getElementsByTagName("message")[0];
            errorStr = (messageError.childNodes[0] == null) ? "Server internal error" : messageError.childNodes[0].nodeValue;
            result = new Error(500, errorStr);
        } else {
            result = "ok";
        }

        if (callback) {
            callback(method, result, responseXML, aData);
        }

        return result;
    },

    _getXmlHttp: function() {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
        let xhr = new XMLHttpRequest();
        xhr.timeout = UtilsOBM.getPrefSafe("extensions.obm.network.connection.timeout", 10000);
        return xhr;
    },

    _toBase64: function(input) {
        var keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        var output = "";
        var chr1, chr2, chr3;
        var enc1, enc2, enc3, enc4;
        var i = 0;

        do {
            chr1 = input.charCodeAt(i++);
            chr2 = input.charCodeAt(i++);
            chr3 = input.charCodeAt(i++);

            enc1 = chr1 >> 2;
            enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
            enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
            enc4 = chr3 & 63;

            if (isNaN(chr2)) {
                enc3 = enc4 = 64;
            } else if (isNaN(chr3)) {
                enc4 = 64;
            }

            output = output + keyStr.charAt(enc1) + keyStr.charAt(enc2) +
                keyStr.charAt(enc3) + keyStr.charAt(enc4);
        } while (i < input.length);

        return output;
    }
};

// module
var components = [ObmRESTClient];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}