/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["ObmSyncManager"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");

Components.utils.import("resource://obm-extension/modules/GlobalData.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");


const Cc = Components.classes;
const Ci = Components.interfaces;

function GenericWrappedXPComObject() {}
GenericWrappedXPComObject.prototype = {
    QueryInterface: ChromeUtils.generateQI([]),
    get wrappedJSObject() { return this; }
};

var stringBundle;

function getLocalizedMessage(aString) {
    dump("" + aString + "\n");
    try {
        return stringBundle.GetStringFromName("obm.sync." + aString);
    } catch (e) {
        return aString;
    }
}

function debugError(prefix, logger, aMessage) {
    if (typeof aMessage != "string") {
        var errmsg =
            aMessage.message +
            " file:" +
            aMessage.fileName +
            " line:" +
            aMessage.lineNumber +
            "\n" +
            aMessage.stack;
        logger.debug(prefix + errmsg, false, false);
    } else
        logger.debug(prefix + aMessage + "(" + typeof aMessage + ")", false, false);
}

/** Backup value for "dom.max_script_run_time" pref. */
var _max_script_run_time;
/** Backup value for "dom.max_chrome_script_run_time" pref. */
var _max_chrome_script_run_time;

/**
 * Prevents warning popups from appearing when a script takes too long by
 * disabling the timeouts.
 */
function _disableTimeSecurity() {
    _max_script_run_time = UtilsOBM.getPrefSafe("dom.max_script_run_time", 10);
    _max_chrome_script_run_time = UtilsOBM.getPrefSafe(
        "dom.max_chrome_script_run_time",
        20
    );
    UtilsOBM.setPrefSafe("dom.max_script_run_time", 0);
    UtilsOBM.setPrefSafe("dom.max_chrome_script_run_time", 0);
    OBMGlobalData.addrSynchronizing = true;
}

/**
 * Restores previous values for script timeouts.
 */
function _restoreTimeSecurity() {
    UtilsOBM.setPrefSafe("dom.max_script_run_time", _max_script_run_time);
    UtilsOBM.setPrefSafe(
        "dom.max_chrome_script_run_time",
        _max_chrome_script_run_time
    );
    OBMGlobalData.addrSynchronizing = false;
}

function ObmSyncManager() {
    stringBundle = Cc["@mozilla.org/intl/stringbundle;1"]
        .getService(Ci.nsIStringBundleService)
        .createBundle("chrome://obm-extension/locale/errors_warnings.properties");
    this.logger = new ObmLogger();

    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("obmSyncManager.js: ");
    this.logger.debug("new ObmSyncManager()", false, false);
    this.init()
}

ObmSyncManager.prototype = {
    QueryInterface: ChromeUtils.generateQI([Ci.obmISyncManager, Ci.nsIObserver]),
    classID: Components.ID("{e37931ed-104f-4cf1-9b5c-73fab2d65424}"),
    contractID: "@obm.org/syncmanager;1",
    classDescription: "OBM Sync Manager",
    synchronizationErrorAckRequired: false,

    // {{{ START_COMPAT_TB_3
    _xpcom_categories: [{
        category: "app-startup",
        service: true
    }],
    // END_COMPAT_TB_3 }}}

    init: function() {
        this.initTimers()
    },

    // Human-readable operation names, for the interface
    _humanOpNames: {
        logUserIn: "login",
        getSync: "checking",
        getAddressBookSync: "checking",
        getSyncWithSortedChanges: "checking",
        mergeRemoteDiff: "merging",
        sendLocalDiff: "sending",
        listCategories: "merging"
    },

    shouldManuallySynchronize: function(syncReadiness) {
        let isReadyForSync = [
            SyncReadiness.SYNC_ACK_REQUIRED,
            SyncReadiness.SYNC_OFFLINE,
            SyncReadiness.SYNC_NOTIFICATION_ACK_REQUIRED,
            SyncReadiness.SYNC_WINDOW_OPEN
        ].indexOf(syncReadiness) < 0;
        return isReadyForSync;
    },

    shouldAutoSynchronize: function(syncReadiness) {
        return syncReadiness == SyncReadiness.SYNC_OK;
    },

    showSyncErrorMessage: function(syncReadiness) {
        let messageName = this.getSyncReadinessMessageKey(syncReadiness);
        if (messageName == null) {
            return;
        }

        let win = this._getXulBrowserWindow();
        if (win != null) {
            win.setJSStatus(
                "OBM Sync - " + UtilsOBM.getLocalizedMessage(messageName)
            );
        }
    },

    isSyncWindowOpen: function() {
        let wWatcher = Components.classes[
            "@mozilla.org/embedcomp/window-watcher;1"
        ].getService(Components.interfaces.nsIWindowWatcher);
        // Don't synchronize if there is already a sync window open
        return wWatcher.getWindowByName("OBM_SYNCHRO", null) != null;
    },

    isNotificationDialogOpen: function() {
        let wWatcher = Components.classes[
            "@mozilla.org/embedcomp/window-watcher;1"
        ].getService(Components.interfaces.nsIWindowWatcher);

        // Don't synchronize if there is a notification dialog open
        let we = wWatcher.getWindowEnumerator();
        while (we.hasMoreElements()) {
            let testw = we.getNext();
            if (testw.name.indexOf("notificationDialog:") == 0) {
                this.logger.debug(
                    "refresh canceled : notification window open",
                    false,
                    false
                );
                return true;
            }
        }
        return false;
    },

    getSyncReadiness: function() {
        let usernamePref = {},
            passwordPref = {};
        UtilsOBM.getOBMCredentials(usernamePref, passwordPref);

        if (UtilsOBM.getPrefSafe("extensions.obm.server", "") == "") {
            return SyncReadiness.SYNC_SERVER_MISSING;
        } else if (usernamePref.value == "") {
            return SyncReadiness.SYNC_USERNAME_MISSING;
        } else if (passwordPref.value == UtilsOBM.PASSWORD_NOT_CONFIGURED) {
            return SyncReadiness.SYNC_PASSWORD_MISSING;
        } else if (this.synchronizationErrorAckRequired) {
            return SyncReadiness.SYNC_ACK_REQUIRED;
        } else if (!UtilsOBM.isOnline()) {
            return SyncReadiness.SYNC_OFFLINE;
        } else if (this.isNotificationDialogOpen()) {
            return SyncReadiness.SYNC_NOTIFICATION_ACK_REQUIRED;
        } else if (this.isSyncWindowOpen()) {
            return SyncReadiness.SYNC_WINDOW_OPEN;
        } else {
            return SyncReadiness.SYNC_OK;
        }
    },

    getSyncReadinessMessageKey: function(syncReadiness) {
        switch (syncReadiness) {
            case SyncReadiness.SYNC_SERVER_MISSING:
                return "obm.syncerror.serverPrefEmpty";
            case SyncReadiness.SYNC_USERNAME_MISSING:
                return "obm.syncerror.usernamePrefEmpty";
            case SyncReadiness.SYNC_PASSWORD_MISSING:
                return "obm.syncerror.passwordPrefEmpty";
            default:
                return null;
        }
    },

    syncWithDialog: function(aParent) {
        function hasAllLoginData(username, password, obmServer) {
            obmServer.value = UtilsOBM.getPrefSafe("extensions.obm.server", "");
            return (
                UtilsOBM.getOBMCredentials(username, password) &&
                username.value != "" &&
                password.value != "" &&
                password.value != UtilsOBM.PASSWORD_NOT_CONFIGURED &&
                obmServer.value
            );
        }
        let syncReadiness = this.getSyncReadiness();
        if (!this.shouldManuallySynchronize(syncReadiness)) {
            return;
        }

        try {
            let username = {},
                password = {},
                obmServer = {};
            if (!hasAllLoginData(username, password, obmServer)) {
                if (!username.value || !obmServer.value) {
                    let promptService = Components.classes[
                        "@mozilla.org/embedcomp/prompt-service;1"
                    ].getService(Components.interfaces.nsIPromptService);
                    let title = UtilsOBM.getLocalizedMessage("obmconf");
                    let message = UtilsOBM.getLocalizedMessage("obmconfsyncserver");
                    promptService.alert(null, title, message);

                    if (!UtilsOBM.showWindow("preferences/connection.xul", aParent)) {
                        return;
                    }
                } else {
                    if (!UtilsOBM.showWindow("preferences/password.xul", aParent) &&
                        !UtilsOBM.getTemporaryPassword(password)
                    ) {
                        return;
                    }
                }
            }

            if (hasAllLoginData(username, password, obmServer)) {
                // Only open the sync dialog if we now have all data. For
                // example, it could be that the user didn't enter the
                // username/password in the connection prefs dialog
                UtilsOBM.showWindow("sync-dialog.xul", aParent, "OBM_SYNCHRO");
            }
        } catch (e) {
            this.logger.error(e, true, false);
        }
    },

    // Must be declared in the myComponent interface idl
    syncAll: function() {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
        var online = UtilsOBM.isOnline();
        if (!online) {
            this.logger.debug(
                "ObmSyncManager.syncAll(): we're offline, no sync",
                false,
                false
            );
            return;
        }
        this.logger.debug("ObmSyncManager.syncAll()", false, false);
        _disableTimeSecurity();
        var ObserverServ = Cc["@mozilla.org/observer-service;1"].getService(
            Ci.nsIObserverService
        );
        Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRAddressBookSync.js")
        var abSync = new ObmRAddressBookSync();
        if (
            UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) ||
            (UtilsOBM.getPrefSafe("extensions.obm.calTasks.is_sync_enabled", true) &&
                Ci.calICalendar)
        ) {
            Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
            var calSync = new ObmRCalendarSync();
        }

        /*
                The operations array stores structs like:
                    - Name of action to perform
                    - function to call in order to perform the action
            */
        var opIdx = 0;
        var operations = [];
        if (
            UtilsOBM.getPrefSafe("extensions.obm.addressbooks.is_sync_enabled", true)
        ) {
            this.logger.debug(
                "Read prefs: extensions.obm.addressbooks.is_sync_enabled=true",
                false,
                false
            );
            operations.push([
                "card",
                function(arg) {
                    abSync.mergeAddressBook(arg);
                }
            ]);
        }
        if (
            UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) ||
            UtilsOBM.getPrefSafe("extensions.obm.calTasks.is_sync_enabled", true)
        ) {
            this.logger.debug(
                "Read prefs: extensions.obm.calEvents.is_sync_enabled=true",
                false,
                false
            );
            this.logger.debug(
                "OR extensions.obm.calTasks.is_sync_enabled=true",
                false,
                false
            );

            operations.push([
                "cal",
                function(arg) {
                    calSync.mergeAllCalendars(arg);
                }
            ]);
        }
        if (
            UtilsOBM.getPrefSafe(
                "extensions.obm.calendar.resourcesSyncRangeUpdated",
                true
            )
        ) {
            Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRResourceSync.js");
            var resourceSync = new ObmRResourceSync();
            operations.push([
                "resources",
                function(arg) {
                    resourceSync.subscribeToAllResources(arg);
                }
            ]);
        }

        /*
              the addressbooks naming changes require us to reset extensions.obm.addressbooks property
              on the first run of the new extension code
              => add a new preference extensions.obm.extensionUpgrade24 = true
            */
        this.logger.debug(
            "check extensions.obm.extensionUpgrade24: " +
            UtilsOBM.getPrefSafe("extensions.obm.extensionUpgrade24", false),
            false,
            false
        );
        if (
            UtilsOBM.getPrefSafe("extensions.obm.extensionUpgrade24", false) === false
        ) {
            this.logger.debug(
                "set extensions.obm.extensionUpgrade24 to true",
                false,
                false
            );
            // forget all ab
            UtilsOBM.setPrefSafe("extensions.obm.addressbooks", "");
            // reset lastSync to force complete sync from the server
            UtilsOBM.setPrefSafe(
                "extensions.obm.changes.nsIAbDirectory.lastSync",
                "0"
            );
            UtilsOBM.setPrefSafe("extensions.obm.extensionUpgrade24", true);
        }

        // Keep a pointer on the logger to use in the following struct
        var this_logger = this.logger;
        var listener = {
            onStart: function() {
                this_logger.debug("[syncAll] onStart", false, false);
            },

            onNewOperation: function(aOpName, aOpData) {
                this_logger.debug(
                    "[syncAll] onNewOperation: opName=" + aOpName + ", opData=" + aOpData,
                    false,
                    false
                );
                var msg = new GenericWrappedXPComObject();
                msg.opName = aOpName;
                msg.opData = aOpData;
                ObserverServ.notifyObservers(
                    msg,
                    "verbose-obm-sync-progress",
                    "newOperation"
                );
            },
            onOperationDone: function(aSuccess, aResultData) {
                this_logger.debug(
                    "[syncAll] onOperationDone: success=" +
                    aSuccess +
                    ", resultData=" +
                    aResultData,
                    false,
                    false
                );
                var msg = new GenericWrappedXPComObject();
                msg.opRes = aSuccess;
                ObserverServ.notifyObservers(
                    msg,
                    "verbose-obm-sync-progress",
                    "operationDone"
                );
            },
            onProgress: function(aIdx, aTotal, aType, aEntity) {
                this_logger.debug(
                    "[syncAll] onProgress: opIdx=" +
                    aIdx +
                    ", total=" +
                    aTotal +
                    ", type=" +
                    aType +
                    ", entity=" +
                    aEntity,
                    false,
                    false
                );
                var msg = new GenericWrappedXPComObject();
                msg.idx = aIdx + 1;
                msg.total = aTotal;
                msg.type = aType;
                msg.entity = aEntity;
                ObserverServ.notifyObservers(
                    msg,
                    "verbose-obm-sync-progress",
                    "operationProgress"
                );
            },
            onError: function(aMessage) {
                debugError("[syncAll] onError: ", this_logger, aMessage);
                var msg = new GenericWrappedXPComObject();
                msg.message = aMessage;
                ObserverServ.notifyObservers(
                    msg,
                    "verbose-obm-sync-progress",
                    "operationSetError"
                );
            },
            onStop: function(aRes, aMessage) {
                debugError(
                    "[syncAll] onStop: res=" + aRes + ", message=",
                    this_logger,
                    aMessage
                );

                var msg = new GenericWrappedXPComObject();
                msg.message = aMessage;
                if (aRes == Components.results.NS_OK) {
                    ObserverServ.notifyObservers(
                        msg,
                        "verbose-obm-sync-progress",
                        "operationSetDone"
                    );
                    this.nextOperation();
                } else {
                    ObserverServ.notifyObservers(
                        msg,
                        "verbose-obm-sync-progress",
                        "operationSetErrorFatal"
                    );
                    ObserverServ.notifyObservers(null, "obmCalendarSyncCompleted", 0);
                    _restoreTimeSecurity();
                }
            },
            nextOperation: function() {
                if (opIdx >= operations.length) {
                    this_logger.debug(
                        "[syncAll] nextOperation: No more operation.",
                        false,
                        false
                    );
                    ObserverServ.notifyObservers(
                        null,
                        "verbose-obm-sync-progress",
                        "syncDone"
                    );
                    // re-compute the lightning "blue bar"
                    var observer = Components.classes[
                        "@mozilla.org/observer-service;1"
                    ].getService(Components.interfaces.nsIObserverService);
                    observer.notifyObservers(null, "onItipItemCreation", 0);
                    _restoreTimeSecurity();
                    return;
                }

                this_logger.debug(
                    "[syncAll] nextOperation: " + operations[opIdx][0],
                    false,
                    false
                );
                var msg = new GenericWrappedXPComObject();
                msg.opSetName = operations[opIdx][0];
                ObserverServ.notifyObservers(
                    msg,
                    "verbose-obm-sync-progress",
                    "newOperationSet"
                );
                try {
                    operations[opIdx++][1](this);
                } catch (e) {
                    this.onError(e);
                }
            }
        };

        listener.nextOperation();
    },

    syncAddressBook: function() {
        this.logger.debug("syncAddressBook()", false, false);

        if (!UtilsOBM.getPrefSafe("extensions.obm.addressbooks.is_sync_enabled", true)) {
            this.logger.debug(
                "syncAddressBook() Address book sync disabled, skipping...",
                false,
                false
            );

            return;
        }

        Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRAddressBookSync.js")
        let abSync = new ObmRAddressBookSync();

        let operations = [
            [
                "AddressBook",
                function(arg) {
                    abSync.mergeAddressBook(arg);
                }
            ]
        ];
        let listener = this._getListenerForSyncOperations(operations);

        _disableTimeSecurity();

        listener.nextOperation();
    },

    syncCalendar: function(aCalendar) {
        this.logger.debug("syncCalendar(" + aCalendar.name + ")", false, false);

        Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
        let calSync = new ObmRCalendarSync();
        let operations = [
            [
                "Calendar",
                function(arg) {
                    calSync.mergeOneCalendar(aCalendar, arg);
                }
            ]
        ];
        let listener = this._getListenerForSyncOperations(operations);

        _disableTimeSecurity();

        listener.nextOperation();
    },

    syncAllCalendars: function() {
        this.logger.debug("syncAllCalendars()", false, false);

        Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
        let calSync = new ObmRCalendarSync();
        let operations = [
            [
                "Calendar",
                function(arg) {
                    calSync.mergeAllCalendars(arg);
                }
            ]
        ];
        let listener = this._getListenerForSyncOperations(operations);

        _disableTimeSecurity();

        listener.nextOperation();
    },

    purgeCalendar: function(calName) {
        this.logger.debug("purgeCalendar(" + calName + ")", false, false);

        _disableTimeSecurity();

        Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
        let calSync = new ObmRCalendarSync();
        let opIdx = 0;
        let operations = [
            [
                "Calendar",
                function(arg) {
                    calSync.purge(calName, arg);
                }
            ]
        ];
        let listener = {
            onStart: function() {},

            onNewOperation: function(aOpName) {},
            onOperationDone: function(aResult) {},
            onProgress: function(aIdx, aTotal, aType, aEntity) {},
            onError: function(aMessage) {},
            onStop: function(aRes, aMessage) {
                if (aRes == Components.results.NS_OK) {
                    this.nextOperation();
                } else {
                    var promptService = Components.classes[
                        "@mozilla.org/embedcomp/prompt-service;1"
                    ].getService(Components.interfaces.nsIPromptService);
                    var title = UtilsOBM.getLocalizedMessage("obm.syncerror.srvError");
                    var message = UtilsOBM.getLocalizedMessage(
                        "obm.syncerror.srvErrorExplained" + aMessage
                    );
                    promptService.alert(null, title, message);

                    _restoreTimeSecurity();
                }
            },
            nextOperation: function() {
                if (opIdx >= operations.length) {
                    _restoreTimeSecurity();
                    return;
                }

                try {
                    operations[opIdx++][1](this);
                } catch (e) {
                    this.onError(e);
                }
            }
        };

        listener.nextOperation();
    },

    _getListenerForSyncOperations: function(operations) {
        let opIdx = 0;
        let syncManager = this;
        let window = this._getXulBrowserWindow();
        let promptService = Components.classes[
            "@mozilla.org/embedcomp/prompt-service;1"
        ].getService(Components.interfaces.nsIPromptService);
        let observerService = Cc["@mozilla.org/observer-service;1"].getService(
            Ci.nsIObserverService
        );

        let listener = {
            onStart: function() {},

            onNewOperation: function(aOpName) {
                // let opName = syncManager._humanOpNames[aOpName];
            },

            onOperationDone: function(aResult) {},

            onProgress: function(aIdx, aTotal, aType, aEntity) {},

            onError: function(aMessage) {},

            onStop: function(aRes, aMessage) {
                if (aRes == Components.results.NS_OK) {
                    this.nextOperation();
                } else {
                    observerService.notifyObservers(null, "obmCalendarSyncCompleted", 0);
                    let alertUser = UtilsOBM.getPrefSafe(
                        "extensions.obm.backgroundErrorMustBeAcknowledged",
                        false
                    );

                    if (alertUser) {
                        let title = UtilsOBM.getLocalizedMessage("obmconnector");

                        syncManager.synchronizationErrorAckRequired = true;
                        promptService.alert(null, title, aMessage);
                        syncManager.synchronizationErrorAckRequired = false;
                    }

                    _restoreTimeSecurity();
                }
            },

            nextOperation: function() {
                let syncReadiness = syncManager.getSyncReadiness();
                if (!syncManager.shouldManuallySynchronize(syncReadiness)) {
                    syncManager.logger.debug(
                        "Synchronization canceled: an error must be acknowledged. Alert open for user",
                        false,
                        false
                    );
                    return;
                }

                if (opIdx >= operations.length) {
                    _restoreTimeSecurity();

                    return;
                }

                let op = operations[opIdx++];

                this.head = "OBM Sync - " + getLocalizedMessage(op[0]) + ": ";

                try {
                    op[1](this);
                } catch (e) {
                    this.onError(e);
                }
            }
        };

        return listener;
    },

    _getXulBrowserWindow: function() {
        let windowsMediator = Cc[
            "@mozilla.org/appshell/window-mediator;1"
        ].getService(Ci.nsIWindowMediator);
        let windows = windowsMediator.getEnumerator("mail:3pane");

        if (windows.hasMoreElements()) {
            return windows
                .getNext()
                .QueryInterface(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIWebNavigation)
                .QueryInterface(Ci.nsIDocShellTreeItem)
                .treeOwner.QueryInterface(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIXULWindow).XULBrowserWindow;
        }

        return null;
    },

    initTimers: function() {
        this.logger.error("=========== InitTimers");
        if (this.mContactTimer) this.mContactTimer.cancel();
        if (this.mCalendarTimer) this.mCalendarTimer.cancel();

        let timeout =
            UtilsOBM.getPrefSafe("extensions.obm.contact.autorefresh.timeout", 0) *
            60 *
            1000;

        if (timeout != 0) {
            this.logger.error("=========== InitTimers ------> Create CONTACT Timer");

            this.mContactTimer = Cc["@mozilla.org/timer;1"].createInstance(
                Ci.nsITimer
            );
            this.mContactTimer.init(this, timeout, Ci.nsITimer.TYPE_REPEATING_SLACK);
        }

        timeout =
            (UtilsOBM.getPrefSafe("extensions.obm.calendar.autorefresh.timeout", 0) ||
                // {{{ START_COMPAT_TB_3
                UtilsOBM.getPrefSafe("calendar.autorefresh.timeout", 0)) *
            // END_COMPAT_TB_3 }}}
            60 *
            1000;

        if (timeout > 0) {
            this.logger.error("=========== InitTimers ------> Create calendar Timer");
            this.mCalendarTimer = Cc["@mozilla.org/timer;1"].createInstance(
                Ci.nsITimer
            );

            let self = this;

            function setupRepeatingCalendarTimer() {
                self.mCalendarTimer.init(
                    self,
                    timeout,
                    Ci.nsITimer.TYPE_REPEATING_SLACK
                );
            }
            let firstSyncDelay = UtilsOBM.getPrefSafe(
                "extensions.obm.firstSyncDelay",
                0
            );
            if (firstSyncDelay > 0) {
                // If set, the first refresh is delayed some time between now
                // and the autorefresh interval.
                let delay = UtilsOBM.getRandomInt(0, firstSyncDelay * 60 * 1000);
                this.mCalendarTimer.initWithCallback({
                        notify: function() {
                            self.syncAddressBook();
                            self.syncAllCalendars();
                            setupRepeatingCalendarTimer();
                        }
                    },
                    delay,
                    Ci.nsITimer.TYPE_ONE_SHOT
                );
            } else {
                // Otherwise set up the fixed repeating timer right away
                setupRepeatingCalendarTimer();
            }
        }
    },

    checkPurgeCalendars: function() {
        let username = {},
            password = {};
        if (
            UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) ||
            !UtilsOBM.getOBMCredentials(username, password) ||
            !username.value
        ) {
            // Not ready to purge yet, bail out
            return;
        }

        let calMgr = cal.getCalendarManager();
        let name = username.value.split("@")[0];
        let lastPurgeAsk = UtilsOBM.getPrefSafe("extensions.obm.lastPurgeAsk", 0);

        for (let calendar of calMgr.getCalendars({})) {
            if (calendar.type != "obm") {
                continue;
            }

            let calName = calendar.getProperty("X-OBM-NAME");
            if (calname == name) {
                if (lastPurgeAsk != 0) {
                    lastPurgeAsk = cal.dtz.jsDateToDateTime(
                        UtilsOBM.restDateToJSDate(lastPurgeAsk),
                        cal.dtz.calendarDefaultTimezone
                    );
                    let elapsed = cal.now().subtractDate(lastPurgeAsk);
                    if (elapsed.days >= 15) {
                        this.purgeCalendar(calendar);
                    }
                }
                UtilsOBM.setPrefSafe(
                    "extensions.obm.lastPurgeAsk",
                    new Date().getTime()
                );

                // We are only interested in purging the user's own calendar,
                // break out now.
                break;
            }
        }
    },

    observe: function(aSubject, aTopic, aData) {
        this.logger.debug("Observe SyncManager");
        switch (aTopic) {
            case "profile-after-change":
                try {
                    Components.utils.import(
                        "resource://obm-extension/modules/UtilsOBM.jsm"
                    );
                    Components.utils.import(
                        "resource://obm-extension/modules/GlobalData.jsm"
                    );
                    var { cal } = ChromeUtils.import(
                        "resource://calendar/modules/calUtils.jsm",
                        null
                    );
                } catch (e) {
                    Components.utils.reportError(e);
                }

                this.initLogger();
                this.initTimers();
                this.checkPurgeCalendars();

                break;
                // {{{ START_COMPAT_TB_3 This could be removed
            case "app-startup":
                Cc["@mozilla.org/observer-service;1"]
                    .getService(Ci.nsIObserverService)
                    .addObserver(this, "profile-after-change", false);

                break;
                // END_COMPAT_TB_3 }}}
            case "timer-callback":
                this.logger.debug("timer-callback");
                try {
                    let syncReadiness = this.getSyncReadiness();
                    if (this.shouldAutoSynchronize(syncReadiness)) {
                        if (aSubject == this.mContactTimer) {
                            this.syncAddressBook();
                        } else if (aSubject == this.mCalendarTimer) {
                            this.syncAllCalendars();
                        }
                    } else {
                        this.showSyncErrorMessage(syncReadiness);
                    }
                } catch (e) {
                    this.logger.debug("Error during timer-callback: " + e);
                }
                break;
        }
    }
};

var SyncReadiness = {
    SYNC_SERVER_MISSING: "SYNC_SERVER_MISSING",
    SYNC_USERNAME_MISSING: "SYNC_USERNAME_MISSING",
    SYNC_PASSWORD_MISSING: "SYNC_PASSWORD_MISSING",
    SYNC_ACK_REQUIRED: "SYNC_ACK_REQUIRED",
    SYNC_OFFLINE: "SYNC_OFFLINE",
    SYNC_NOTIFICATION_ACK_REQUIRED: "SYNC_NOTIFICATION_ACK_REQUIRED",
    SYNC_WINDOW_OPEN: "SYNC_WINDOW_OPEN",
    SYNC_OK: "SYNC_OK"
};

function ServerPreferenceEmptyException() {}

function UsernamePreferenceEmptyException() {}

function PasswordPreferenceEmptyException() {}

var components = [ObmSyncManager];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}