/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYBOLS = ["obmVCardHandler"];

var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function obmVCardHandler() {
    try {
        Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
    } catch (e) {
        Components.utils.reportError(e);
    }
    this.logger = new ObmLogger();

    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("obmVCardHandler:");
}

obmVCardHandler.prototype = {
    QueryInterface: ChromeUtils.generateQI([Components.interfaces.obmIVCardHandler]),
    classID: Components.ID("{431dd1bc-6d57-42f1-b08f-18c928dd5169}"),
    contractID: "@obm.org/vcard-handler;1",
    classDescription: "OBM text/x-vcard handler",
    getCurrentVCard: function(aMsgWindow, aDbView, aCallback) {

        var msgURI = null;
        if (aDbView) {
            msgURI = aDbView.URIForFirstSelectedMessage;
        }

        if (msgURI == null) {
            if (aCallback)
                aCallback.onResponse(null, "");
            return;
        }
        //this.logger.debug("getCalendarURIFromVCard : msgURI="+msgURI,false,false);

        // let's retrieve the whole mail (its source text)
        // the streamlistener will extract the vcard
        var messenger = Components.classes["@mozilla.org/messenger;1"]
            .createInstance(Components.interfaces.nsIMessenger);
        var msgService = messenger.messageServiceFromURI(msgURI);
        this._streamListener._logger = this.logger;
        this._streamListener._callback = aCallback;
        msgService.CopyMessage(msgURI, this._streamListener, false, null, aMsgWindow, {});
    },

    /**
     * the stream listener called during the retrieval of the mail
     * TODO This will create a static object shared on all prototype instances,
     * really wanted?
     */
    _streamListener: {
        _content: "",
        _logger: null,
        _callback: null,
        onDataAvailable: function(request, context, inputStream, offset, count) {
            //this._logger.debug("~~~~~~ onDataAvailable ("+count+")", false, false);
            try {
                var sis = Components.classes["@mozilla.org/scriptableinputstream;1"]
                    .createInstance(Components.interfaces.nsIScriptableInputStream);
                sis.init(inputStream);
                this._content += sis.read(count);
            } catch (ex) {}
        },
        onStartRequest: function(request, context) {
            this._content = "";
        },
        onStopRequest: function(aRequest, aContext, aStatusCode) {
            //this._logger.debug("~~~~~~ onStopRequest", false, false);

            // we now have the mail
            this._content = this._content.replace(/\r/g, "");

            // we extract the header to know the content type of the mail
            var posendHeader = this._content.indexOf("\n\n");
            var header = this._content.substr(0, posendHeader + 1); // keep a \n
            var headersParser = Components.classes["@mozilla.org/messenger/mimeheaders;1"]
                .createInstance(Components.interfaces.nsIMimeHeaders);
            headersParser.initialize(header, header.length);

            var contentType = headersParser.extractHeader("Content-Type", false);
            var vcard = null;
            var obmCalendarURI = "";
            if (contentType && contentType.indexOf('multipart/mixed') == 0) {
                // the mail contains attachments
                //var m = /boundary="([^"]+)"/gi.exec(contentType); // it gives random results !??  FIXME: use RegExp instead
                //if (m != null) {
                var bs = contentType.indexOf('boundary="');
                var be = contentType.indexOf('"', bs + 10);
                if (bs != -1 && be != -1) {
                    var boundary = contentType.substr(bs + 10, be - bs - 10);
                    //this._logger.debug(' boundary="'+boundary+'"',false,false);
                    var body = this._content.substr(posendHeader + 2);
                    var attachments = body.split(boundary);
                    var obmad = UtilsOBM.getAddressbookManager();
                    for (var i = 0; i < attachments.length; i++) {
                        if (attachments[i].search(/^\s*Content-type\s*:\s*text\/x-vcard;/i) == 0) {
                            this._logger.debug("~~~~~~    attachement " + i + " is a vcard", false, false);
                            var svcard = attachments[i];
                            var svcardposheader = svcard.indexOf("\n\n");
                            var svcardheader = svcard.substr(0, svcardposheader);
                            var svcard = svcard.substr(svcardposheader + 2, svcard.length - svcardposheader - 4); // remove the 2 last char which are "--"

                            if (svcardheader.search(/\s*Content-transfer-encoding\s*:\s*base64/i) != -1) {
                                // we have to decode from base64
                                var wWatcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
                                    .getService(Components.interfaces.nsIWindowWatcher);
                                if (wWatcher.activeWindow) {
                                    var win = wWatcher.activeWindow.QueryInterface(Components.interfaces.nsIDOMWindowInternal);
                                    svcard = win.atob(svcard.replace(/\n/g, ""));
                                }
                            }
                            var start = svcard.search(/begin:vcard/i);
                            var end = svcard.search(/end:vcard/i);
                            if (start != -1 && end != -1) {
                                vcard = obmad.escapedVCardToAbCard(escape(svcard.substr(start, end + 9 - start)));

                                var vcardcontent = svcard.substr(start + 11, end - start - 11);

                                // we won't use nsIAbCard::getProperty, TB seems to ignore the CALURI field during
                                // the parsing of a vcard, so let's try to retrieve by ourselves
                                var caluri = (new RegExp("\\nCALURI:\\s*([^\\s]*)", "gi")).exec(vcardcontent);
                                if (caluri) {
                                    obmCalendarURI = caluri[1];
                                    // store the CALURI into a custom1 field, because there is no
                                    // CALURI field support in TB3
                                    vcard.setProperty("Custom1", "X-OBM-CALURI:" + obmCalendarURI);
                                } else {
                                    this._logger.debug('CALURI not found', false, false);
                                }
                            } else {
                                this._logger.debug('begin/end vcard not found', false, false);
                            }
                        }
                    }
                }
                //else
                //  this._logger.debug(' no boundary',false,false);
            }
            //else
            //  this._logger.debug(' no multipart',false,false);
            if (this._callback)
                this._callback.onResponse(vcard, obmCalendarURI);
        }
    }
};

var components = [obmVCardHandler];
if ("generateNSGetFactory" in XPCOMUtils) {
    var NSGetFactory = XPCOMUtils.generateNSGetFactory(components);
} else {
    var NSGetModule = XPCOMUtils.generateNSGetModule(components);
}