/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/
var EXPORTED_SYMBOLS = ["StartShutdownObserver"];
var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js')
    // unixFileWriter
var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

// NOTE: This line loads the overlay for calItipUtils.jsm. Do not
// remove this line even though it is not used in this file.

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function StartShutdownObserver() {
    Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
    this.wrappedJSObject = this;
    this.initLogger();
    this.setupObservers();
    this.startSynchronize();
}

StartShutdownObserver.prototype = {
    // nsISupports interface
    QueryInterface: ChromeUtils.generateQI([Ci.nsIObserver, Ci.nsIConsoleListener, Ci.nsIMsgFolderListener]),
    classID: Components.ID("{16896900-4776-11dd-ae16-0800200c9a66}"),
    contractID: "@obm.org/start-observer;1",
    classDescription: "OBM Connector start observer",
    _xpcom_factory: {
        instance_: null,
        createInstance(outer, iid) {
            if (outer) {
                throw Cr.NS_ERROR_NO_AGGREGATION;
            }
            if (!this.instance_) {
                this.instance_ = new StartShutdownObserver();
            }
            return this.instance_.QueryInterface(iid);
        },
    },
    get QueryInterface() {
        return ChromeUtils.generateQI([Ci.nsIObserver, Ci.nsIConsoleListener, Ci.nsIMsgFolderListener]);
    },
    // {{{ START_COMPAT_TB_3
    _xpcom_categories: [{
        category: "app-startup",
        service: true
    }],
    // END_COMPAT_TB_3 }}}

    // nsIObserver/nsIConsoleListener interface
    observe: function(aSubject, aTopic, aData) {
        if (aTopic == "final-ui-startup") {
            this.startupObserver(aSubject, aTopic, aData);
        } else if (aTopic == "quit-application") {
            this.quitObserver(aSubject, aTopic, aData);
        } else if (aTopic == "profile-after-change") {
            try {
                Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
                var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);

                UtilsOBM.getOrigin.init();
            } catch (e) {
                Components.utils.reportError(e);
            }

            try {
                Components.utils.import("resource://gre/modules/mailServices.js");
                Components.utils.import("resource://gre/modules/iteratorUtils.jsm");
            } catch (e) {
                // Ignored, we're running in TB3 that doesn't have MailServices
            }

            // this.initLogger();
            // this.addObservers();
        } else if (aTopic == "app-startup") {
            // {{{ START_COMPAT_TB_3 This could be removed
            Cc["@mozilla.org/observer-service;1"]
                .getService(Ci.nsIObserverService)
                .addObserver(this, "profile-after-change", false);
            // END_COMPAT_TB_3 }}}
        } else if (aTopic) {
            this.logObserver(aTopic);
        }
    },

    initLogger: function initLogger() {
        this.logger = new ObmLogger();

        this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
        this.logger.setHeader("StartShutdownObserver: ");

        var consoleService = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
        consoleService.registerListener(this);
    },
    setupObservers: function() {
        Services.obs.addObserver(this, "profile-after-change");
        Services.obs.addObserver(this, "final-ui-startup");
        Services.obs.addObserver(this, "quit-application");
        this.logger.debug("StartShutdownObserver added listeners", false, false);
    },
    // addObservers: function() {
    //     if (!this.observerAdded) {
    //         var observerService = Cc["@mozilla.org/observer-service;1"]
    //             .getService(Ci.nsIObserverService);

    //         observerService.addObserver(this, "final-ui-startup", false);
    //         observerService.addObserver(this, "quit-application", false);
    //         // observerService.addObserver(this, "profile-after-change", false);
    //         this.observerAdded = true;

    //         //this.logger.debug("StartShutdownObserver added listeners", false, false);
    //     }
    // },

    startSynchronize: function initFirstSync() {
        let syncManager = UtilsOBM.getSyncManager();
        syncManager.syncWithDialog();
    },

    logObserver: function(aMessage) {
        this.logger.debug("observer triggered: " + aMessage.message, false, false);
    },

    startupObserver: function(aSubject, aTopic, aData) {
        if (typeof(MailServices) != 'undefined') {
            // Notify of new email, make sure our header is saved when receiving them
            const mfn = Components.interfaces.nsIMsgFolderNotificationService;
            MailServices.mfn.addListener(this, mfn.msgsClassified);
            let customHeaders = UtilsOBM.getPrefSafe("mailnews.customDBHeaders", "").split(" ");
            if (customHeaders.indexOf("x-obm-notification-email") < 0) {
                customHeaders.push("x-obm-notification-email");
                UtilsOBM.setPrefSafe("mailnews.customDBHeaders", customHeaders.join(" ").trim());
            }
        }

        // Migrate between Lightning versions
        this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
        this.timer.initWithCallback({
            notify: function(timer) {
                UtilsOBM.checkLightningUpgrade();
            }
        }, 1000, this.timer.TYPE_ONE_SHOT);

        // Synchronize on startup
        if (UtilsOBM.getPrefSafe("extensions.obm.firstsync", true)) {
            UtilsOBM.setPrefSafe("extensions.obm.firstsync", false);
        } else {
            let syncOnStart = UtilsOBM.getPrefSafe("extensions.obm.syncOnStart", true);
            let firstSyncDelay = UtilsOBM.getPrefSafe("extensions.obm.firstSyncDelay", 0);

            // If the first sync is delayed, then it will be taken care of by
            // the autorefresh timeout in the sync manager
            if (firstSyncDelay <= 0 && UtilsOBM.isOnline() && syncOnStart) {
                this.startSynchronize();
            }
        }
    },

    quitObserver: function(aSubject, aTopic, aData) {
        if (typeof(MailServices) != 'undefined') {
            MailServices.mfn.removeListener(this);
        }

        // TODO : reset pref for obm-sync alert support:
        //UtilsOBM.resetPref("extensions.obm.sync-supports-alerts");

        // Synchronize on quit
        var online = UtilsOBM.isOnline();
        var pref = UtilsOBM.getPrefSafe("extensions.obm.syncOnQuit", false);

        if (online && pref) {
            this.startSynchronize();
        }
    },

    msgAdded: function() {},
    msgsDeleted: function() {},
    msgsMoveCopyCompleted: function() {},
    msgKeyChanged: function() {},
    folderAdded: function() {},
    folderDeleted: function() {},
    folderMoveCopyCompleted: function() {},
    folderRenamed: function() {},
    itemEvent: function() {},

    msgsClassified: function(aMsgs, aJunkProcessed, aTraitsProcessed) {
        if (!UtilsOBM.getPrefSafe("extensions.obm.proactiveSync", false)) {
            return;
        }

        let foundHeader = false;
        for (let msg of fixIterator(aMsgs.enumerate(), Ci.nsIMsgDBHdr)) {
            if (msg.getProperty("x-obm-notification-email")) {
                foundHeader = true;
                break;
            }
        }

        if (foundHeader) {
            this.logger.notice("Received Email with X-OBM-NOTIFICATION-EMAIL, refreshing all calendars", false, false);

            // This is a background sync, just do it.
            Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
            let calSync = new ObmRCalendarSync();
            calSync.mergeAllCalendars({
                onStart: function() {},
                onNewOperation: function() {},
                onOperationDone: function() {},
                onProgress: function() {},
                onError: function() {},
                onStop: function() {}
            });
        }
    }
};

/****
 **** module registration
 ****/


this.NSGetFactory = XPCOMUtils.generateNSGetFactory([StartShutdownObserver]);