var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
//Components.utils.import("resource://calendar/modules/calItipUtils.jsm");
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");

var DEFAULT_ADD_LISTENERS_DELAY = 500;
var DEFAULT_DELETE_MESSAGE_DELAY = 1000;
var DEFAULT_DELETE_MESSAGE_PERMANENTLY = false;
var DEFAULT_DELETE_ENABLE = true;

var gObmDisplayedMessage = null;

const obmOnItipItem = {
    obmCalendarSyncInProgress: false,
    ltnImipObserve: function() {},
    observe: function ltnOnItipItem_observe(subject, topic, state) {
        if (topic != "onItipItemCreation" || !gMessageDisplay.displayedMessage) {
            return;
        }

        var itipItem = null;

        try {
            if (!subject) {
                var sinkProps = msgWindow.msgHeaderSink.properties;
                // This property was set by lightningTextCalendarConverter.js
                itipItem = sinkProps.getPropertyAsInterface("itipItem", Components.interfaces.calIItipItem);
            }
        } catch (e) {
            // This will throw on every message viewed that doesn't have the
            // itipItem property set on it. So we eat the errors and move on.
            // XXX TODO: Only swallow the errors we need to. Throw all others.
        }

        if (!itipItem) {
            return;
        }

        let isReady = UtilsOBM.isPrefSet("extensions.obm.server.uuid") && UtilsOBM.isPrefSet("extensions.obm.server.domain");

        if (!isReady) {
            showBarWithoutButtons(UtilsOBM.getLocalizedMessage("obm.calendar.not.ready"));
            return;
        }

        if (this.obmCalendarSyncInProgress) {
            showBarWithoutButtons(UtilsOBM.getLocalizedMessage("obm.synchronization.in.progress"));
        } else {
            this.ltnImipObserve(subject, topic, state);

            let enabled = UtilsOBM.getPrefSafe("extensions.obm.imip.delete.enable", DEFAULT_DELETE_ENABLE);

            if (enabled) {
                gObmDisplayedMessage = gMessageDisplay.displayedMessage;

                let delay = UtilsOBM.getPrefSafe("extensions.obm.imip.add.listeners.delay", DEFAULT_ADD_LISTENERS_DELAY);
                setTimeout(obmSetupImipButtons, delay);
            }
        }
    }
};

const obmCalendarSyncObserver = {
    observe: function obmCalendarSyncObserver_observe(subject, topic, state) {
        if (topic == "obmCalendarSyncStarted") {
            obmOnItipItem.obmCalendarSyncInProgress = true;
        } else if (topic == "obmCalendarSyncCompleted") {
            obmOnItipItem.obmCalendarSyncInProgress = false;
        }

        Components
            .classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService)
            .notifyObservers(null, "onItipItemCreation", 0);
    }
};

function showBarWithoutButtons(message) {
    let imipBar = document.getElementById("imip-bar");

    hideElement("imip-button1");
    hideElement("imip-button2");
    imipBar.setAttribute("label", message);
    imipBar.setAttribute("collapsed", "false");
}

function obmSetupImipButtons() {
    document.getElementById("imip-button1").onclick = onClickListener;
    document.getElementById("imip-button2").onclick = onClickListener;
}

function onClickListener() {
    if (gObmDisplayedMessage) {
        var msgHdr = gObmDisplayedMessage;
        var msgs = Components.classes["@mozilla.org/array;1"].createInstance(Components.interfaces.nsIMutableArray);
        let delay = UtilsOBM.getPrefSafe("extensions.obm.imip.delete.message.delay", DEFAULT_DELETE_MESSAGE_DELAY);
        let permanentDelete = UtilsOBM.getPrefSafe("extensions.obm.imip.delete.message.permanently", DEFAULT_DELETE_MESSAGE_PERMANENTLY);

        msgs.appendElement(msgHdr, false);

        setTimeout(function(aMsg) { msgHdr.folder.deleteMessages(aMsg, null, permanentDelete, false, null, false); }, delay, msgs);
    }

    gObmDisplayedMessage = null;
}

function obmImipOnLoad() {
    var obs = Components
        .classes["@mozilla.org/observer-service;1"]
        .getService(Components.interfaces.nsIObserverService);

    obs.addObserver(obmOnItipItem, "onItipItemCreation", false);
    obs.addObserver(obmCalendarSyncObserver, "obmCalendarSyncStarted", false);
    obs.addObserver(obmCalendarSyncObserver, "obmCalendarSyncCompleted", false);

    obmOnItipItem.ltnImipObserve = ltnImipBar.observe;
    ltnImipBar.observe = function() {};
}

function obmImipOnUnload() {
    var obs = Components
        .classes["@mozilla.org/observer-service;1"]
        .getService(Components.interfaces.nsIObserverService);

    removeEventListener("messagepane-loaded", obmImipOnLoad, true);
    removeEventListener("messagepane-unloaded", obmImipOnUnload, true);
    document.getElementById("imip-button1").removeEventListener("onclick", onClickListener, false);
    document.getElementById("imip-button2").removeEventListener("onclick", onClickListener, false);

    obs.removeObserver(obmOnItipItem, "onItipItemCreation");
    obs.removeObserver(obmCalendarSyncObserver, "obmCalendarSyncStarted");
    obs.removeObserver(obmCalendarSyncObserver, "obmCalendarSyncCompleted");

    gObmDisplayedMessage = null;
}

addEventListener("messagepane-loaded", obmImipOnLoad, true);
addEventListener("messagepane-unloaded", obmImipOnUnload, true);