/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/

Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");

var gOBMPrefs = {
    init: function() {
        window.addEventListener("unload", this, false);
        Services.obs.addObserver(this, "network:offline-status-changed", false);
        this.updateCommands();

        // let prefs = document.getElementById("prefs-button");
        // prefs.addEventListener("command", gOBMPrefs.loadConnectionWindow())
    },
    destroy: function() {
        Services.obs.removeObserver(this, "network:offline-status-changed");
    },

    updateCommands: function() {
        let offline = Services.io.offline;

        document.getElementById('subscription').hidden =
            (offline ||
                UtilsOBM.getPrefSafe("extensions.obm.singleCalendar", true) ||
                !UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true) ||
                !UtilsOBM.isLightningInstalledAndEnabled());

        document.getElementById('import-button').hidden =
            (offline || !UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true));

        document.getElementById('purge-button').hidden =
            (offline || !UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", true));

        document.getElementById('subscription-button').disabled = !UtilsOBM.getOBMCredentials({}, {});

        document.getElementById("extensions.obm.imip.delete.message.permanently").disabled = !document.getElementById("extensions.obm.imip.delete.enable").value;
    },

    loadSubscriptionWindow: function() {
        let paginationEnabled = UtilsOBM.getPrefSafe("extensions.obm.pagination.enabled", true) &&
            UtilsOBM.capabilitySupported("CALENDAR_HANDLER_SUPPORTS_PAGINATION");
        let windowURI = (paginationEnabled ? "preferences/paginatedSubscription.xul" : "preferences/subscription.xul");
        UtilsOBM.showWindow(windowURI, window);
    },

    loadConnectionWindow: function() {
        UtilsOBM.showWindow('preferences/connection.xul', window);
    },

    loadSyncWindow: function() {
        UtilsOBM.showWindow('preferences/synchronization.xul', window);
    },

    // Observer and Event Listener
    handleEvent: function(aEvent) {
        if (aEvent.type == "unload") {
            this.destroy();
        }
    },

    observe: function(aSubject, aTopic, aState) {
        if (aTopic == "network:offline-status-changed") {
            this.updateCommands();
        }
    },
    updatePref: function(pref) {
        let prefValue;
        pref.tagName == 'checkbox' ? prefValue = pref.checked : prefValue = pref.value;
        UtilsOBM.setPrefSafe(pref.getAttribute("preference"), prefValue)
    }

};
document.addEventListener("dialogaccept", function(event) {
    event.preventDefault()
    window.close()
});
document.addEventListener("dialogcancel", function(event) {
    event.preventDefault()
    window.close()
});

let prefs = [
    "notificationList",
    "deleteInvitationsWhenDone",
    "deleteInvitationsPermanently"
];
prefs.forEach(p => {
    let pref = document.getElementById(p);
    let prefValue = UtilsOBM.getPrefSafe(pref.getAttribute("preference"), false);
    if (pref.tagName == 'checkbox') {
        pref.checked = prefValue
        pref.onclick = function() {
            gOBMPrefs.updatePref(pref)
        }
    } else {
        pref.value = prefValue
        pref.onselect = function() {
            gOBMPrefs.updatePref(pref)
        }
    }
})