/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter

const HTML_NS = "http://www.w3.org/1999/xhtml";
const Cc = Components.classes;
const Ci = Components.interfaces;

var UO_logger = new ObmLogger()

UO_logger.setLevelWriter(UO_logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
UO_logger.setHeader("subscribe.js: ");

var gSearchInput;
var userServerError = false;
var resourceServerError = false;

function GenericWrappedXPComObject() {};
GenericWrappedXPComObject.prototype = {
    QueryInterface: ChromeUtils.generateQI([]),
    get wrappedJSObject() { return this; }
};

function displayResourceError(aMessage) {
    var loading = document.getElementById('loading-resource');
    loading.removeChild(loading.firstChild);
    var result = document.getElementById('loading-resource.label');
    result.setAttribute("value", aMessage);
}

function displayUserError(aMessage) {
    var loading = document.getElementById('loading-user');
    loading.removeChild(loading.firstChild);
    var result = document.getElementById('loading-user.label');
    result.setAttribute("value", aMessage);
}

function createTreeItem(calInfo) {
    let obmCalMgr = UtilsOBM.getCalendarManager();
    let subs = (obmCalMgr.getObmCalendar(calInfo.uid) != null);
    let treerow = document.createElement("treeitem").appendChild(document.createElement("treerow"));

    treerow.setAttribute("id", "_" + calInfo.uid);
    treerow.appendChild(document.createElement("treecell")).setAttribute("properties", (!calInfo.write ? "readOnly" : "readWrite"));
    treerow.appendChild(document.createElement("treecell")).setAttribute("label", calInfo.lastname + " " + calInfo.firstname);
    treerow.appendChild(document.createElement("treecell")).setAttribute("label", calInfo.mail);
    treerow.appendChild(document.createElement("treecell")).setAttribute("value", subs);

    return treerow.parentNode;
}

function createResourceTreeItem(resInfo) {
    var subs = false;
    // TODO. NEW_OBM_CALENDAR_MANAGER. use obmICalendarsManager::getObmCalendar() instead
    if (_resourceIsSubscribed(resInfo.mail)) {
        subs = true;
    }
    let treerow = document.createElement("treeitem").appendChild(document.createElement("treerow"));
    treerow.setAttribute("id", "_" + resInfo.id);
    treerow.setAttribute("properties", "_" + resInfo.mail);
    treerow.appendChild(document.createElement("treecell")).setAttribute("properties", "readOnly");
    treerow.appendChild(document.createElement("treecell")).setAttribute("label", resInfo.name);
    treerow.appendChild(document.createElement("treecell")).setAttribute("label", resInfo.description);
    treerow.appendChild(document.createElement("treecell")).setAttribute("value", subs);
    return treerow.parentNode;
}

function fillSubscriptions(aCalendarsArray) {
    var loading = document.getElementById('loading-user');
    var tree = document.getElementById('user-calendars-tree');
    loading.setAttribute("hidden", "true");
    tree.setAttribute("hidden", "false");

    var treeChildren = document.getElementById('user-calendars-treechildren');
    UtilsOBM.emptyDOMNode(treeChildren);
    var calendars = aCalendarsArray.QueryInterface(Ci.nsIArray);
    var count = calendars.length;

    var items = {},
        sortedKeys = [];

    for (var i = 0; i < count; i++) {
        var calInfo = calendars.queryElementAt(i, Ci.obmIRCalendarInfo);
        var hash = calInfo.lastname + " " + calInfo.firstname + " " + calInfo.uid;
        items[hash] = createTreeItem(calInfo);
        var inserted = false;
        for (var j in sortedKeys) {
            if (hash <= sortedKeys[j]) {
                sortedKeys.splice(j, null, hash);
                inserted = true;
                break;
            }
        }
        if (!inserted) sortedKeys.push(hash);
    }

    for (var key of sortedKeys) {
        treeChildren.appendChild(items[key]);
    }

}

function fillResourceSubscriptions(aResourcesArray) {
    var loading = document.getElementById('loading-resource');
    var tree = document.getElementById('resource-calendars-tree');
    loading.setAttribute("hidden", "true");
    tree.setAttribute("hidden", "false");

    var treeChildren = document.getElementById('resource-calendars-treechildren');
    UtilsOBM.emptyDOMNode(treeChildren);
    var allResources = aResourcesArray.QueryInterface(Ci.nsIArray);
    var resources = filterResourceByRights(allResources);
    var count = resources.length;

    UO_logger.debug(" count :" + count, false, false);

    var items = {},
        sortedKeys = [];

    for (var i = 0; i < count; i++) {
        var resInfo = resources[i];
        var hash = resInfo.name + " " + resInfo.id;
        items[hash] = createResourceTreeItem(resInfo);
        var inserted = false;
        for (var j in sortedKeys) {
            if (hash <= sortedKeys[j]) {
                sortedKeys.splice(j, null, hash);
                inserted = true;
                break;
            }
        }
        if (!inserted) sortedKeys.push(hash);
    }
    sortedKeys.forEach(function(key) { treeChildren.appendChild(items[key]) })
}

function filterResourceByRights(allResources) {
    var resources = [];
    var count = allResources.length;

    for (var i = 0; i < count; i++) {
        var resInfo = allResources.queryElementAt(i, Ci.obmIRResourceInfo);
        if (resInfo.read == true) {
            resources.push(resInfo);
        }
    }
    return resources;
}

function _resourceIsSubscribed(aResInfoMail) {
    var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
    var calendars = calMgr.getCalendars({});
    for (var calendar of calendars) {
        if (calendar.type == "ics" && calMgr.getCalendarPref_(calendar, "X-OBM-RESOURCE-EMAIL") == aResInfoMail) {
            return calendar;
        }
    }
    return null;
}

function _subscribeToResource(aId, aName, aEmail, aReadOnly) {
    var uid = aId;
    var name = "[" + UtilsOBM.getLocalizedMessage("obm.resource.label") + "] " + aName;
    var email = aEmail;
    var color = UtilsOBM.getRandomColor();
    var readonly = aReadOnly;
    //UO_logger.debug("_subscribeToResource: " + uid + ", " + name + ", " +  email + ", " + color + ", " + readonly + "\n", false, false);

    var calMgr = Cc["@mozilla.org/calendar/manager;1"].getService(Ci.calICalendarManager);
    UtilsOBM.setPrefSafe(
        "extensions.obm.changes.calICalendar." +
        uid +
        ".readOnly", readonly
    );

    var uri = UtilsOBM.createResourceUri(email);

    // Components.utils.import("chrome://obm-extension/content/components/calOBMCalendar.js")
    // var newCalendar = new calOBMCalendar();
    // newCalendar.uri = uri;

    var newCalendar = calMgr.createCalendar("ics", uri);
    calMgr.registerCalendar(newCalendar);
    newCalendar.name = name;
    newCalendar.setProperty("X-OBM-DEFAULT-NAME", name.toString());
    newCalendar.setProperty("X-OBM-RESOURCE-EMAIL", email.toString());
    newCalendar.setProperty("X-OBM-EMAIL", email.toString());
    newCalendar.setProperty("color", color.toString());

    UtilsOBM.setPrefSafe("extensions.obm.changes." +
        Ci.calICalendar + "." +
        uid + "." +
        "lastSync",
        "1970-01-01T00:00:00Z");

}

function saveSubscriptionSettings() {
    if (userServerError)
        return;

    var treeChildren = document.getElementById('user-calendars-treechildren');

    var checkedCalendars = {};
    for (var treeItem of Array.slice(treeChildren.childNodes)) {
        let treeRow = treeItem.firstChild;
        // subscribed ?
        var treecells = Array.slice(treeRow.childNodes);
        var subscribed = treecells[3].getAttribute("value") == "true" ? true : false;
        var calId = treeRow.id.substring(1);
        if (subscribed) {
            checkedCalendars[calId] = {
                readOnly: treecells[0].getAttribute("properties"),
                name: treecells[1].getAttribute("label"),
                email: treecells[2].getAttribute("label")
            };
        }
    }
    var calMgr = cal.getCalendarManager();
    var obmCalMgr = UtilsOBM.getCalendarManager();
    var currentlySubscribed = calMgr.getCalendars({});
    // unsubscribe from non-checked calendars
    for (var calendar of currentlySubscribed) {
        var uri = UtilsOBM.getCalendarOBMName(calendar);
        if (calendar.type == "obm" &&
            !(uri.split(";").pop() in checkedCalendars)) {
            // both have to be called
            calMgr.unregisterCalendar(calendar);
            calMgr.removeCalendar(calendar);
        }
    }

    // addd checked calendars
    for (let checkedCalendarUID in checkedCalendars) {
        let checkedCalendarInfos = checkedCalendars[checkedCalendarUID];
        let calendarName = checkedCalendarInfos.name;
        // e-mail of the calendar owner, used for attendee checking
        let email = checkedCalendarInfos.email;
        let color = UtilsOBM.getRandomColor();
        let readOnly = checkedCalendarInfos.readOnly == "readOnly";
        let calendar = obmCalMgr.getObmCalendar(checkedCalendarUID);
        if (calendar) {
            // update infos
            obmCalMgr.updateObmCalendar(calendar, checkedCalendarUID, calendarName, email, readOnly);
        } else {
            obmCalMgr.subscribeObmCalendar(checkedCalendarUID, calendarName, email, readOnly);
        }
    }
    saveResourceSubscriptionSettings();
}

function saveResourceSubscriptionSettings() {
    if (resourceServerError) return;

    // Build a map of currently subscribed resource calendars
    let subscribedResourceMap = {};
    for (let calendar of UtilsOBM.getSubscribedOBMResource()) {
        let uri = UtilsOBM.getResourceOBMMail(calendar);
        let mail = uri.split(";").pop();
        if (calendar.type == "ics") {
            subscribedResourceMap[mail] = calendar;
        }
    }

    // Go through the checked calendars, subscribe to any that are not
    // subscribed.
    let treeChildren = document.getElementById('resource-calendars-treechildren');
    for (let treeItem of Array.slice(treeChildren.childNodes)) {
        let treeRow = treeItem.firstChild;
        let treecells = Array.slice(treeRow.childNodes);
        let calId = treeRow.id.substring(1);
        let calMail = treeRow.getAttribute("properties").substring(1);
        let calReadOnly = (treecells[0].getAttribute("properties") == "readOnly");
        let calName = treecells[1].getAttribute("label");
        let subscribed = (treecells[3].getAttribute("value") == "true");

        if (subscribed) {
            if (calMail in subscribedResourceMap) {
                delete subscribedResourceMap[calMail];
            } else {
                let color = UtilsOBM.getRandomColor();
                _subscribeToResource(calId, calName, calMail, calReadOnly);
            }
        }
    }

    // Anything remaining in subscribedResourceMap needs to be deleted because
    // its subscribed but not checked.
    let calMgr = cal.getCalendarManager();
    for (let calendar in subscribedResourceMap) {
        calMgr.unregisterCalendar(subscribedResourceMap[calendar]);
        calMgr.removeCalendar(subscribedResourceMap[calendar]);
    }
}

function setupSearchInput() {
    if (!gSearchInput) {
        gSearchInput = document.getElementById("searchInputt");
    }
    document.getElementById("user-calendars-tree").focus();
}

function onSearchInputFocus(input) {
    input.select();
}

/*
var gQuickSearchFocusEl = null;
*/
function onClearSearch() {
    if (gSearchInput && !gSearchInput.showingSearchCriteria) // ignore the text box value if it's just showing the search criteria string
    {
        gSearchInput.value = "";
        updateList(gSearchInput);
        // this needs to be on a timer otherwise we end up messing up the focus while the Search("") is still happening
        setTimeout("restoreSearchFocusAfterClear();", 0);
    }
}

function restoreSearchFocusAfterClear() {
    gSearchInput.clearButtonHidden = 'true';
    //gQuickSearchFocusEl = null;
}

function updateList(input) {
    var treeItems = document.getElementsByTagName("treeitem");

    for (var i = 0; i < treeItems.length; i++) {
        var treeItem = treeItems.item(i);
        var treeCells = treeItem.getElementsByTagName("treecell");
        var cell1 = treeCells.item(1).getAttribute("label").toLowerCase(); // second cell
        var cell2 = treeCells.item(2).getAttribute("label").toLowerCase(); // third cell
        var search = input.value.toLowerCase().trim();
        var searchMatches = search == "" || (cell1.indexOf(search) > -1 || cell2.indexOf(search) > -1);

        treeItem.setAttribute("hidden", !searchMatches);
    }
}

function setupResourceSearchInput() {
    UO_logger.debug("setupResourceSearchInput()", false, false);
    gResourceSearchInput = document.getElementById("resourceSearchInputt");
    document.getElementById("resource-calendars-tree").focus();
}

function onClearResourceSearch() {
    if (gResourceSearchInput && !gResourceSearchInput.showingSearchCriteria) {
        gResourceSearchInput.value = "";
        updateList(gResourceSearchInput);
        setTimeout("restoreSearchFocusAfterClear();", 0);
    }
}

function restoreSearchFocusAfterClear() {
    gResourceSearchInput.clearButtonHidden = 'true';
}

function SyncObserver() {}
SyncObserver.prototype = {
    observe: function(subject, topic, data) {
        switch (data) {
            case "newOperation":
                break;
            case "operationDone":
                this.lastData = subject.wrappedJSObject.opResData;
                break;
            case "newOperationSet":
                break;
            case "operationSetError":
                var msg = subject.wrappedJSObject.message;
                displayUserError(msg);
                userServerError = true;
                this.unregister();
                break;
            case "operationSetDone":
                break;
            case "allDone":
                ResourcesIsSupported();
                fillSubscriptions(this.lastData);
                this.unregister();
                break;
            default:
        }
    },
    register: function() {
        var observerService = Components.classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService);
        observerService.addObserver(this, "subscribe-obm-sync", false);
    },
    unregister: function() {
        var observerService = Components.classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService);
        observerService.removeObserver(this, "subscribe-obm-sync");
    }
}

function ResourcesIsSupported() {
    let currentVersion = UtilsOBM.getPrefSafe("extensions.obm.current.sync.version");
    let minVersion = UtilsOBM.getPrefSafe("extensions.obm.minversion.sync.resource");
    var resourceIsSupported = UtilsOBM.compareVersions(currentVersion, minVersion) >= 0 ? true : false;
    if (resourceIsSupported == true) {
        listAllOBMResources();
        var resourceTab = document.getElementById("resourceintabs");
        resourceTab.setAttribute("hidden", "false");
    }
}

function ResourcesSyncObserver() {}
ResourcesSyncObserver.prototype = {
    observe: function(subject, topic, data) {
        switch (data) {
            case "newOperation":
                break;
            case "operationDone":
                this.lastData = subject.wrappedJSObject.opResData;
                break;
            case "newOperationSet":
                break;
            case "operationSetError":
                var msg = subject.wrappedJSObject.message;
                displayResourceError(msg);
                resourceServerError = true;
                this.unregister();
                break;
            case "operationSetDone":
                break;
            case "allDone":
                fillResourceSubscriptions(this.lastData);
                this.unregister();
                break
            default:
        }
    },
    register: function() {
        var observerService = Components.classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService);
        observerService.addObserver(this, "subscribe-obm-sync-resources", false);
    },
    unregister: function() {
        var observerService = Components.classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService);
        observerService.removeObserver(this, "subscribe-obm-sync-resources");
    }
}

function listAllOBMCalendars() {
    var ObserverServ = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);

    Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
    let calSync = new ObmRCalendarSync();
    var opIdx = 0;
    var operations = [];

    operations[0] = ["Calendar", function(arg) { calSync.getSubscribleCalendars(arg, -1, -1, '') }];

    var listener = {
        onStart: function() {},

        onNewOperation: function(aOpName) {
            var msg = new GenericWrappedXPComObject();
            msg.opName = aOpName;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "newOperation");
        },
        onOperationDone: function(aSuccess, aResultData) {
            var msg = new GenericWrappedXPComObject();
            msg.opRes = aSuccess;
            msg.opResData = aResultData;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "operationDone");
        },
        onProgress: function(aIdx, aTotal, aType) {},
        onError: function(aMessage) {
            dump("EXCEPTION: " + aMessage + "\n");
            var msg = new GenericWrappedXPComObject();
            msg.message = aMessage;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "operationSetError");
        },
        onStop: function(aRes, aMessage) {
            var msg = new GenericWrappedXPComObject();
            msg.message = aMessage;
            if (aRes == Components.results.NS_OK) {
                ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "operationSetDone");
                this.nextOperation();
            } else {
                ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "operationSetError");
            }
        },
        nextOperation: function() {
            if (opIdx >= operations.length) {
                ObserverServ.notifyObservers(null, "subscribe-obm-sync", "allDone");
                return;
            }

            var msg = new GenericWrappedXPComObject();
            msg.opSetName = operations[opIdx][0];
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync", "newOperationSet");
            try {
                operations[opIdx++][1](this);
            } catch (e) {
                listener.onError(e);
            }
        }
    };
    listener.nextOperation();
}

function listAllOBMResources() {
    var ObserverServ = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);

    Components.utils.import("chrome://obm-extension/content/components/obm-sync-rest/obmRCalendarSync.js")
    let calSync = new ObmRCalendarSync();

    var opIdx = 0;
    var operations = [];

    operations[0] = ["Resource", function(arg) { calSync.getSubscribleResources(arg, -1, -1, '') }];

    var listener = {
        onStart: function() {},

        onNewOperation: function(aOpName) {
            var msg = new GenericWrappedXPComObject();
            msg.opName = aOpName;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "newOperation");
        },
        onOperationDone: function(aSuccess, aResultData) {
            UO_logger.debug("onOperationDone", false, false);
            var msg = new GenericWrappedXPComObject();
            msg.opRes = aSuccess;
            msg.opResData = aResultData;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "operationDone");
        },
        onProgress: function(aIdx, aTotal, aType) {},
        onError: function(aMessage) {
            dump("EXCEPTION: " + aMessage + "\n");
            var msg = new GenericWrappedXPComObject();
            msg.message = aMessage;
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "operationSetError");
        },
        onStop: function(aRes, aMessage) {
            var msg = new GenericWrappedXPComObject();
            msg.message = aMessage;
            if (aRes == Components.results.NS_OK) {
                ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "operationSetDone");
                this.nextOperation();
            } else {
                ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "operationSetError");
            }
        },
        nextOperation: function() {
            if (opIdx >= operations.length) {
                ObserverServ.notifyObservers(null, "subscribe-obm-sync-resources", "allDone");
                return;
            }

            var msg = new GenericWrappedXPComObject();
            msg.opSetName = operations[opIdx][0];
            ObserverServ.notifyObservers(msg, "subscribe-obm-sync-resources", "newOperationSet");
            try {
                operations[opIdx++][1](this);
            } catch (e) {
                listener.onError(e);
            }
        }
    };
    listener.nextOperation();
}

function onLoad() {
    var obs = new SyncObserver();
    var resobs = new ResourcesSyncObserver();
    obs.register();
    resobs.register();
    setupSearchInput();
    setupResourceSearchInput();
    setTimeout(listAllOBMCalendars, 100);
}

window.addEventListener("load", onLoad, false);

document.addEventListener("dialogaccept", function(event) {
    event.preventDefault(); // Prevent the dialog closing.
    saveSubscriptionSettings();
    window.close();
});