/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/

Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");
Components.utils.import("resource://obm-extension/modules/GlobalData.jsm");
var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter

/* Base Synchronizer */

function Base_Synchronizer() {
    //Init logger
    this.logger = new ObmLogger()

    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("Base_Synchronizer: ");
    this.logger.debug("Base_Synchronizer()", false, false);
}

Base_Synchronizer.prototype = {

    get mSynchronizing() {
        return OBMGlobalData.addrSynchronizing; // this value is set by ObmSyncManager
    },

    shutdown: function() {
        this.mProxy.unsubscribreToObserverService();
    },
    notifyItemAdded: function(aContainer, aItem) {
        this.logger.debug("notifyItemAdded", false, false);
        if (this.mSynchronizing) {
            this.logger.debug("mSynchronizing", false, false);
            return;
        }

        aContainer = aContainer.QueryInterface(this.mSupportedContainerType);

        if (!this.isContainerManagedByOBM(aContainer)) {
            this.logger.debug("! isContainerManagedByOBM", false, false);
            return;
        }

        var containerName = this.getContainerOBMName(aContainer);

        var itemID = this.getItemID(aItem);
        var newId = false;
        if (!itemID) {
            newId = true;
            // assign an ID to the new created item if necessary
            // (i.e. address book contact)
            itemID = "" + Math.round(Math.random() * 1000) + "-" + Math.round(Math.random() * 1000);

            this.logger.debug("itemID:" + itemID, false, false);
        }

        UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "added", itemID);
        if (newId) {
            this.setItemID(aContainer, aItem, itemID);
            //in tb3 will raise onPropertyChanged of addressbook listenner
            //so call after pushing in added list
        }
    },

    notifyItemModified: function(aContainer, aItem) {
        this.logger.debug("notifyItemModified", false, false);
        if (this.mSynchronizing) {
            this.logger.debug("mSynchronizing", false, false);
            return;
        }

        aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
        aItem = aItem.QueryInterface(this.mSupportedItemType);

        if (!this.isContainerManagedByOBM(aContainer)) {
            return;
        }

        var containerName = this.getContainerOBMName(aContainer);

        var itemID = this.getItemID(aItem);
        this.logger.debug("itemID:" + itemID, false, false);

        UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "updated", itemID);
    },

    notifyItemParticipationChanged: function(aContainer, aItem) {
        this.logger.debug("notifyItemParticipationChanged", false, false);
        if (this.mSynchronizing) {
            this.logger.debug("mSynchronizing", false, false);
            return;
        }

        aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
        aItem = aItem.QueryInterface(this.mSupportedItemType);

        if (!this.isContainerManagedByOBM(aContainer)) {
            return;
        }

        var containerName = this.getContainerOBMName(aContainer);

        var itemID = this.getItemID(aItem);
        this.logger.debug("itemID:" + itemID, false, false);

        UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "participationChanged", itemID);
    },

    notifyItemDeleted: function(aContainer, aItem) {
        this.logger.debug("notifyItemDeleted", false, false);
        if (this.mSynchronizing) {
            this.logger.debug("mSynchronizing", false, false);
            return;
        }

        aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
        aItem = aItem.QueryInterface(this.mSupportedItemType);

        if (!this.isContainerManagedByOBM(aContainer)) {
            return;
        }

        var containerName = this.getContainerOBMName(aContainer);

        var itemID = this.getItemID(aItem);
        if (!itemID) {
            return;
        }

        UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "removed", itemID);
    },

    notifyContainerDeleted: function(aContainer) {
        this.logger.debug("notifyContainerDeleted", false, false);
        if (this.mSynchronizing) {
            this.logger.debug("mSynchronizing", false, false);
            return;
        }

        aContainer = aContainer.QueryInterface(this.mSupportedContainerType);

        if (!this.isContainerManagedByOBM(aContainer)) {
            this.logger.debug("! isContainerManagedByOBM", false, false);
            return;
        }

        var containerName = this.getContainerOBMName(aContainer);
        UtilsOBM.deletePreferenceBranch("extensions.obm.changes." + this.mSupportedContainerType.toString() + "." + this.getContainerOBMName(aContainer));

        if (this.mSupportedContainerType.toString() == "calICalendar") {
            UtilsOBM.removeAllCalendarSyncItems(aContainer);
        } else {
            // don't fill a property for removed calendars
            UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), "list", this.mSupportedContainerType.toString(), "removed", this.getContainerOBMName(aContainer));
        }


        // Delete addressbook from addressbook list
        // TODO. OBM_NEW_AD_MANAGER. properly define API instead...
        // 		var name = this.getContainerOBMName(aContainer);
        let namesPref = UtilsOBM.getPrefSafe("extensions.obm.addressbooks", null);
        var names = namesPref ? namesPref.split(",") : [];
        var newNames = [];
        for (var i = 0; i < names.length; i++) {
            if (names[i] != containerName) {
                newNames.push(names[i]);
            }
        }
        UtilsOBM.setPrefSafe("extensions.obm.addressbooks", newNames.join(","));
        this.logger.debug("Update extensions.obm.addressbooks to " + newNames.join(","), false, false);

    },

    /* PRIVATE FUNCTIONS */
    isContainerManagedByOBM: function(aContainer) {
        return this.getContainerOBMName(aContainer) != null;
    },

    //FIXME: unused?
    _removeItemFromChangesList: function(aContainer, aChangeType, aItemID) {
        UtilsOBM.resetAllItemsChangesLists(this.mSupportedContainerType.toString(), this.getContainerOBMName(aContainer), this.mSupportedItemType.toString(), aChangeType, aItemID);
    }
};