/******************************************************************************
 Copyright (C) 2011-2012 Linagora

 This program is free software: you can redistribute it and/or modify it under
 the terms of the GNU Affero General Public License as published by the Free
 Software Foundation, either version 3 of the License, or (at your option) any
 later version, provided you comply with the Additional Terms applicable for OBM
 connector by Linagora pursuant to Section 7 of the GNU Affero General Public
 License, subsections (b), (c), and (e), pursuant to which you must notably (i)
 retain the “Message sent thanks to OBM, Free Communication by Linagora”
 signature notice appended to any and all outbound messages (notably e-mail and
 meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
 well as between Linagora and linagora.com, and (iii) refrain from infringing
 Linagora intellectual property rights over its trademarks and commercial
 brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
 for more details.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License and
 its applicable Additional Terms for OBM along with this program. If not, see
 <http://www.gnu.org/licenses/> for the GNU Affero General   Public License
 version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
 applicable to OBM connectors.
 ******************************************************************************/
var { ObmLogger } = ChromeUtils.import('chrome://obm-extension/content/components/logger/logger.js');
// unixFileWriter
var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm", null);
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");

var CalendarConflictResolver = function(calendar, calSync, serverPriority) {
    if (!calendar) {
        throw new Components.Exception("Calendar not set", Components.results.NS_ERROR_XPC_NEED_OUT_OBJECT);
    }
    if (!calSync) {
        throw new Components.Exception("OBMRCalendarSync (calSync) not set", Components.results.NS_ERROR_XPC_NEED_OUT_OBJECT);
    }
    this.calendar = calendar;
    this.calSync = calSync;
    this.serverPriority = serverPriority || false;
    this.logger = new ObmLogger()

    this.logger.setLevelWriter(this.logger.LOAD_LEVEL); //, UtilsOBM.getWriter());
    this.logger.setHeader("CalendarConflictResolver: ");
    this.logger.debug("new CalendarConflictResolver()", false, false);
};


CalendarConflictResolver.prototype = {
    calendar: null,
    serverPriority: false,
    logger: null,
    _getConflict: function(conflicts, id) {
        for (var i in conflicts) {
            if (conflicts[i].id == id) {
                return conflicts[i];
            }
        }
        return false;
    },
    _getRemoteChange: function(remoteDiff, type, id) {
        //this.logger.debug("getRemotechange:begins",false,false);
        if (type == "updated") {
            for (var i in remoteDiff.updated) {
                if (remoteDiff.updated[i].extId == id) {
                    return remoteDiff.updated[i];
                }
            }
        } else if (type == "removed") {
            for (var i in remoteDiff.removed) {
                if (remoteDiff.removed[i] == id) {
                    return id;
                }
            }
        } else if (type == "participationChanged") {
            for (var i in remoteDiff.participationChanged) {
                if (remoteDiff.participationChanged[i].extId == id) {
                    return remoteDiff.participationChanged[i];
                }
            }
        }
        return null;
    },
    // return a ical string in UTC
    _getInUtc: function(str) {
        if (str[(str.length - 1)] == "Z") {
            return str;
        }
        this.logger.debug("creating dateTime " + str, false, false);
        try {
            var dt = cal.createDateTime(str);
        } catch (e) {
            this.logger.debug("creating dateTime FAILED" + str, false, false);
        }
        this.logger.debug("dateTime created" + str, false, false);
        dt.timezone = cal.dtz.defaultTimezone;
        dt = dt.getInTimezone(cal.dtz.UTC);
        this.logger.debug("returning dateTime" + str, false, false);
        return dt.icalString;
    },
    _getAllParticipationChanges: function(diff) {
        var back = [];
        if (diff.recurrenceChange) {
            back.push({
                extId: diff.extId,
                action: "participationChanged",
                sequence: diff.sequence,
                notification: diff.notification
            });
        }
        for (var recurrenceId of diff.recurrenceIds) {
            back.push({
                extId: diff.extId,
                action: "participationChanged",
                sequence: diff.sequence,
                notification: diff.notification,
                recurrenceId: recurrenceId
            });
        }
        return back;
    },
    _getEventFromCalendar: function(remoteEventId, callback) {
        var listener = { localEvent: null };
        listener.onOperationComplete = function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            //<TO_REMOVE>
            // Fix second onOperationComplete call (bug LT 0.8)
            // Fixed in LT 0.9
            try {
                if (aDetail.toString() == "Can't deduce item type based on QI") {
                    dump("_calGetItem, aDetail = Can't deduce item type based on QI\n");
                    return;
                }
            } catch (e) {
                dump("_calGetItem, aDetail: not an object?\n");
                dump(e);
            }
            // FIXME : bug lightning ? aStatus for getItem is allways 0
            if (!Components.isSuccessCode(aStatus) || !this.localEvent) {
                // nothing to do ?
                callback(null);
            } else {
                var localEvent;
                try {
                    localEvent = aDetail.QueryInterface(Components.interfaces.calIEvent)
                } catch (e) {
                    localEvent = this.localEvent;
                }

                callback(localEvent);
            }
        };
        listener.onGetResult = function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };
        this.calendar.getItem(remoteEventId, listener);
    },
    _mergeParticipationChange: function(remoteEvt, conflict, callback) {
        var self = this;
        this._getEventFromCalendar(remoteEvt.extId, function(localEvent) {
            if (!localEvent) {
                self.logger.debug("_resolveConflicts: LocalEvent not found", false, false);
                callback(false);
            }
            var localOccurrence = remoteEvt.recurrenceId ? localEvent.getOccurrenceFor(remoteEvt.recurrenceId) : localEvent;
            //get my status on local storage
            var ownerEmail = cal.getCalendarManager().getCalendarPref_(localOccurrence.calendar, "X-OBM-EMAIL"),
                ownerAttendeeLocal = localOccurrence.getAttendeeById("MAILTO:" + ownerEmail),
                ownerAttendeeRemote = remoteEvt.getAttendeeByEmail(ownerEmail);
            // apply my local participationState to the remote one
            if (ownerAttendeeLocal && ownerAttendeeRemote) {
                ownerAttendeeRemote.status = ownerAttendeeLocal.participationStatus;
            }
            callback(true);
        });
    },
    /**
     * foreach of our local participation change, we apply it on the remote evt if necessary, and sent it to the server
     */
    _mergeParticipationChangeOnAnUpdatedEvent: function(remoteEvt, conflict, localChanges, callback) {
        var self = this,
            applyRemote = false;
        this._getEventFromCalendar(remoteEvt.extId, function(localEvent) {
            if (!localEvent) {
                self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: LocalEvent not found", false, false);
                callback(false);
            }
            self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent localChanges: " + JSON.stringify(localChanges), false, false);
            var response = [],
                remoteEventExceptions = remoteEvt.recurrence.getEventExceptions({}),
                remoteEventExceptionsIndex = [],
                updateParticipationAndModifyRemoteEvt = function(localEvt, remoteEvt, localChange, ownerEmail) {
                    ownerAttendeeLocal = localEvt.getAttendeeById("MAILTO:" + ownerEmail),
                        ownerAttendeeRemote = remoteEvt.getAttendeeByEmail(ownerEmail);
                    // local user not found in local event...
                    if (!ownerAttendeeLocal) {
                        return;
                    }
                    if (ownerAttendeeRemote) {
                        if (ownerAttendeeRemote.status != ownerAttendeeLocal.participationStatus) {
                            ownerAttendeeRemote.status = ownerAttendeeLocal.participationStatus;
                            response.push(localChange);
                            applyRemote = true;
                            self.logger.debug("updateParticipationAndModifyRemoteEvt updating remote & local", false, false);
                        }
                    } else {
                        // not an attendee in remote evt and attendee in local... hope that won't happen
                        return;
                    }
                };

            for (var e of remoteEventExceptions) {
                remoteEventExceptionsIndex.push(self._getInUtc(cal.dtz.jsDateToDateTime(UtilsOBM.restDateToJSDate(e.recurrenceId)).icalString));
            }
            self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: remoteEventExceptionsIndex: " + JSON.stringify(remoteEventExceptionsIndex), false, false);
            for (var localChange of localChanges) {
                self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent localChange: " + JSON.stringify(localChange), false, false);
                var ownerEmail = cal.getCalendarManager().getCalendarPref_(localEvent.calendar, "X-OBM-EMAIL");
                if (!localChange.recurrenceId) {
                    var localSequence = UtilsOBM.getSequence(localEvent);
                    var remoteSequence = remoteEvt.generation;
                    var sequenceIsOutdated = (localSequence < remoteSequence);
                    //if local event is a simple event I check the sequence (OBMFULL-3773)
                    if (!localEvent.recurrenceInfo && sequenceIsOutdated) {
                        self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: working on simple Event, sequence is outdated ", false, false);
                        applyRemote = true;
                    } else {
                        updateParticipationAndModifyRemoteEvt(localEvent, remoteEvt, localChange, ownerEmail);
                    }
                } else {
                    self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: working on recurrence " + localChange.recurrenceId, false, false);
                    var localRecurrenceIdString = self._getInUtc(localChange.recurrenceId),
                        localRecurrenceIdDatetime = cal.createDateTime(localRecurrenceIdString);
                    self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: localChange recurrenceId " + localRecurrenceIdString, false, false);
                    var localException = localEvent.recurrenceInfo.getExceptionFor(localRecurrenceIdDatetime, false);
                    self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: localException " + localException, false, false);
                    var remoteExceptionIndex = remoteEventExceptionsIndex.indexOf(localRecurrenceIdString);
                    self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: remoteExceptionIndex " + remoteExceptionIndex, false, false);
                    var remoteException = (remoteExceptionIndex > -1) ? remoteEventExceptions[remoteExceptionIndex] : null;
                    self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: remoteException " + remoteException, false, false);
                    if (!localException) {
                        //no problem, doesn't exist locally
                        self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: localException not found", false, false);
                        continue;
                    } else if (!remoteException) {
                        // just updated an occurrence locally, thus creating an exception. Should send that to the server
                        self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: remote exception not found, pushing localChange", false, false);
                        response.push(localChange);
                    } else {
                        self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: full diff of localException and remoteException", false, false);
                        updateParticipationAndModifyRemoteEvt(localException, remoteException, localChange, ownerEmail);
                    }
                }
                self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent: End of " + localChange.recurrenceId, false, false);
            }
            self.logger.debug("_mergeParticipationChangeOnAnUpdatedEvent response: " + JSON.stringify(response), false, false);
            callback(true, response, applyRemote);
        });
    },
    _updateParticipation: function(participation, callback) {
        this.logger.debug("_updateParticipation:()", false, false);
        this.logger.debug("_updateParticipation: id = " + participation.id, false, false);
        this.logger.debug("_updateParticipation: extId = " + participation.extId, false, false);
        this.logger.debug("_updateParticipation: recurrenceId = " + participation.recurrenceId, false, false);
        var self = this,
            newLocalEvent = null;
        this._getEventFromCalendar(participation.extId, function(localEvent) {
            if (!localEvent) {
                // this case should not appear
                self.logger.debug("_updateParticipation: localEvent not found " + participation.id, false, false);
            } else {
                if (!participation.recurrenceId) {
                    self.logger.debug("_updateParticipation: localEvent.id= " + localEvent.id, false, false);
                    newLocalEvent = localEvent.clone();
                    self._updateParticipationInOccurrence(participation, newLocalEvent, localEvent);
                } else {
                    var recurrenceInfo = localEvent.recurrenceInfo;
                    if (recurrenceInfo) {
                        var dt = cal.createDateTime(participation.recurrenceId);
                        var exception = recurrenceInfo.getExceptionFor(dt, true);
                        if (exception) {
                            var newException = exception.clone(),
                                newLocalEvent = localEvent.clone();
                            self._updateParticipationInOccurrence(participation, newException, exception);
                            newLocalEvent.recurrenceInfo.modifyException(newException, false);
                        } else {
                            self.logger.debug("unable to find exception for occurrence recurrenceId " + participation.recurrenceId, false, false);
                        }
                    }
                }
            }
            callback(newLocalEvent);
        });
    },
    _updateParticipationInOccurrence: function(participation, occurrence, occurrenceRO) {
        for (var attendee of participation.getAttendees({})) {
            // Do we really need to guard against the server sending a wrong participation status?
            this.logger.debug("_updateParticipation attendee.email:" + attendee.email, false, false);
            var newParticipationStatus = this._getAcceptableParticipationState(attendee.status);
            var localAttendee = occurrenceRO.getAttendeeById("MAILTO:" + attendee.email);
            if (localAttendee == null) {
                this.logger.debug("_updateParticipation: localAttendee " + attendee.email + " not found", false, false);
                continue;
            }
            // Can't modifiy localAttendee because it is immutable
            var newLocalAttendee = localAttendee.clone();
            this.logger.debug("_updateParticipation actual participationStatus:" + localAttendee.participationStatus, false, false);
            newLocalAttendee.participationStatus = newParticipationStatus;
            this.logger.debug("_updateParticipation new status:" + newLocalAttendee.participationStatus, false, false);
            if (localAttendee.participationStatus != newLocalAttendee.participationStatus) {
                // Can't modify directly localEvent because it is immutable too...
                occurrence.removeAttendee(localAttendee);
                occurrence.addAttendee(newLocalAttendee);
                this.logger.debug("_updateParticipation: modification done.", false, false);
            }
        }
        return occurrence;
    },
    _getAcceptableParticipationState: function(participationState) {
        var acceptableParticipationStates = ["ACCEPTED", "DECLINED", "NEEDS-ACTION"];
        var defaultParticipationState = "NEEDS-ACTION";
        var newParticipationState = defaultParticipationState;

        for (var state of acceptableParticipationStates) {
            if (state == participationState) {
                newParticipationState = state;
                break;
            }
        }
        return newParticipationState;
    },
    _isInUpdated: function(remoteDiff, extId) {
        for (var i in remoteDiff.updated) {
            if (remoteDiff.updated[i].extId == extId) {
                return true;
            }
        }
        return false;
    },
    _findConflicts: function(localDiff, remoteDiff, localToSend, remoteToApply) {
        // find conflicts
        var conflicts = [];
        for (var i in remoteDiff.updated) {
            var id = remoteDiff.updated[i].extId;
            this.logger.debug("remoteDiff.updated: " + remoteDiff.updated[i].extId, false, false);
            if (localDiff.event[id] && localDiff.event[id].action) {
                conflicts.push({ id: id, local: localDiff.event[id].action, remote: "updated" });
            } else {
                remoteToApply.updated.push(remoteDiff.updated[i]);
            }
        }
        for (var i in remoteDiff.participationChanged) {
            var remoteData = remoteDiff.participationChanged[i],
                id = remoteDiff.participationChanged[i].extId;
            // OBM-SYNC send occurrences participationChanges both in updated and in participationChanged XML tags.
            // We resolve those through _resolveConflictUpdated()
            if (this._isInUpdated(remoteDiff, id)) {
                continue;
            }
            this.logger.debug("remoteDiff.participationChanged: " + remoteDiff.participationChanged[i].extId, false, false);
            if (localDiff.event[id] && localDiff.event[id].action) {
                if (localDiff.event[id].action == "participationChanged") {
                    if (!remoteData.recurrenceId && localDiff.event[id].recurrenceChange ||
                        localDiff.event[id].recurrenceIds &&
                        localDiff.event[id].recurrenceIds.indexOf(remoteData.recurrenceId) > -1) {
                        conflicts.push({ id: id, local: localDiff.event[id].action, remote: "participationChanged", recurrenceId: remoteData.recurrenceId });
                    }
                } else {
                    conflicts.push({ id: id, local: localDiff.event[id].action, remote: "participationChanged" });
                }
            } else {
                remoteToApply.participationChanged.push(remoteDiff.participationChanged[i]);
            }
        }
        for (var i in remoteDiff.removed) {
            var id = remoteDiff.removed[i];
            this.logger.debug("remoteDiff.removed: " + id, false, false);
            if (localDiff.event[id] && localDiff.event[id].action) {
                conflicts.push({ id: id, local: localDiff.event[id].action, remote: "removed" });
            } else {
                remoteToApply.removed.push(remoteDiff.removed[i]);
            }
        }

        // alter localDiff
        for (var i in localDiff.event) {
            // add the extid inside the struct, bc we loose the key on occurrence participation changes
            localDiff.event[i].extId = i;

            if (!this._getConflict(conflicts, i)) {
                // if there is no conflict on this item, it's clear we should send the change to our server

                // but if it's a participationChange event this can happen on several exceptions
                // and so generate several update queries
                if (localDiff.event[i].action == "participationChanged") {
                    var pChanges = this._getAllParticipationChanges(localDiff.event[i]);
                    this.logger.debug("participationChanges expanded from: " + JSON.stringify(localDiff.event[i]), false, false);
                    this.logger.debug("participationChanges expanded to: " + JSON.stringify(pChanges), false, false);
                    for (var pChange of pChanges) {
                        localToSend[i + "-" + pChange.recurrenceId] = pChange;
                    }
                } else {
                    localToSend[i] = localDiff.event[i];
                }
            }
        }
        return conflicts;
    },
    _resolveConflictRemoved: function(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved) {
        let localEvent = localDiff.event[conflict.id];

        if (conflict.remote == "removed") {
            // event flagged as 'removed from local DB' AND "remove from remote DB"
            // So : nothing to do.
            remoteToApply.removed.push(conflict.id);
        } else if (conflict.remote == "updated") {
            // local event is flagged as deleted, remote event as updated
            // check sequence
            var localSequence = localEvent.generation || 0;
            var remoteEvt = this._getRemoteChange(remoteDiff, conflict.remote, conflict.id);
            var remoteSequence = remoteEvt.generation;
            if (remoteSequence > localSequence) {
                remoteToApply.updated.push(remoteEvt);
            } else if (remoteSequence < localSequence || this._hashEqual(localEvent, remoteEvt)) {
                localToSend[conflict.id] = localEvent;
            } else {
                // sequence are the same : fallback to synchro preferences
                if (this.serverPriority) {
                    remoteToApply.updated.push(remoteEvt);
                } else {
                    localToSend[conflict.id] = localEvent;
                }
            }
        } else if (conflict.remote == "participationChanged") {
            // ignore participation change as we remove the event
            localToSend[conflict.id] = localEvent;
        }
        onConflictResolved();
    },

    _hashEqual: function(localEvent, remoteEvent) {
        let localHash = UtilsOBM.getHash(localEvent),
            remoteHash = UtilsOBM.getHash(remoteEvent),
            equal = remoteHash != null && (localHash === remoteHash);

        this.logger.debug("_hashEqual(): Local Hash is " + localHash + ", remote Hash is " + remoteHash + ". Returning " + equal + ".");

        return equal;
    },

    _resolveConflictParticipationChanged: function(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved) {
        this.logger.debug("remote state: " + conflict.remote, false, false);
        if (conflict.remote == "participationChanged") {
            // Here we face a protocol problem:
            // our local "participationChanged" means we changed our _own_ participation status
            // the remote "participationChange" means someone in the attendees updated the participation status
            // no way to know who it is, and it can be more than 1 person
            // So we apply remote participation change, then set back our participation change,
            // and send our participation status to the server
            //

            // alter remote evt to set our local participationStatus
            var remoteEvt = this._getRemoteChange(remoteDiff, conflict.remote, conflict.id),
                logger = this.logger;
            this._mergeParticipationChange(remoteEvt, conflict, function(res) {
                // if res==false, lightning didn't find an event with this uid.
                // we can ignore all data regarding this event
                if (res) {
                    logger.debug("ParticipationChange conflict: found event " + conflict.id + " in local database", false, false);
                    remoteToApply[conflict.remote].push(remoteEvt);
                    // send our local participationChange to the server
                    var localEvt = {
                        extId: conflict.id,
                        action: "participationChanged",
                        sequence: localDiff.event[conflict.id].sequence,
                        notification: localDiff.event[conflict.id].notification,
                        recurrenceId: conflict.recurrenceId
                    };
                    localToSend[conflict.id + "-" + conflict.recurrenceId] = localEvt;
                } else {
                    logger.debug("ParticipationChange conflict: event " + conflict.id + " not found in local database", false, false);
                }
                onConflictResolved();
            });

        } else if (conflict.remote == "updated") {
            // event has been participation changed locally and updated remotely
            var localEmail = cal.getCalendarManager().getCalendarPref_(this.calendar, "X-OBM-EMAIL");
            var remoteEvt = this._getRemoteChange(remoteDiff, conflict.remote, conflict.id);
            var localParticipationChanges = this._getAllParticipationChanges(localDiff.event[conflict.id]);
            this.logger.debug("remoteEvt: " + remoteEvt, false, false);
            this.logger.debug("localParticipationChanges: " + JSON.stringify(localParticipationChanges), false, false);
            this._mergeParticipationChangeOnAnUpdatedEvent(remoteEvt, conflict, localParticipationChanges, function(success, pChanges, applyRemote) {
                if (success) {
                    for (var pChange of pChanges) {
                        localToSend[conflict.id + "-" + pChange.recurrenceId] = pChange;
                    }
                    if (applyRemote) {
                        remoteToApply[conflict.remote].push(remoteEvt);
                    }
                }
                onConflictResolved();
            });

        } else if (conflict.remote == "removed") {
            // the event has been removed remotely : we don't care about participation change
            remoteToApply.removed.push(conflict.id);
            onConflictResolved();
        }
    },

    _resolveConflictUpdated: function(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved) {
        var serverPriority = this.serverPriority;
        this.logger.debug("_resolveConflictUpdated remote change : " + conflict.remote, false, false);
        if (conflict.remote == "removed") {
            // don't have sequence from server on delete events... :-(
            if (serverPriority) {
                remoteToApply.removed.push(conflict.id);
            } else {
                localToSend[conflict.id] = localDiff.event[conflict.id];
            }
            onConflictResolved();
        } else if (conflict.remote == "updated") {
            var self = this;
            var resolveThroughSequence = function(localEvent) {
                var localSequence = UtilsOBM.getSequence(localEvent);
                var remoteSequence = remoteEvt.generation;
                self.logger.debug("localSequence: " + localSequence + " , remoteSequence: " + remoteSequence, false, false);
                if (remoteSequence > localSequence) {
                    remoteToApply.updated.push(remoteEvt);
                    return true;
                } else if (remoteSequence < localSequence) {
                    localToSend[conflict.id] = localDiff.event[conflict.id];
                    return true;
                }
                return false;
            };

            var resolveThroughPriority = function() {
                if (serverPriority) {
                    remoteToApply.updated.push(remoteEvt);
                } else {
                    localToSend[conflict.id] = localDiff.event[conflict.id];
                }
            }

            //we should find if this is a real update or only a participationchange on an occurrence
            var remoteEvt = this._getRemoteChange(remoteDiff, conflict.remote, conflict.id);

            // If remoteEvent hash didn't change, apply local change and don't go further
            if (this._hashEqual(localDiff.event[conflict.id], remoteEvt)) {
                localToSend[conflict.id] = localDiff.event[conflict.id];
                return onConflictResolved();
            }

            var newEvent = this.calSync._createLocalEvent(remoteEvt, this.calendar);


            self.logger.debug("Fetching local event " + remoteEvt.extId + " from calendar", false, false);
            this._getEventFromCalendar(remoteEvt.extId, function(localEvent) {
                if (!localEvent) {
                    // this shouldn't happen : localEvent marqued as "updated" in the changelog
                    // but not found in the calendar
                    self.logger.debug("Local event " + remoteEvt.extId + " not found => applying remote", false, false);
                    remoteToApply.updated.push(remoteEvt);
                    return onConflictResolved();
                }

                // J#3455 Applying remote status on local event before resolving conflict because we can't rely on
                // this only "sequence" resolver (which would wrongly erase remote status by local status).
                // Changing the participation state doesn't increment the sequence.
                // J#3898 Also it is effective only if the localEvent isn't a reinitialization of a recurrent event.
                if (!UtilsOBM.localEventIsReinitialized(localEvent, newEvent)) {
                    self.logger.debug("Applying remote participationChanges on local updated evt", false, false);
                    localEvent = UtilsOBM.applyRemoteParticipationChanges(newEvent, localEvent, true);
                    remoteToApply.updated.push(localEvent);
                    localToSend[conflict.id] = localDiff.event[conflict.id];
                }

                if (resolveThroughSequence(localEvent)) {
                    self.logger.debug("_resolveConflictUpdated remote change resolved through sequence", false, false);
                    return onConflictResolved();
                }
                if (!newEvent.recurrenceInfo || UtilsOBM.hasPropertiesDifferences(newEvent, localEvent)) {
                    resolveThroughPriority();
                    return onConflictResolved();
                }

                //test occurrences
                var occurrenceIds = newEvent.recurrenceInfo.getExceptionIds({});
                var occurrencesChanges = false;
                for (var occurrenceId of occurrenceIds) {
                    var newOccurrence = newEvent.recurrenceInfo.getExceptionFor(occurrenceId, true);
                    var localOccurrence = localEvent.recurrenceInfo.getExceptionFor(occurrenceId, true);
                    if (UtilsOBM.hasPropertiesDifferences(newOccurrence, localOccurrence)) {
                        self.logger.debug("Found real modifications between the two events, on an exception => sequence based conflict resolution", false, false);
                        occurrencesChanges = true;
                    }
                }
                if (occurrencesChanges) {
                    resolveThroughPriority();
                    return onConflictResolved();
                }
                return onConflictResolved();
            });


        } else if (conflict.remote == "participationChanged") {
            // event is updated on local, and participationChanged on remote side...
            // => update local event with remote participation change and send the updated event
            localToSend[conflict.id] = localDiff.event[conflict.id];
            this._updateParticipation(this._getRemoteChange(remoteDiff, conflict.remote, conflict.id), function(newLocalEvent) {
                if (newLocalEvent) {
                    remoteToApply.updated.push(newLocalEvent);
                }
                onConflictResolved();
            });
        }

    },
    _resolveConflictAdded: function(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved) {
        // hardcore: we created an event locally and it is already updated on remote side...
        var remoteEvt = this._getRemoteChange(remoteDiff, conflict.remote, conflict.id);
        if (remoteEvt.owner == this.calSync._login.split("@")[0]) {
            //created in this tb, accepted by an attendee in other tb, created in obm by attendee sync :
            //do not send update or attendee status will be lost
        } else {
            //created in Ui, accepeted by mail before sync in this tb : add to updated to send participation
            localToSend[conflict.id] = localDiff.event[conflict.id];
        }
        onConflictResolved();
    },

    resolveConflicts: function(localDiff, remoteDiff, onConflictsResolved) {
        var conflicts = [],
            remoteToApply = { participationChanged: [], updated: [], removed: [] },
            localToSend = {},
            self = this,
            onConflictResolved = function() {
                if (conflicts.length) {
                    var conflict = conflicts.shift()
                    if (conflict.local == "removed") {
                        self.logger.debug("Launching _resolveConflictRemoved on " + conflict.id, false, false);
                        self._resolveConflictRemoved(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved);
                    } else if (conflict.local == "updated") {
                        self.logger.debug("Launching _resolveConflictUpdated on " + conflict.id, false, false);
                        self._resolveConflictUpdated(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved);
                    } else if (conflict.local == "added") {
                        self.logger.debug("Launching _resolveConflictAdded on " + conflict.id, false, false);
                        self._resolveConflictAdded(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved);
                    } else if (conflict.local == "participationChanged") {
                        self.logger.debug("Launching _resolveConflictParticipationChanged on " + conflict.id, false, false);
                        self._resolveConflictParticipationChanged(conflict, localDiff, remoteDiff, localToSend, remoteToApply, onConflictResolved);
                    } else {
                        self.logger.debug("Unexpected conflict state: " + conflict.local, false, false);
                        onConflictResolved();
                    }
                } else {
                    self.logger.debug("All conflicts resolved, returning", false, false);
                    onConflictsResolved(localToSend, remoteToApply);
                }
            };
        conflicts = this._findConflicts(localDiff, remoteDiff, localToSend, remoteToApply);
        this.logger.debug("Will launch conflict resolution, I got " + conflicts.length + " conflicts", false, false);
        onConflictResolved();
    }
};