/******************************************************************************
Copyright (C) 2011-2012 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/

var EXPORTED_SYMBOLS = ["UtilsOBM"];

var { Services } = ChromeUtils.import("resource://gre/modules/Services.jsm");
var { OBMGlobalData } = ChromeUtils.import("resource://obm-extension/modules/GlobalData.jsm");
var { OBMXPathFirst } = ChromeUtils.import("resource://obm-extension/modules/XMLUtilsOBM.jsm");
var { XPCOMUtils } = ChromeUtils.import("resource://gre/modules/XPCOMUtils.jsm");

var { ObmLogger } = ChromeUtils.import("chrome://obm-extension/content/components/logger/logger.js");

var { ObmCalendarsManager } = ChromeUtils.import("chrome://obm-extension/content/components/obmCalendarsManager.js");
var { ObmAddressBooksManager } = ChromeUtils.import('chrome://obm-extension/content/components/obmAddressBooksManager.js');

var { cal } = ChromeUtils.import("resource://calendar/modules/calUtils.jsm");

const Cc = Components.classes;
const Ci = Components.interfaces;

/** @ignore */
function XMLSerializer() {
    return Components.classes[
        "@mozilla.org/xmlextras/xmlserializer;1"
    ].createInstance(Components.interfaces.nsIDOMSerializer);
}

/** @ignore */
function XMLHttpRequest() {
    return Components.classes[
        "@mozilla.org/xmlextras/xmlhttprequest;1"
    ].createInstance(Components.interfaces.nsIXMLHttpRequest);
}

/** @ignore */
function DOMParser() {
    return Components.classes[
        "@mozilla.org/xmlextras/domparser;1"
    ].createInstance(Components.interfaces.nsIDOMParser);
}

/** @ignore */
function DateTime() {
    return Components.classes["@mozilla.org/calendar/datetime;1"].createInstance(
        Components.interfaces.calIDateTime
    );
}

/**
 * return the current password manager (see PasswordManager)
 */
var _gPasswdMgr = null;

function getPasswordManager() {
    if (_gPasswdMgr == null) _gPasswdMgr = new PasswordManager();
    return _gPasswdMgr;
}

/**
 * wrapper around the mozilla password manager. It uses the right components,
 * depending of the version of Mozilla
 */
function PasswordManager() {
    if ("@mozilla.org/login-manager;1" in Components.classes) {
        this.manager = Components.classes[
            "@mozilla.org/login-manager;1"
        ].getService(Components.interfaces.nsILoginManager);
        this.useLoginManager = true;
    } else {
        // TB2 ?
        this.manager = Components.classes[
            "@mozilla.org/passwordmanager;1"
        ].getService(Components.interfaces.nsIPasswordManager);
        this.useLoginManager = false;
    }
}

PasswordManager.prototype = {
    setPassword: function(aHost, aUsername, aPassword) {
        // remove the current password for this host if it already exists
        var username = {};
        var password = {};

        var login = this.getPassword(aHost, username, password);
        if (!this.useLoginManager) {
            if (login) {
                this.manager.removeUser(aHost, username.value);
            }
            this.manager.addUser(aHost, aUsername, aPassword);
        } else {
            var fixedPassword =
                aPassword == null || aPassword == "" ?
                UtilsOBM.PASSWORD_NOT_CONFIGURED :
                aPassword;
            if (login && aUsername == username.value) {
                var newLogin = login.clone();
                newLogin.password = fixedPassword;
                this.manager.modifyLogin(login, newLogin);
            } else {
                if (login) this.manager.removeLogin(login);

                var loginInfo = Components.classes[
                    "@mozilla.org/login-manager/loginInfo;1"
                ].createInstance(Components.interfaces.nsILoginInfo);
                loginInfo.init(aHost, null, aHost, aUsername, fixedPassword, "", "");
                this.manager.addLogin(loginInfo);
            }
        }
    },

    getPassword: function(aHost, aUsername /* object */ , aPassword /* object */ ) {
        UO_logger.debug("getPassword", false, false);
        if (!this.useLoginManager) {
            UO_logger.debug("useLoginManager false", false, false);
            var passwords = this.manager.enumerator;
            while (passwords.hasMoreElements()) {
                var password = passwords
                    .getNext()
                    .QueryInterface(Components.interfaces.nsIPassword);
                if (password.host == aHost) {
                    aUsername.value = password.user;
                    aPassword.value = password.password;
                    return true;
                }
            }
        } else {
            UO_logger.debug("useLoginManager true", false, false);
            var logins = this.manager.getAllLogins({});
            for (var i = 0; i < logins.length; i++) {
                UO_logger.debug(
                    "logins[" + i + "].hostname" + logins[i].hostname,
                    false,
                    false
                );
                UO_logger.debug(
                    "logins[" + i + "].username" + logins[i].username,
                    false,
                    false
                );
                // UO_logger.debug(
                //   "logins[" + i + "].password" + logins[i].password,
                //   false,
                //   false
                // );
                if (logins[i].hostname == aHost) {
                    aPassword.value = logins[i].password;
                    aUsername.value = logins[i].username;
                    return logins[i];
                }
            }
        }
        return false;
    }
};

/* Compatibility layer */

// {{ START_COMPAT_TB17
if (!cal.wrapInstance) {
    cal.wrapInstance = function wrapInstance(aObj, aInterface) {
        if (!aObj) return null;
        try {
            return aObj.QueryInterface(aInterface);
        } catch (e) {
            return null;
        }
    };
}
// END_COMPAT_TB17 }}}

// {{{ START_COMPAT_TB31
// See bug 1063085
if (!cal.dtz.dateTimeToJsDate) {
    cal.dtz.dateTimeToJsDate = function(cdt) {
        if (cdt.timezone == cal.floating()) {
            return new Date(
                cdt.year,
                cdt.month,
                cdt.day,
                cdt.hour,
                cdt.minute,
                cdt.second
            );
        } else {
            return new Date(cdt.nativeTime / 1000);
        }
    };
}

// END_COMPAT_TB31 }}}

/* UtilsOBM */

/**
 * UtilsOBM est une classe rassemblant toutes les fonctions "utilitaires" de
 * l'extension.
 * <p>
 * Toutes ses fonctions sont statiques.
 *
 * @constructor
 */
var UtilsOBM = {};

UtilsOBM.PASSWORD_NOT_CONFIGURED = "notConfiguredPassword";

UtilsOBM.getCalendarManager = function() {
    return new ObmCalendarsManager();
}

UtilsOBM.getAddressbookManager = function() {
    return new ObmAddressBooksManager();
}

UtilsOBM.getSyncManager = function() {
    var { ObmSyncManager } = ChromeUtils.import("chrome://obm-extension/content/components/obmSyncManager.js");
    return new ObmSyncManager();
};

UtilsOBM.getOrigin = function UO_getOrigin() {
    if (
        UtilsOBM.getOrigin.obmVersion == "unknown" ||
        UtilsOBM.getOrigin.ltnVersion == "unknown"
    ) {
        UO_logger.error(
            "Addon version has not been retrieved yet, check async timing",
            true,
            false
        );
    }
    var origin =
        "thunderbird[ext:" +
        UtilsOBM.getOrigin.obmVersion +
        ", light: " +
        UtilsOBM.getOrigin.ltnVersion +
        "]";

    return origin;
};

UtilsOBM.getOrigin.ltnVersion = "unknown";
UtilsOBM.getOrigin.obmVersion = "unknown";

UtilsOBM.getOrigin.init = function UO_getOrigin_init() {
    const LIGHTNING_UUID = "{e2fda1a4-762b-4020-b5ad-a41df1933103}";
    const OBM_UUID = "obm-connector@aliasource.fr";

    function saveAddonVersion(ltnVersion, obmVersion) {
        if (ltnVersion) {
            // For OBM 2.4 we have a problem with version comparator implemented
            // in obm-sync, so we fool lightning version here...
            var match = ltnVersion.match(/(\d+)obm/);
            var version = match ? parseInt(match[1], 10) : null;
            if (version && version > 9) {
                ltnVersion = ltnVersion.replace(version + "obm", "09obm");
            }
        }
        UtilsOBM.getOrigin.ltnVersion = ltnVersion || "unknown";
        UtilsOBM.getOrigin.obmVersion = obmVersion || "unknown";
    }

    var emClass = Components.classes["@mozilla.org/extensions/manager;1"];
    if (emClass) {
        var em = emClass.getService(Components.interfaces.nsIExtensionManager);
        saveAddonVersion(
            em.getItemForID(LIGHTNING_UUID).version,
            em.getItemForID(OBM_UUID).version
        );
    } else {
        Components.utils.import("resource://gre/modules/AddonManager.jsm");
        AddonManager.getAddonsByIDs([LIGHTNING_UUID, OBM_UUID], function(aAddons) {
            saveAddonVersion(aAddons[0].version, aAddons[1].version);
        });
    }
};

UtilsOBM.logToFile = function(aString) {
    UO_logger.notice(aString, false, false);
};

/**
 * Transforme une date renvoyée par OBM en un object JavaScript de type Date.
 *
 * @param {String}
 *            aXMLDate Une date au format ISO-8601
 * @type Date
 */
UtilsOBM.xmlDateToJSDate = function UO_xmlDateToJSDate(aXMLDate) {
    var jsDate = new Date();
    try {
        jsDate.setISO8601(aXMLDate);
        return jsDate;
    } catch (e) {
        return "";
    }
};

/**
 * Transforme un object JavaScript de type Date en une date compr�hensible par
 * OBM.
 *
 * @param {Date}
 *            aJSDate Un object JavaScript de type Date
 * @type String
 */
UtilsOBM.jsDateToXMLDate = function UO_jsDateToXMLDate(aJSDate) {
    var xmlDate = aJSDate.toISO8601String();
    return xmlDate;
};

UtilsOBM.restDateToJSDate = function UO_restDateToJSDate(aRestDate) {
    var jsDate = new Date();
    try {
        jsDate.setTime(aRestDate);
        return jsDate;
    } catch (e) {
        return "";
    }
};

UtilsOBM.jsDateToRestDate = function UO_jsDateToRestDate(aJSDate) {
    var rDate = aJSDate.getTime();
    return rDate;
};

UtilsOBM.capabilitySupported = function(aCapName, aDefault = "false") {
    let prefValue = UtilsOBM.getPrefSafe(
        "extensions.obm.server.capabilities." + aCapName,
        aDefault
    );
    return prefValue == "true" || prefValue == "yes" || prefValue == "on";
};

// {{{ START_COMPAT_TB17
// UtilsOBM.getPrefSafe can be replaced with Preferences.get in TB24+
UtilsOBM.getPrefSafe = function(aPrefName, aDefault) {
    try {
        var { Preferences } = Components.utils.import("resource://gre/modules/Preferences.jsm");
        UtilsOBM.getPrefSafe = Preferences.get.bind(Preferences);
    } catch (e) {
        // UtilsOBM.getPrefSafe = cal.getPrefSafe.bind(cal);
    }
    return UtilsOBM.getPrefSafe(aPrefName, aDefault);
};
UtilsOBM.setPrefSafe = function(aPrefName, aValue) {
    var { Preferences } = Components.utils.import("resource://gre/modules/Preferences.jsm");
    UtilsOBM.setPrefSafe = Preferences.set.bind(Preferences);
    return UtilsOBM.setPrefSafe(aPrefName, aValue);
};
UtilsOBM.isPrefSet = function(aPrefName) {
    try {
        var { Preferences } = Components.utils.import("resource://gre/modules/Preferences.jsm");
        UtilsOBM.isPrefSet = Preferences.isSet.bind(Preferences);
    } catch (e) {
        UtilsOBM.isPrefSet = function(x) {
            let sp = Services.prefs;
            let nIPB = Components.interfaces.nsIPrefBranch;
            return sp.getPrefType(x) != nIPB.PREF_INVALID && sp.prefHasUserValue(x);
        };
    }
    return UtilsOBM.isPrefSet(aPrefName);
};
UtilsOBM.resetPref = function(aPrefName) {
    try {
        var { Preferences } = Components.utils.import("resource://gre/modules/Preferences.jsm");
        UtilsOBM.resetPref = Preferences.reset.bind(Preferences);
    } catch (e) {
        UtilsOBM.resetPref = function(x) {
            Services.prefs.clearUserPref(x);
        };
    }
    return UtilsOBM.resetPref(aPrefName);
};
// END_COMPAT_TB17 }}}

/**
 * Gets a localized preference, see: https://developer.mozilla.org/en-US/docs/Code_snippets/Preferences#nsIPrefLocalizedString
 * Use this function instead of setCharPref if the value is or was a localized
 * preference where the name of the file it is translated in is the default
 * value.
 *
 * @param {String} aPrefName     The complete name of the preference.
 * @param {String} aDefaultValue The default value to use if the preference doesn't exist.
 */
UtilsOBM.getLocalizedPref = function(aPrefName, aDefaultValue) {
    const nIPL = Components.interfaces.nsIPrefLocalizedString;
    try {
        return Services.prefs.getComplexValue(aPrefName, nIPL).data;
    } catch (e) {
        return aDefaultValue;
    }
};

/**
 * Sets a localized preference, see: https://developer.mozilla.org/en-US/docs/Code_snippets/Preferences#nsIPrefLocalizedString
 * Use this function instead of setCharPref if the value is or was a localized
 * preference where the name of the file it is translated in is the default
 * value.
 *
 * @param {String} aPrefName     The complete name of the preference.
 * @param {String} aValue        The value to set the preference to.
 */
UtilsOBM.setLocalizedPref = function(aPrefName, aValue) {
    const nIPL = Components.interfaces.nsIPrefLocalizedString;
    var pls = Components.classes[
        "@mozilla.org/pref-localizedstring;1"
    ].createInstance(nIPL);
    pls.data = aValue;
    var prefs = Components.classes[
        "@mozilla.org/preferences-service;1"
    ].getService(Components.interfaces.nsIPrefBranch);
    prefs.setComplexValue(aPrefName, nIPL, pls);
    prefs.savePrefFile(null);
};

UtilsOBM.deletePreferenceBranch = function(aBranch) {
    try {
        Services.prefs.deleteBranch(aBranch);
    } catch (e) {
        //do nothing
    }
};

UtilsOBM.getPreferenceBranchChildren = function(aBranch) {
    return Services.prefs.getChildList(aBranch, {});
};

UtilsOBM.getNotificationStatus = function(calendar, itemId) {
    //this.logger.debug("getEmailNotification: "+itemId);
    //UO_logger.debug("getNotificationStatus",false,false);

    var infos = UtilsOBM.getCalendarSyncItem(calendar, itemId);
    //UO_logger.debug("getNotificationStatus returning :" + (infos.notification ? "true" : "false") ,false,false);
    return infos.notification ? true : false;
    //UO_logger.debug("getNotificationStatus",false,false);
};

UtilsOBM.setNotificationStatus = function(
    calendar,
    itemId,
    itemAction,
    sendNotification
) {
    UO_logger.debug(
        "setEmailNotification: " +
        calendar.id +
        " " +
        itemId +
        " " +
        itemAction +
        " " +
        sendNotification,
        false,
        false
    );
    var infos = UtilsOBM.getCalendarSyncItem(calendar, itemId);
    infos.notification = sendNotification;
    UtilsOBM.setCalendarSyncItem(calendar, itemId, infos);
};

/**
 * store the id (& recurrence id) of a cancelled event
 *
 */
UtilsOBM.storeEventCancelled = function(evt) {
    // find all calendars matching attendees...
    var id = evt.id;
    var rid = evt.recurrenceId ? evt.recurrenceId.nativeTime || "" : "";
    var uid = id + "_" + rid;

    UtilsOBM.setPrefSafe("extensions.obm.cancelledEvents." + uid, true);
};

UtilsOBM.isEventCancelled = function(evt) {
    var id = evt.id;
    var rid = evt.recurrenceId ? evt.recurrenceId.nativeTime || "" : "";
    var uid = id + "_" + rid;

    return UtilsOBM.getPrefSafe("extensions.obm.cancelledEvents." + uid, false);
};

/*
 * systeme d'enregistrement des modifications locales à envoyer sur le serveur sous forme de :
 * itemId: { action: added|removed|updated, notification: true|false }
 *
 * ex :
 * UtilsOBM.setCalendarSyncItem( aCalendar, aItem, { action: "added" } );
 *
 * var itemProperties = UtilsOBM.getCalendarSyncItem( aCalendar, aItem );
 * debug ( itemProperties.action );
 *
 */

/**
 * @private
 */
UtilsOBM.getCalendarSyncItem = function(calendar, item) {
    var id = "";
    if (typeof item == "string") {
        id = item;
    } else {
        id = item.id;
    }
    //UO_logger.debug("getCalendarSyncItem: extensions.obm.changesv2."+calendar.id+"."+id,false,false);
    var existing = UtilsOBM.getPrefSafe(
        "extensions.obm.changesv2." + calendar.id + "." + id,
        ""
    );
    //UO_logger.debug("getCalendarSyncItem fetched "+existing,false,false);
    var container = {};
    if (existing.length) {
        //UO_logger.debug("getCalendarSyncItem: "+existing,false,false);
        container = JSON.parse(existing);
    }
    return container;
};

/**
 * @private
 */
UtilsOBM.setCalendarSyncItem = function(calendar, item, obj) {
    var id = "";
    if (typeof item == "string") {
        //UO_logger.debug("setCalendarSyncItem : item is a string :"+item,false,false);
        id = item;
    } else {
        //UO_logger.debug("setCalendarSyncItem : typeof item :"+(typeof item),false,false);
        //UO_logger.debug("setCalendarSyncItem : item.id :"+item.id,false,false);
        id = item.id;
    }
    //UO_logger.debug("setCalendarSyncItem : "+calendar.id+"."+id+" :"+JSON.stringify(obj),false,false);
    UtilsOBM.setPrefSafe(
        "extensions.obm.changesv2." + calendar.id + "." + id,
        JSON.stringify(obj)
    );
};

/**
 * @private
 */
UtilsOBM.removeCalendarSyncItem = function(calendar, item) {
    var id = "";
    if (typeof item == "string") {
        id = item;
    } else {
        id = item.id;
    }
    UO_logger.debug(
        "removeCalendarSyncItem, " + calendar + "." + id,
        false,
        false
    );
    UtilsOBM.resetPref("extensions.obm.changesv2." + calendar.id + "." + id);
};

UtilsOBM.removeAllCalendarSyncItems = function(calendar) {
    UO_logger.debug(
        "removeAllCalendarSyncItems, " + typeof calendar,
        false,
        false
    );
    UO_logger.debug("removeAllCalendarSyncItems, " + calendar.id, false, false);
    UO_logger.debug(
        "branch = extensions.obm.changesv2." + calendar.id,
        false,
        false
    );
    UtilsOBM.deletePreferenceBranch("extensions.obm.changesv2." + calendar.id);
    //UO_logger.debug("/removeAllCalendarSyncItems",false,false);
};

UtilsOBM.getAllCalendarSyncItems = function(calendar) {
    var prefbranchkey = "extensions.obm.changesv2." + calendar.id;
    var keys = UtilsOBM.getPreferenceBranchChildren(prefbranchkey);
    UO_logger.debug(
        "Keys for branch " + prefbranchkey + ": " + JSON.stringify(keys),
        false,
        false
    );
    var back = {};
    for (var i in keys) {
        var k = keys[i].substr(prefbranchkey.length + 1);
        back[k] = UtilsOBM.getCalendarSyncItem(calendar, k);
    }
    return back;
};

UtilsOBM.addEventToChangesList = function UO_addEventToChangesList(
    aEvent,
    aChangeType
) {
    UO_logger.debug("UtilsOBM.addEventToChangesList", false, false);
    UO_logger.debug(
        "addEventToChangesList:    - aChangeType: " + aChangeType,
        false,
        false
    );
    UO_logger.debug(
        "addEventToChangesList:    - aItemID: " + aEvent.id,
        false,
        false
    );
    UO_logger.debug(
        "addEventToChangesList:    - aItem recurrenceId: " + aEvent.recurrenceId,
        false,
        false
    );

    if (aEvent.id == null || aEvent.id == "") {
        UO_logger.error(
            "addEventToChangesList id is null : (" + aEvent.id + ")",
            true,
            false
        );
        return;
    }

    // get last recorded change for this item
    var existingChange = UtilsOBM.getCalendarSyncItem(aEvent.calendar, aEvent);
    existingChange.action = existingChange.action || "";
    existingChange.hash = UtilsOBM.getHash(aEvent);

    if (aChangeType == "updated") {
        // if we were in "participationChanged" now there is a major change so we're in a "updated"
        if (existingChange.action == "participationChanged") {
            existingChange.action = "updated";
            // remove recurrenceIds markers
            if (existingChanges.recurrenceChange) {
                delete existingChanges.recurrenceChange;
                delete existingChanges.recurrenceIds;
            }
        }
        if (!existingChange.action) {
            existingChange.action = "updated";
        }
    } else if (aChangeType == "removed") {
        // if we already are in added then we remove the item without sending to server
        if (existingChange.action == "added") {
            UtilsOBM.removeCalendarSyncItem(aEvent.calendar, aEvent);
            return;
        }
        existingChange.action = "removed";
        existingChange.generation = UtilsOBM.getSequence(aEvent);
        // remove recurrenceIds markers of participationChanged action (if any)
        if (existingChange.recurrenceChange) {
            delete existingChange.recurrenceChange;
            delete existingChange.recurrenceIds;
        }
    } else if (
        aChangeType == "participationChanged" &&
        (!existingChange.action || existingChange.action == aChangeType)
    ) {
        existingChange.action = "participationChanged";
        existingChange.sequence = UtilsOBM.getSequence(aEvent);
        // to mark an update on the recurrence
        existingChange.recurrenceChange = existingChange.recurrenceChange || false;
        // to mark an update on an occurence (content are recurrenceId of occurences where participation was modified)
        existingChange.recurrenceIds = existingChange.recurrenceIds || [];
        if (aEvent.recurrenceId) {
            existingChange.recurrenceIds.push(aEvent.recurrenceId.icalString);
        } else {
            existingChange.recurrenceChange = true;
        }
    } else if (aChangeType == "added") {
        // If we try to add a previous removed event without synchronization in-between, it is an update action.
        if (existingChange.action == "removed") {
            existingChange.action = "updated";
        } else {
            existingChange.action = "added";
        }
        // remove recurrenceIds markers of participationChanged action (if any)
        if (existingChange.recurrenceChange) {
            delete existingChange.recurrenceChange;
            delete existingChange.recurrenceIds;
        }
    }
    UtilsOBM.setCalendarSyncItem(aEvent.calendar, aEvent, existingChange);
    UO_logger.debug(
        "addEventToChangesList: " + JSON.stringify(existingChange),
        false,
        false
    );
};

UtilsOBM.removeEventFromChangesList = function(aCalendar, aItemId) {
    UtilsOBM.removeCalendarSyncItem(aCalendar, aItemId);
};

/*
 * systeme d'enregistrement des modifications locales autres que les events calendar
 * à envoyer sur le serveur
 */

UtilsOBM.getItemType = function UO_getItemType(aItem) {
    var itemType;

    if (aItem instanceof Components.interfaces.calIEvent) {
        itemType = "calIEvent";
    } else if (aItem instanceof Components.interfaces.calITodo) {
        itemType = "calITodo";
    }

    return itemType;
};

UtilsOBM._getItemsChanges = function UO_getItemsChanges(
    aContainerType,
    aContainerName,
    aItemType
) {
    var changeTypes = ["added", "updated", "removed", "participationChanged"];

    var itemsChanges = new HashMap();

    for (let type of changeTypes) {
        var pref = UtilsOBM.getPrefSafe(
            "extensions.obm.changes." +
            aContainerType +
            "." +
            aContainerName +
            "." +
            aItemType +
            "." +
            type,
            ""
        );
        var changes = pref != "" ? pref.split(",") : new Array();
        var goodChanges = new Array();
        for (var i = 0; i < changes.length; i++) {
            if (changes[i] == null || changes[i] == "") {
                UO_logger.error(
                    "_getItemsChanges id is null (" +
                    "id:" +
                    changes[i] +
                    ", " +
                    aContainerType +
                    ", " +
                    aContainerName +
                    ", " +
                    aItemType +
                    ")",
                    true,
                    false
                );
            } else {
                goodChanges.push(changes[i]);
            }
        }
        itemsChanges.put(type, goodChanges);
    }

    return itemsChanges;
};

UtilsOBM._setItemsChanges = function UO_setItemsChanges(
    aContainerType,
    aContainerName,
    aItemType,
    aChanges
) {
    var changeTypes = aChanges.keySet();

    for (let changeType of changeTypes) {
        UtilsOBM.setPrefSafe(
            "extensions.obm.changes." +
            aContainerType +
            "." +
            aContainerName +
            "." +
            aItemType +
            "." +
            changeType,
            aChanges.get(changeType).join(",")
        );
    }
};

UtilsOBM.removeItemFromChangesList = function UO_removeItemFromChangesList(
    aContainerType,
    aContainerName,
    aItemType,
    aChangeType,
    aItemID
) {
    UO_logger.debug("UtilsOBM.removeItemFromChangesList", false, false);
    UO_logger.debug(
        "removeItemFromChangesList:    - aContainerType: " + aContainerType,
        false,
        false
    );
    UO_logger.debug(
        "removeItemFromChangesList:    - aContainerName: " + aContainerName,
        false,
        false
    );
    UO_logger.debug(
        "removeItemFromChangesList:    - aItemType: " + aItemType,
        false,
        false
    );
    UO_logger.debug(
        "removeItemFromChangesList:    - aChangeType: " + aChangeType,
        false,
        false
    );
    UO_logger.debug(
        "removeItemFromChangesList    - aItemID: " + aItemID,
        false,
        false
    );
    var allChanges = UtilsOBM._getItemsChanges(
        aContainerType,
        aContainerName,
        aItemType
    );
    var typeChanges = allChanges.get(aChangeType);
    var typeChanges2 = [];
    for (let itemID of typeChanges) {
        if (itemID != aItemID) {
            typeChanges2.push(itemID);
        }
    }
    allChanges.put(aChangeType, typeChanges2);
    UtilsOBM._setItemsChanges(
        aContainerType,
        aContainerName,
        aItemType,
        allChanges
    );
};

UtilsOBM.addItemToChangesList = function UO_addItemToChangesList(
    aContainerType,
    aContainerName,
    aItemType,
    aChangeType,
    aItemID
) {
    UO_logger.debug("UtilsOBM.addItemToChangesList", false, false);
    UO_logger.debug(
        "addItemToChangesList:    - aContainerType: " + aContainerType,
        false,
        false
    );
    UO_logger.debug(
        "addItemToChangesList:    - aContainerName: " + aContainerName,
        false,
        false
    );
    UO_logger.debug(
        "addItemToChangesList:    - aItemType: " + aItemType,
        false,
        false
    );
    UO_logger.debug(
        "addItemToChangesList:    - aChangeType: " + aChangeType,
        false,
        false
    );
    UO_logger.debug(
        "addItemToChangesList    - aItemID: " + aItemID,
        false,
        false
    );

    if (aItemID == null || aItemID == "") {
        UO_logger.error(
            "addItemToChangesList id is null : (" +
            aItemID +
            ", " +
            aContainerType +
            ", " +
            aContainerName +
            ", " +
            aItemType +
            ", " +
            aChangeType +
            ")",
            true,
            false
        );
        return;
    }

    if (aItemType == "calIEvent") {
        //find calendar
        var calMgr = Components.classes[
            "@mozilla.org/calendar/manager;1"
        ].getService(Components.interfaces.calICalendarManager);
        var calendars = calMgr.getCalendars({});
        var calendar = null;
        for (var i in calendars) {
            if (
                calMgr.getCalendarPref_(calendars[i], "X-OBM-NAME") == aContainerName
            ) {
                calendar = calendars[i];
                break;
            }
        }
        if (!calendar) {
            UO_logger.debug(
                "GRAVE: no calendar found with name " + aContainerName,
                false,
                false
            );
            return;
        }

        var listener = {};
        listener.localEvent = null;
        listener.onOperationComplete = function(
            aCalendar,
            aStatus,
            aOperationType,
            aId,
            aDetail
        ) {
            if (listener.localEvent == null) {
                UO_logger.debug("GRAVE: event not found ", false, false);
                return;
            }
            UtilsOBM.addEventToChangesList(listener.localEvent, aChangeType);
        };
        listener.onGetResult = function(
            aCalendar,
            aStatus,
            aItemType,
            aDetail,
            aCount,
            aItems
        ) {
            if (aCount > 0) {
                this.localEvent = aItems[0];
            }
        };
        // WARNING: this method is supposed to be asynchronous !!
        calendar.getItem(aItemID, listener);
        return;
    }

    var itemsChanges = UtilsOBM._getItemsChanges(
        aContainerType,
        aContainerName,
        aItemType
    );
    var itemsArray;

    function removeItem(element, index, array) {
        return element != this.itemID;
    }
    var toChangeType = "none";

    if (aChangeType == "updated") {
        // type == "updated" && already in "added" => don't add to "updated"
        if (itemsChanges.get("added").indexOf(aItemID) != -1) {
            toChangeType = "none";

            // type == "updated" && already in "participationChanged" => delete from "participationChanged"
            // add to "update"
        } else if (
            itemsChanges.get("participationChanged").indexOf(aItemID) != -1
        ) {
            itemsArray = itemsChanges.get("participationChanged");
            var o = {};
            o.itemID = aItemID;
            itemsArray = itemsArray.filter(removeItem, o);
            itemsChanges.put("participationChanged", itemsArray);

            // Add this event to "Has been updated" list
            toChangeType = "updated";
        } else {
            // type == "updated" and event out-of-sync => add to "added"
            // TODO removed : conflicts with ics
            /*if (aItemID.match(/(\w+\-){3}(\w)+/) ) {
                          //remove from updated
                          itemsArray = itemsChanges.get("updated");
                          var o = {}; o.itemID = aItemID;
                          itemsArray = itemsArray.filter(removeItem, o);
                          itemsChanges.put("updated", itemsArray);
                          //add to added
                          toChangeType = "added";
                      } else {
                          toChangeType = "updated";
                      }*/
            toChangeType = "updated";
        }
    }

    if (aChangeType == "removed") {
        // type == "removed" && already in "added" => delete from "added"
        // don't add to "removed"
        if (itemsChanges.get("added").indexOf(aItemID) != -1) {
            itemsArray = itemsChanges.get("added");
            var o = {};
            o.itemID = aItemID;
            itemsArray = itemsArray.filter(removeItem, o);
            itemsChanges.put("added", itemsArray);
            toChangeType = "none";
        }
        // type == "removed" && already in "updated" => delete from "updated"
        // add to "removed"
        else if (itemsChanges.get("updated").indexOf(aItemID) != -1) {
            itemsArray = itemsChanges.get("updated");
            var o = {};
            o.itemID = aItemID;
            itemsArray = itemsArray.filter(removeItem, o);
            itemsChanges.put("updated", itemsArray);

            toChangeType = "removed";

            // type == "removed" && already in "participationChanged" => delete from "participationChanged"
            // add to "remove"
        } else if (
            itemsChanges.get("participationChanged").indexOf(aItemID) != -1
        ) {
            itemsArray = itemsChanges.get("participationChanged");
            var o = {};
            o.itemID = aItemID;
            itemsArray = itemsArray.filter(removeItem, o);
            itemsChanges.put("participationChanged", itemsArray);

            toChangeType = "removed";
        } else {
            // type == "removed" && not in "added" or "updated"

            // TODO removed : conflicts with ics
            /*if (aItemID.match(/(\w+\-){3}(\w)+/)) {
                      //event out-of-sync do no add to "removed"
                      toChangeType = "none";
                  }*/
            toChangeType = "removed";
        }
    }

    if (aChangeType == "added") {
        toChangeType = "added";
    }

    if (aChangeType == "participationChanged") {
        // type == "participationChanged" && already in "updated" => leave as is
        if (itemsChanges.get("updated").indexOf(aItemID) != -1) {
            // Leave as is
            // type == "participationChanged" && already in "removed" => leave as is
            // This case should never happen anyway!
        } else if (itemsChanges.get("removed").indexOf(aItemID) != -1) {
            // Leave as is
            // type == "participationChanged" && already in "added" => leave as is
        } else if (itemsChanges.get("added").indexOf(aItemID) != -1) {
            // Leave as is
            // type == "participationChanged" && already in "participationChanged" => leave as is
        } else if (
            itemsChanges.get("participationChanged").indexOf(aItemID) != -1
        ) {
            // Leave as is
            // Event has not been flagged yet. let's flag it as "participationChanged"
        } else {
            toChangeType = "participationChanged";
        }
    }

    if (toChangeType != "none") {
        itemsArray = itemsChanges.get(toChangeType);
        // already a item with this change type
        if (itemsArray.indexOf(aItemID) != -1) {} else {
            itemsArray.push(aItemID);
            itemsChanges.put(toChangeType, itemsArray);
        }
    }

    UtilsOBM._setItemsChanges(
        aContainerType,
        aContainerName,
        aItemType,
        itemsChanges
    );
    UO_logger.debug("addItemToChangesList\r\n" + itemsChanges, false, false);
};

// FIXME: unused?
UtilsOBM.resetAllItemsChangesLists = function UO_resetAllItemsChangesLists(
    aContainerType,
    aContainerName,
    aItemType
) {
    var changes = new HashMap();
    changes.put("added", []);
    changes.put("updated", []);
    changes.put("removed", []);
    changes.put("participationChanged", []);

    UtilsOBM._setItemsChanges(aContainerType, aContainerName, aItemType, changes);
};

/*
 * XXX HACK no synchronous interface to get an item see bug #296832 or could an
 * observer be used instead ??
 */
/**
 * R�cup�re l'item de l'agenda, de fa�on synchrone.
 *
 * @param {calICalendar}
 *            aCalendar
 * @param {String}
 *            aItemID
 * @type calIItemBase
 */
UtilsOBM.getCalendarItem_Synchronous = function UO_getCalendarItem_Synchronous(
    aCalendar,
    aItemID
) {
    var listener = UtilsOBM._createListener(
        function(aCalendar, aStatus, aOperationType, aId, aDetail) {
            this.done = true;
        },
        function(aCalendar, aStatus, aItemType, aDetail, aCount, aItems) {
            if (aItems.length == 1) this.item = aItems[0];
        }
    );
    aCalendar.getItem(aItemID, listener);
    UtilsOBM._waitForDefined(listener.done);
    return listener.item;
};

/**
 * Retourne le nom OBM de l'agenda. (ex : "usera")
 *
 * @param {calICalendar}
 *            aCalendar
 * @type String
 */
UtilsOBM.getCalendarOBMName = function UO_getCalendarOBMName(aCalendar) {
    return cal.getCalendarManager().getCalendarPref_(aCalendar, "X-OBM-NAME");
};
UtilsOBM.getResourceOBMMail = function UO_getResourceOBMMail(aCalendar) {
    return cal
        .getCalendarManager()
        .getCalendarPref_(aCalendar, "X-OBM-RESOURCE-EMAIL");
};

/**
 * Retourne la liste des agendas OBM souscris
 *
 * @param {calICalendar}
 *            aCalendar
 * @type Array
 */
UtilsOBM.getSubscribedOBMCalendar = function UO_getSubscribedOBMCalendar() {
    var calendars = cal.getCalendarManager().getCalendars({});
    var OBMcalendars = new Array();
    for (let calendar of calendars) {
        if (calendar.type == "obm") {
            OBMcalendars.push(calendar);
        }
    }
    return OBMcalendars;
};
(UtilsOBM.getSubscribedOBMResource = function UO_getSubscribedOBMResource() {
    var resMgr = cal.getCalendarManager();
    var resources = cal.getCalendarManager().getCalendars({});
    var OBMresources = new Array();
    for (let resource of resources) {
        if (
            resource.type == "ics" &&
            resMgr.getCalendarPref_(resource, "X-OBM-RESOURCE-EMAIL") != null
        ) {
            OBMresources.push(resource);
        }
    }
    return OBMresources;
}),
(UtilsOBM.getOBMCalendarByUid = function UO_getOBMCalendarByUid(uid) {
    let calendars = UtilsOBM.getSubscribedOBMCalendar();

    for (let calendar of calendars) {
        if (UtilsOBM.getCalendarOBMName(calendar) == uid) {
            return calendar;
        }
    }

    return null;
}),
/**
 * Retourne Vrai si Ligthning est installe et active.
 *
 * @type boolean
 */
(UtilsOBM.isLightningInstalledAndEnabled = function UO_isLightningInstalledAndEnabled() {
    return !!Components.interfaces.calICalendarManager;
});
/**
 * Reset syncOnQuit pref after a reset.
 */
UtilsOBM.setSyncOnQuitAfterReset = function UO_setSyncOnQuitAfterReset() {
    var resetPrefName = "extensions.obm.syncOnQuit.beforeReset";
    var normalPrefName = "extensions.obm.syncOnQuit";

    if (UtilsOBM.isPrefSet(resetPrefName)) {
        var beforeReset = UtilsOBM.getPrefSafe(resetPrefName, false);

        UtilsOBM.setPrefSafe(normalPrefName, beforeReset);
        UtilsOBM.resetPref(resetPrefName);
    }
};

/**
 * Retourne le gestionnaire d'agendas de Calendar.
 *
 * @type calICalendarManager
 */
UtilsOBM.createAttendee = function UO_createAttendee(aCalendar, aOwnerName) {
    var calAttendee = Components.classes[
        "@mozilla.org/calendar/attendee;1"
    ].createInstance(Components.interfaces.calIAttendee);

    calAttendee.id = "MAILTO:" + aOwnerName;
    calAttendee.commonName =
        cal.getCalendarManager().getCalendarPref_(aCalendar, "X-OBM-NAME") +
        " (" +
        aOwnerName +
        ")";
    calAttendee.participationStatus = "";
    calAttendee.role = "";
    return calAttendee;
};

UtilsOBM.addOwnerToAttendees = function UO_addOwnerToAttendees(
    aCalendar,
    allAttendees
) {
    var ownerName = cal
        .getCalendarManager()
        .getCalendarPref_(aCalendar, "X-OBM-EMAIL");
    var addOwner = true;
    if (allAttendees == null || allAttendees.length == 0) {
        if (allAttendees == null) {
            allAttendees = [];
        }
    } else {
        for (let attendee of allAttendees) {
            if (attendee.id == "MAILTO:" + ownerName) {
                addOwner = false;
                break;
            }
            Components.reportError(
                "attendee -> " + attendee.id + " -> " + attendee.commonName
            );
        }
    }
    if (addOwner) {
        allAttendees.unshift(UtilsOBM.createAttendee(aCalendar, ownerName));
    }
    return allAttendees;
};
/**
 * Retourne la liste des agendas OBM souscrits.
 *
 * @type Array[String]
 */
UtilsOBM.getSubscribedCalendarsList = function UO_getSubscribedCalendarsList() {
    var calendarManager = cal.getCalendarManager();

    var calendars = calendarManager.getCalendars({});

    var array = new Array();

    for (let calendar of calendars) {
        if (calendar.type == "obm") {
            array.push(UtilsOBM.getCalendarOBMName(calendar));
        }
    }

    return array;
};

/**
 * Cr�� un <i>listener</i> d'op�ration vide, c'est-�-dire dont les <i>handlers</i>
 * ne "font rien".
 *
 * @type calIOperationListener
 */
UtilsOBM.createEmptyListener = function UO_createEmptyListener() {
    return UtilsOBM._createListener(function() {}, function() {});
};

/**
 * Cr�� un <i>listener</i> d'op�ration.
 *
 * @private
 * @param {function}
 *            aOnOperationComplete
 * @param {function}
 *            aOnGetResult
 * @type calIOperationListener
 */
UtilsOBM._createListener = function UO_createListener(
    aOnOperationComplete,
    aOnGetResult
) {
    return {
        onOperationComplete: aOnOperationComplete,
        onGetResult: aOnGetResult
    };
};

/**
 * Retourne apr�s XX millisecondes.
 *
 * @param {integer}
 *            aDuration Nombre de millisecondes � attendre.
 */
UtilsOBM.sleep = function UO_sleep(aDuration) {
    //dump("XXX: " + aDuration + "\n");
    var thread = Components.classes["@mozilla.org/thread;1"].createInstance(
        Components.interfaces.nsIThread
    );
    thread.currentThread.sleep(aDuration);
};

/**
 * Boucle tant que le param�tre est "undefined".
 *
 * @private
 * @param {Object}
 *            aProperty
 */
UtilsOBM._waitForDefined = function UO_waitForDefined(aProperty) {
    // YOB241006 var thread = Components.classes["@mozilla.org/thread;1"]
    // YOB241006 .createInstance(Components.interfaces.nsIThread);
    while (aProperty === undefined) {
        UtilsOBM.sleep(5);
    }
};

/**
 * Vide un noeud DOM, c'est � dire supprime tous ses fils.
 *
 * @param {DOMNode}
 *            aNode
 */
UtilsOBM.emptyDOMNode = function UO_emptyDOMNode(aNode) {
    while (aNode.firstChild) {
        aNode.removeChild(aNode.firstChild);
    }
};

/**
 * G�n�re une couleur al�atoire, au format HTML #xxyyzz.
 *
 * @type String
 */
UtilsOBM.getRandomColor = function UO_getRandomColor() {
    var result = "#";
    for (var i = 0; i < 3; i++) {
        var hexValue = Math.round(Math.random() * 255).toString(16);
        result += (hexValue.length == 1 ? "0" : "") + hexValue;
    }
    return result;
};

UtilsOBM.getLocalizedMessage = function UO_getLocalizedMessage(aString) {
    var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
        .getService(Components.interfaces.nsIStringBundleService)
        .createBundle("chrome://obm-extension/locale/errors_warnings.properties");
    return bundle.GetStringFromName(aString);
};

UtilsOBM.getLocalizedMessageWithParams = function UO_getLocalizedMessageWithParams(
    aString,
    params
) {
    var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
        .getService(Components.interfaces.nsIStringBundleService)
        .createBundle("chrome://obm-extension/locale/errors_warnings.properties");

    return bundle.formatStringFromName(aString, params, params.length);
};

UtilsOBM.showAlert = function UO_showAlert(aString) {
    var aMessage = UtilsOBM.getLocalizedMessage(aString);
    var aTitle = UtilsOBM.getLocalizedMessage("obmconnector");
    var alertsService = Components.classes[
        "@mozilla.org/alerts-service;1"
    ].getService(Components.interfaces.nsIAlertsService);
    alertsService.showAlertNotification(
        "chrome://obm-extension/skin/icons/obm_logo.png",
        aTitle,
        aMessage,
        false,
        "",
        null
    );
};

UtilsOBM.resetOBMCredentials = function() {
    var login = {},
        password = {};
    var gotCredentials = UtilsOBM.getOBMCredentials(login, password);
    if (gotCredentials) {
        UtilsOBM.setOBMCredentials("", "");
    }
};

UtilsOBM.getOBMCredentials = function UO_getOBMCredentials(
    aUsername /* object */ ,
    aPassword /* object */
) {
    const HOST = "obm-obm-obm"; // sort of key
    var passwordManager = getPasswordManager();
    var retour = false;

    UO_logger.debug("UtilsOBM.getOBMCredentials()", false, false);

    if (passwordManager.getPassword(HOST, aUsername, aPassword)) {
        retour =
            aUsername && aUsername.value && aUsername.value.length ? true : false;
    }

    UO_logger.debug("aUsername.value:" + aUsername.value, false, false);
    // UO_logger.debug("aPassword.value:" + aPassword.value, false, false);

    if (UtilsOBM.getPrefSafe("extensions.obm.rememberpassword", true) == false) {
        UO_logger.debug("extensions.obm.rememberpassword:", false, false);
        retour &= UtilsOBM.getTemporaryPassword(aPassword);
    }
    return retour;
};

UtilsOBM.setOBMCredentials = function UO_setOBMCredentials(
    aUsername,
    aPassword
) {
    const HOST = "obm-obm-obm"; // sort of key
    var passwordManager = getPasswordManager();
    passwordManager.setPassword(HOST, aUsername, aPassword);
};

UtilsOBM.setTemporaryPassword = function UO_setTemporaryPassword(aPassword) {
    OBMGlobalData.password = aPassword;
};

UtilsOBM.getTemporaryPassword = function UO_getTemporaryPassword(
    aPassword /* object */
) {
    if (OBMGlobalData.password) {
        aPassword.value = OBMGlobalData.password;
        return true;
    } else {
        return false;
    }
};

UtilsOBM.checkConnectionInformation = function UO_checkLoginInformation() {
    var pref = UtilsOBM.getPrefSafe("extensions.obm.server");
    return pref && UtilsOBM.getOBMCredentials({}, {});
};

UtilsOBM.getDateTime = function UO_getDateTime(timestamp) {
    var d = cal.dtz.jsDateToDateTime(new Date());
    // d.nativeTime = timestamp * 1000;

    return d.getInTimezone(cal.dtz.defaultTimezone);
};

// renvoie la date du jour, au format DateTime pour comparaison...
UtilsOBM.getNowDateTime = function UO_getNowDateTime() {
    return cal.now();
    /*
     * var nowDate = new Date(); var retour =
     * Components.classes["@mozilla.org/calendar/datetime;1"].createInstance(Components.interfaces.calIDateTime);
     *
     * retour.year = nowDate.getFullYear(); retour.month = nowDate.getMonth();
     * retour.day = nowDate.getDate(); retour.hour = nowDate.getHours();
     * retour.minute = nowDate.getMinutes(); retour.second =
     * nowDate.getSeconds();
     *
     * return retour;
     */
};
/* Paul Sowden -- http://delete.me.uk/2005/03/iso8601.html -- AFL license */
/** @ignore */
Date.prototype.setISO8601 = function setISO8601(string) {
    var regexp =
        "([0-9]{4})(-([0-9]{2})(-([0-9]{2})" +
        "(T([0-9]{2}):([0-9]{2})(:([0-9]{2})(.([0-9]+))?)?" +
        "(Z|(([-+])([0-9]{2}):([0-9]{2})))?)?)?)?";
    var d = string.match(new RegExp(regexp));

    var offset = 0;
    var date = new Date(d[1], 0, 1);

    if (d[3]) {
        date.setMonth(d[3] - 1);
    }
    if (d[5]) {
        date.setDate(d[5]);
    }
    if (d[7]) {
        date.setHours(d[7]);
    }
    if (d[8]) {
        date.setMinutes(d[8]);
    }
    if (d[10]) {
        date.setSeconds(d[10]);
    }
    if (d[12]) {
        date.setMilliseconds(Number("0." + d[12]) * 1000);
    }
    if (d[14]) {
        offset = Number(d[16]) * 60 + Number(d[17]);
        offset *= d[15] == "-" ? 1 : -1;
    }

    offset -= date.getTimezoneOffset();
    time = Number(date) + offset * 60 * 1000;
    this.setTime(Number(time));
};

/** @ignore */
Date.prototype.toISO8601String = function toISO8601String(format, offset) {
    /*
     * accepted values for the format [1-6]: 1 Year: YYYY (eg 1997) 2 Year and
     * month: YYYY-MM (eg 1997-07) 3 Complete date: YYYY-MM-DD (eg 1997-07-16) 4
     * Complete date plus hours and minutes: YYYY-MM-DDThh:mmTZD (eg
     * 1997-07-16T19:20+01:00) 5 Complete date plus hours, minutes and seconds:
     * YYYY-MM-DDThh:mm:ssTZD (eg 1997-07-16T19:20:30+01:00) 6 Complete date
     * plus hours, minutes, seconds and a decimal fraction of a second
     * YYYY-MM-DDThh:mm:ss.sTZD (eg 1997-07-16T19:20:30.45+01:00)
     */
    var date;

    if (!format) {
        format = 6;
    }
    if (!offset) {
        offset = "Z";
        date = this;
    } else {
        var d = offset.match(/([-+])([0-9]{2}):([0-9]{2})/);
        var offsetnum = Number(d[2]) * 60 + Number(d[3]);
        offsetnum *= d[1] == "-" ? -1 : 1;
        date = new Date(Number(Number(this) + offsetnum * 60000));
    }

    var zeropad = function(num) {
        return (num < 10 ? "0" : "") + num;
    };

    var str = "";
    str += date.getUTCFullYear();
    if (format > 1) {
        str += "-" + zeropad(date.getUTCMonth() + 1);
    }
    if (format > 2) {
        str += "-" + zeropad(date.getUTCDate());
    }
    if (format > 3) {
        str +=
            "T" + zeropad(date.getUTCHours()) + ":" + zeropad(date.getUTCMinutes());
    }
    if (format > 5) {
        var secs = Number(
            date.getUTCSeconds() +
            "." +
            (date.getUTCMilliseconds() < 100 ? "0" : "") +
            zeropad(date.getUTCMilliseconds())
        );
        str += ":" + zeropad(secs);
    } else if (format > 4) {
        str += ":" + zeropad(date.getUTCSeconds());
    }

    if (format > 3) {
        str += offset;
    }
    return str;
};

UtilsOBM.isValideDate = function UO_isValideDate(string) {
    var regexp =
        "[0-9]{4}-[0-9]{2}-[0-9]{2}" +
        "T[0-9]{2}:[0-9]{2}:[0-9]{2}.[0-9]+" +
        "Z|[-+][0-9]{2}:[0-9]{2}";
    var d = string.match(new RegExp(regexp));
    return d != null;
};

// (Exception supprimees -> nouveaux evenements crees)
UtilsOBM.addExceptionItem = function UO_addExceptionItem(
    aContainer,
    aContainerName,
    aItemID
) {
    var aItem = UtilsOBM.getCalendarItem_Synchronous(aContainer, aItemID);

    // is Occurence Event
    var newAddedItemsIDs = [];
    try {
        if (aItem.recurrenceInfo != null) {
            var aExceptionIDs = aItem.recurrenceInfo.getExceptionIds({});
            // has Exception
            if (aExceptionIDs.length != 0) {
                for (let aExceptionID of aExceptionIDs) {
                    var ocurrence = aItem.recurrenceInfo.getOccurrenceFor(
                        aExceptionID,
                        false
                    );
                    UtilsOBM.logToFile("exception : " + aExceptionID + "\n");
                    UtilsOBM.logToFile("ocurrence title : " + ocurrence.title + "\n");
                    UtilsOBM.logToFile(
                        "ocurrence recurrenceStartDate : " +
                        ocurrence.recurrenceStartDate +
                        "\n"
                    );

                    var uuidgen = Components.classes[
                        "@mozilla.org/uuid-generator;1"
                    ].getService(Components.interfaces.nsIUUIDGenerator);
                    var uuid = uuidgen.generateUUID().toString();

                    var newItem = ocurrence.clone();
                    newItem.recurrenceInfo = null;
                    newItem.recurrenceId = null;
                    newItem.parentItem = null;
                    newItem.title = ocurrence.title;
                    newItem.id = uuid.substring(1, uuid.length - 1);

                    var aListenerAdd = {};
                    aListenerAdd.onOperationComplete = function(cal, res, op, id, item) {
                        // Components.reportError("adopt -> cal : " + cal + ", res : " +
                        // res + ",op : " + op + ", id : " + id + ", item : " + item);
                        this.id = id;
                        this.done = true;
                    };
                    aContainer.adoptItem(newItem, aListenerAdd);
                    UtilsOBM._waitForDefined(aListenerAdd.done);
                    newAddedItemsIDs.push(aListenerAdd.id);
                }
            }
        }
    } catch (e) {}
    return newAddedItemsIDs;
};

var current_window = {};
UtilsOBM.setWindow = function UO_setWindow(window) {
    current_window = window;
};
UtilsOBM.getWindow = function UO_getWindow() {
    return current_window;
};

UtilsOBM.showWindow = function showWindow(aName, aParent, aId) {
    let wWatcher = Components.classes[
        "@mozilla.org/embedcomp/window-watcher;1"
    ].getService(Components.interfaces.nsIWindowWatcher);

    let uri = "chrome://obm-extension/content/" + aName;
    let options = "modal,centerscreen,chrome,resizable=no,dialog";
    wWatcher.openWindow(aParent, uri, aId || "", options, null);

    return !OBMGlobalData.canceled;
};

// Used in pref windows
UtilsOBM.validateIntegers = function validateIntegers(event) {
    if (isNaN(Number(event.target.value))) {
        var newValue = parseInt(event.target.value);
        event.target.value = isNaN(newValue) ? "" : newValue;
        event.preventDefault();
    }
};

UtilsOBM.validateNaturalNums = function validateNaturalNums(event) {
    UtilsOBM.validateIntegers(event);
    var num = event.target.value;
    if (num < 0) {
        event.target.value = -1 * num;
        event.preventDefault();
    }
};

UtilsOBM.importCalendar = function UO_importCalendar(aCalendar) {
    var calMgr = Components.classes["@mozilla.org/calendar/manager;1"].getService(
        Components.interfaces.calICalendarManager
    );
    var calName = calMgr.getCalendarPref_(aCalendar, "X-OBM-NAME");
    var containerType = Components.interfaces.calICalendar;
    var itemType = Components.interfaces.calIEvent;

    var listener = {};
    var self = this;
    listener.events = [];
    listener.imported = 0;
    listener.onOperationComplete = function(
        aCalendar,
        aStatus,
        aOperationType,
        aId,
        aDetail
    ) {
        // <TO_REMOVE>
        // Fix second onOperationComplete call (bug LT 0.8)
        // Fixed in LT 0.9
        try {
            if (aDetail.toString() == "Can't deduce item type based on QI") {
                return;
            }
        } catch (e) {}
        // </TO_REMOVE>
        if (Components.isSuccessCode(aStatus)) {
            for (let event of this.events) {
                var itemID = event.id;
                if (itemID.match(/(\w+\-){3}(\w)+/) && !event.hasProperty("X-OBM-ID")) {
                    UO_logger.debug(
                        "import item:" + event.title + "(" + itemID + ")",
                        false,
                        false
                    );
                    UtilsOBM.addItemToChangesList(
                        containerType,
                        calName,
                        itemType,
                        "added",
                        itemID
                    );
                    this.imported++;
                }
            }
            UO_logger.notice(this.imported + " events imported", false, false);
            var aMessage =
                calName +
                " : " +
                this.imported +
                " " +
                UtilsOBM.getLocalizedMessage("obm.import.alert");
            var aTitle = UtilsOBM.getLocalizedMessage("obmconnector");
            var promptService = Components.classes[
                "@mozilla.org/embedcomp/prompt-service;1"
            ].getService(Components.interfaces.nsIPromptService);
            promptService.alert(null, aTitle, aMessage);
        }
    };
    listener.onGetResult = function(
        aCalendar,
        aStatus,
        aItemType,
        aDetail,
        aCount,
        aItems
    ) {
        if (aCount > 0) {
            this.events = this.events.concat(aItems);
        }
    };

    aCalendar.getItems(
        Components.interfaces.calICalendar.ITEM_FILTER_TYPE_EVENT,
        0,
        null,
        null,
        listener
    );
};

UtilsOBM.compareVersions = function UO_compareVersions(aVersionA, aVersionB) {
    var comparator = Components.classes[
        "@mozilla.org/xpcom/version-comparator;1"
    ].getService(Components.interfaces.nsIVersionComparator);

    return comparator.compare(aVersionA, aVersionB);
};

UtilsOBM.getOBMEventId = function UO_getOBMEventId(aExtId, aEvent) {
    if (!aEvent) {
        return "";
    }
    if (aExtId) {
        if (!aEvent.hasProperty("X-OBM-ID")) {
            return "";
        } else {
            return aEvent.getProperty("X-OBM-ID");
        }
    } else {
        return aEvent.id;
    }
};
UtilsOBM.setOBMEventId = function UO_setOBMEventId(
    supportExtId,
    aEvent,
    aOBMId
) {
    if (supportExtId) {
        UO_logger.debug("setOBMEventId: " + typeof aOBMId, false, false);
        aEvent.setProperty("X-OBM-ID", aOBMId);
    } else {
        aEvent.id = aOBMId;
    }
};

UtilsOBM.reset = function UO_reset(cb) {
    let storedPreferences = [];

    UO_logger.debug("Reset = BEGIN =", false, false);
    UO_logger.debug("Reset - CALENDARS", false, false);
    // disable synchronization on start and quit
    let syncOnQuit = UtilsOBM.getPrefSafe("extensions.obm.syncOnQuit", false);
    UtilsOBM.setPrefSafe("extensions.obm.syncOnQuit.beforeReset", syncOnQuit);
    UtilsOBM.setPrefSafe("extensions.obm.syncOnQuit", false);

    if (UtilsOBM.getPrefSafe("extensions.obm.calEvents.is_sync_enabled", false)) {
        let obmCalendars = UtilsOBM.getCalendarManager().getOBMCalendars({});

        for (let calendar of obmCalendars) {
            UO_logger.debug(
                "Reset - Calendar " +
                calendar.name +
                " (readOnly: " +
                calendar.readOnly +
                ")"
            );
            UtilsOBM.applyRemoteCalendarRights(
                calendar,
                UtilsOBM.getCalInfoForLocalCalendar(calendar, calendar.readOnly),
                true
            );
        }

        for (let pref of UtilsOBM.getPreferenceBranchChildren(
                "extensions.obm.changes.calICalendar"
            )) {
            if (/readOnly/.test(pref)) {
                storedPreferences.push({
                    key: pref,
                    value: UtilsOBM.getPrefSafe(pref, false)
                });
            }
        }

        // Reset des categories
        var currentCategories = UtilsOBM.getPrefSafe("calendar.categories.names");
        if (currentCategories && currentCategories != "") {
            currentCategories = currentCategories.split(",");
        } else {
            currentCategories = [];
        }

        var resetCategories = [];

        for (let category of currentCategories) {
            if (!category.match(/\[OBM\]/)) {
                resetCategories.push(category);
            }
        }
        UO_logger.debug(
            "Reset - calendar categories kept: " + resetCategories.join(","),
            false,
            false
        );
        UtilsOBM.setPrefSafe(
            "calendar.categories.names",
            resetCategories.join(",")
        );
    }

    UO_logger.debug("Reset - ADDRESSBOOKS", false, false);

    let addressBookUids = UtilsOBM.getPrefSafe(
        "extensions.obm.addressbooks",
        null
    );
    let abNames = addressBookUids ? addressBookUids.split(",") : [];
    let abManager = Components.classes["@mozilla.org/abmanager;1"].getService(
        Components.interfaces.nsIAbManager
    );

    for (let oldAddressBookUid of["contacts", "users"]) {
        if (!(oldAddressBookUid in abNames)) {
            abNames.push(oldAddressBookUid);
        }
    }

    UO_logger.debug("Reset - addressbooks: abNames=" + abNames, false, false);

    for (let abName of abNames) {
        let iter = abManager.directories;
        let dirPrefId = "ldap_2.servers." + abName;

        while (iter.hasMoreElements()) {
            let data = iter.getNext();

            try {
                if (data instanceof Components.interfaces.nsIAbDirectory) {
                    let uri = data.URI;

                    if (
                        data.dirPrefId == dirPrefId ||
                        (uri && uri.indexOf(abName) > -1)
                    ) {
                        UO_logger.debug(
                            "Reset - addressbooks: deleting addressbook '" + uri + "'"
                        );

                        abManager.deleteAddressBook(uri);
                    }
                }
            } catch (e) {
                UO_logger.debug("unable to remove addressbook " + e, true, true);
            }
        }
    }

    // delete prefs
    UO_logger.debug(
        "Reset - addressbooks: deleting pref 'extensions.obm.changes'",
        false,
        false
    );
    UtilsOBM.deletePreferenceBranch("extensions.obm.changes");
    UO_logger.debug(
        "Reset - deleting pref 'extensions.obm.changesv2'",
        false,
        false
    );
    UtilsOBM.deletePreferenceBranch("extensions.obm.changesv2");
    UO_logger.debug(
        "Reset - addressbooks: deleting pref 'extensions.obm.lastPurgeAsk'",
        false,
        false
    );
    UtilsOBM.deletePreferenceBranch("extensions.obm.lastPurgeAsk");

    UtilsOBM.domainUuid.delete();

    for (let pref of storedPreferences) {
        UtilsOBM.setPrefSafe(pref.key, pref.value);
    }

    var defaultCallback = function() {
        var promptService = Components.classes[
            "@mozilla.org/embedcomp/prompt-service;1"
        ].getService(Components.interfaces.nsIPromptService);

        var check = { value: false };
        var flags =
            promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_0 +
            promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_1;

        var title = UtilsOBM.getLocalizedMessage("obmreinittitle");
        var message = UtilsOBM.getLocalizedMessage("obmreinitmessage");
        var rightnow = UtilsOBM.getLocalizedMessage("obmreinitrightnow");
        var later = UtilsOBM.getLocalizedMessage("obmreinitlater");
        var button = promptService.confirmEx(
            null,
            title,
            message,
            flags,
            rightnow,
            later,
            "",
            null,
            check
        );

        UO_logger.debug("Reset = END =", false, false);

        if (button == 0) {
            // restart now
            UtilsOBM.restartThunderbird();
        }
    };

    if (cb) {
        cb();
    } else {
        defaultCallback();
    }
};

UtilsOBM.getHash = function(aItem) {
    if ("getProperty" in aItem) {
        return aItem.getProperty("X-OBM-HASH");
    }

    return aItem.hash || null;
};

UtilsOBM.getSequence = function(aItem) {
    if ("getProperty" in aItem) {
        return aItem.getProperty("SEQUENCE") || 0;
    }
    if (aItem.generation) return aItem.generation;
    return 0;
};

UtilsOBM.setSequence = function(aItem, seq) {
    seq = seq || 0;
    if ("setProperty" in aItem) {
        return aItem.setProperty("SEQUENCE", seq);
    }
    aItem.generation = seq;
};

function strim(mystr) {
    if (mystr && mystr.replace)
        return mystr.replace(/^\s\s*/, "").replace(/\s\s*$/, "");
    return mystr;
}

UtilsOBM.applyRemoteParticipationChanges = function(source, target, recursive) {
    recursive = recursive || false;
    var mergedEvent = target.clone();
    var sourceAttendees = source.getAttendees({});
    for (let sourceAttendee of sourceAttendees) {
        var targetAttendee = mergedEvent.getAttendeeById(sourceAttendee.id);
        if (targetAttendee) {
            mergedEvent.removeAttendee(targetAttendee);
            targetAttendee = targetAttendee.clone();
            UO_logger.debug("This is a known targetAttendee", false, false);
        } else {
            targetAttendee = sourceAttendee.clone();
            UO_logger.debug("This is a new targetAttendee", false, false);
        }
        UO_logger.debug(
            "targetAttendee: " + targetAttendee + " " + targetAttendee.isMutable,
            false,
            false
        );
        targetAttendee.participationStatus = sourceAttendee.participationStatus;
        UO_logger.debug(
            "targetAttendee: " +
            targetAttendee +
            " participationStatus: " +
            sourceAttendee.participationStatus,
            false,
            false
        );
        mergedEvent.addAttendee(targetAttendee);
    }

    if (recursive) {
        if (source.recurrenceInfo) {
            var exceptionIds = source.recurrenceInfo.getExceptionIds({});
            for (let exceptionId of exceptionIds) {
                UO_logger.debug("source got exception @: " + exceptionId, false, false);
                var sourceException = source.recurrenceInfo.getExceptionFor(
                    exceptionId,
                    false
                );
                var targetException = mergedEvent.recurrenceInfo.getExceptionFor(
                    exceptionId,
                    false
                );
                if (!targetException) {
                    targetException = mergedEvent.recurrenceInfo.getOccurrenceFor(
                        exceptionId
                    );
                }
                targetException = UtilsOBM.applyRemoteParticipationChanges(
                    sourceException,
                    targetException,
                    false
                );
                mergedEvent.recurrenceInfo.modifyException(targetException, false);
            }
        }
    }
    return mergedEvent;
};

UtilsOBM.localEventIsReinitialized = function(localEvent, remoteEvent) {
    var changesSum = function(evt) {
        var majorProps = { RDATE: true, RRULE: true };
        var propStrings = [];
        var prop;
        var comp = evt.icalComponent;
        for (
            prop = comp.getFirstProperty("ANY"); prop; prop = comp.getNextProperty("ANY")
        ) {
            if (majorProps[prop.propertyName]) {
                propStrings.push(prop.icalString);
            }
        }
        var sdate = evt.icalComponent.startTime.getInTimezone(cal.dtz.UTC);
        propStrings.push(sdate.toString());
        return propStrings.join("#");
    };
    var oldItemSum = changesSum(localEvent),
        newItemSum = changesSum(remoteEvent);
    return oldItemSum != newItemSum;
};

UtilsOBM.hasPropertiesDifferences = function(aFirstItem, aSecondItem) {
    var debugmode = UO_logger.getLevelWriter(0) != null;
    if (debugmode)
        UO_logger.debug("hasMajorDifference with debug mode", false, false);
    else UO_logger.debug("hasMajorDifference without debug mode", false, false);

    // Now check regular attributes
    if (
        strim(aFirstItem.title) != strim(aSecondItem.title) ||
        aFirstItem.calendar.id != aSecondItem.calendar.id ||
        aFirstItem.startDate.nativeTime != aSecondItem.startDate.nativeTime ||
        aFirstItem.endDate.nativeTime != aSecondItem.endDate.nativeTime ||
        aFirstItem.recurrenceStartDate.nativeTime !=
        aSecondItem.recurrenceStartDate.nativeTime ||
        strim(aFirstItem.description) != strim(aSecondItem.description) ||
        aFirstItem.privacy != aSecondItem.privacy ||
        aFirstItem.priority != aSecondItem.priority ||
        aFirstItem.status != aSecondItem.status ||
        strim(aFirstItem.getProperty("LOCATION")) !=
        strim(aSecondItem.getProperty("LOCATION"))
    ) {
        if (debugmode) {
            if (strim(aFirstItem.title) != strim(aSecondItem.title))
                UO_logger.debug(
                    " Change in title. old=" +
                    aSecondItem.title +
                    " ; new=" +
                    aFirstItem.title,
                    false,
                    false
                );
            if (aFirstItem.calendar.id != aSecondItem.calendar.id)
                UO_logger.debug(
                    " Change in calendar.id. old=" +
                    aSecondItem.calendar.id +
                    " ; new=" +
                    aFirstItem.calendar.id,
                    false,
                    false
                );
            if (aFirstItem.startDate.nativeTime != aSecondItem.startDate.nativeTime)
                UO_logger.debug(
                    " Change in startDate. old=" +
                    aSecondItem.startDate.nativeTime +
                    " ; new=" +
                    aFirstItem.startDate.nativeTime,
                    false,
                    false
                );
            if (aFirstItem.endDate.nativeTime != aSecondItem.endDate.nativeTime)
                UO_logger.debug(
                    " Change in endDate. old=" +
                    aSecondItem.endDate.nativeTime +
                    " ; new=" +
                    aFirstItem.endDate.nativeTime,
                    false,
                    false
                );
            if (
                aFirstItem.recurrenceStartDate.nativeTime !=
                aSecondItem.recurrenceStartDate.nativeTime
            )
                UO_logger.debug(
                    " Change in recurrenceStartDate. old=" +
                    aSecondItem.recurrenceStartDate.nativeTime +
                    " ; new=" +
                    aFirstItem.recurrenceStartDate.nativeTime,
                    false,
                    false
                );
            if (strim(aFirstItem.description) != strim(aSecondItem.description))
                UO_logger.debug(
                    " Change in description. old=" +
                    aSecondItem.description +
                    " ; new=" +
                    aFirstItem.description,
                    false,
                    false
                );
            if (aFirstItem.privacy != aSecondItem.privacy)
                UO_logger.debug(
                    " Change in privacy. old=" +
                    aSecondItem.privacy +
                    " ; new=" +
                    aFirstItem.privacy,
                    false,
                    false
                );
            if (aFirstItem.priority != aSecondItem.priority)
                UO_logger.debug(
                    " Change in priority. old=" +
                    aSecondItem.priority +
                    " ; new=" +
                    aFirstItem.priority,
                    false,
                    false
                );
            if (aFirstItem.status != aSecondItem.status)
                UO_logger.debug(
                    " Change in status. old=" +
                    aSecondItem.status +
                    " ; new=" +
                    aFirstItem.status,
                    false,
                    false
                );
            if (
                strim(aFirstItem.getProperty("LOCATION")) !=
                strim(aSecondItem.getProperty("LOCATION"))
            )
                UO_logger.debug(
                    " Change in location. old=" +
                    aSecondItem.getProperty("LOCATION") +
                    " ; new=" +
                    aFirstItem.getProperty("LOCATION"),
                    false,
                    false
                );
        }
        return true;
    }

    /*var anEnum = aFirstItem.propertyEnumerator;
       while (anEnum.hasMoreElements()) {
       var property = anEnum.getNext().QueryInterface(Components.interfaces.nsIProperty);
       UO_logger.debug(property.name + " : " + property.value, false, false);
       }*/
    // Check dynamic properties
    // not in DURATION, this property is not always set by lightning, even if startdate and enddate are equals.
    var properties = [
        "STATUS",
        "CATEGORIES",
        "PRIORITY",
        "DUE",
        "COMPLETED",
        "RECURRENCEID",
        "CLASS"
    ];
    for (let property of properties) {
        if (aFirstItem.getProperty(property) != aSecondItem.getProperty(property)) {
            if (debugmode)
                UO_logger.debug(
                    " Change in property " +
                    property +
                    ". old=" +
                    aSecondItem.getProperty(property) +
                    " ; new=" +
                    aFirstItem.getProperty(property),
                    false,
                    false
                );
            return true;
        }
    }
    // Attachments
    // Let's build for 'first' a String representing all attachments URIs sorted in a single String
    var firstAttachments = aFirstItem.getAttachments({});
    var firstAttachmentsArr = [];
    for (i in firstAttachments) {
        firstAttachmentsArr[i] = firstAttachments[i].uri.path;
    }
    // Same for 'second'
    var secondAttachments = aSecondItem.getAttachments({});
    var secondAttachmentsArr = [];
    for (i in secondAttachments) {
        secondAttachmentsArr[i] = secondAttachments[i].uri.path;
    }

    // Final Strings ready to be compared
    s1 = firstAttachmentsArr.sort().join("\r\n");
    s2 = secondAttachmentsArr.sort().join("\r\n");

    if (s1 != s2) {
        if (debugmode)
            UO_logger.debug(
                " Change in attachments. old list=" + s2 + "\n  new list=" + s1,
                false,
                false
            );
        return true;
    }
};

UtilsOBM.getExdates = function(event, asDateTime) {
    asDateTime = asDateTime || false;
    var comp = event.icalComponent;
    var back = [];
    for (
        var i = comp.getFirstProperty("EXDATE"); i; i = comp.getNextProperty("EXDATE")
    ) {
        if (asDateTime) {
            back.push(i.valueAsDatetime);
        } else {
            back.push(i.value);
        }
    }
    back.sort();
    return back;
};

/*
 * HashMap see Java's HashMap API documentation
 */

/** @ignore */
function HashMap() {
    this.length = 0;
    this.data = {};
}

HashMap.prototype.put = function(aKey, aValue) {
    var previous = this.data[aKey];
    this.data[aKey] = aValue;
    if (previous === undefined) {
        this.length++;
    }
    return previous;
};

HashMap.prototype.get = function(aKey) {
    return typeof this.data[aKey] == "undefined" ? undefined : this.data[aKey];
};

HashMap.prototype.keySet = function() {
    var keySet = [];
    for (var key in this.data) {
        keySet.push(key);
    }
    return keySet;
};

HashMap.prototype.values = function() {
    var values = [];
    for (var key in this.data) {
        values.push(this.data[key]);
    }
    return values;
};

HashMap.prototype.size = function() {
    return this.length;
};

HashMap.prototype.remove = function(aKey) {
    if (this.data[aKey] !== undefined) {
        delete this.data[aKey];
        this.length--;
    }
};

HashMap.prototype.clear = function() {
    this.data = {};
    this.length = 0;
};

HashMap.prototype.isEmpty = function() {
    return this.length == 0;
};

HashMap.prototype.containsKey = function(aKey) {
    for (var key in this.data) {
        if (key == aKey) return true;
    }
    return false;
};

HashMap.prototype.containsValue = function(aValue) {
    for (var key in this.data) {
        if (this.data[key] == aValue) return true;
    }
    return false;
};

HashMap.prototype.toString = function() {
    var str = "";
    for (var key in this.data) {
        str += "'" + key + "': '" + this.data[key] + "'\n";
    }
    return str;
};

UtilsOBM.checkLightningUpgrade = function() {
    var mLogger = UO_logger;
    var getCalMgr = function() {
        return Components.classes["@mozilla.org/calendar/manager;1"].getService(
            Components.interfaces.calICalendarManager
        );
    };

    var _getSubscribedOBMCalendar = function() {
        //  this.logger.debug("_getSubscribedOBMCalendar()", false, false);
        var calMgr = getCalMgr();
        var calendars = calMgr.getCalendars({});
        var res = [];
        for (let calendar of calendars) {
            if (calendar.type == "obm") {
                res.push(calendar);
            }
        }
        return res;
    };

    var launch = function() {
        var calendars = _getSubscribedOBMCalendar();
        if (calendars.length) {
            var calMgr = getCalMgr();
            var calendar = calendars[0];
            var obmname = calMgr.getCalendarPref_(calendar, "X-OBM-NAME");
            if (!obmname || !obmname.length) {
                mLogger.debug("Lightning: migration from 0.9 to 1.0", false, false);
                for (let cal of calendars) {
                    mLogger.debug("Migration: " + cal.id, false, false);
                    var obmname = calMgr.getCalendarPref_(calendar, "x-obm-name");
                    var name = calMgr.getCalendarPref_(calendar, "name");
                    var email = calMgr.getCalendarPref_(calendar, "x-obm-email");
                    mLogger.debug(
                        "name: " + name + ", email: " + email + ", obmname: " + obmname,
                        false,
                        false
                    );
                    if (!name || !email || !obmname) {
                        mLogger.debug(
                            "GRAVE: NAME, X-OBM-NAME or X-OBM-EMAIL not found",
                            false,
                            false
                        );
                        mLogger.debug(
                            "name: " + name + ", email: " + email + ", obmname: " + obmname,
                            false,
                            false
                        );
                        continue;
                    }
                    mLogger.debug(
                        "Setting name: " +
                        name +
                        ", email: " +
                        email +
                        ", obmname: " +
                        obmname,
                        false,
                        false
                    );
                    calMgr.setCalendarPref_(calendar, "X-OBM-NAME", obmname);
                    calMgr.setCalendarPref_(calendar, "X-OBM-EMAIL", email);
                    calMgr.setCalendarPref_(calendar, "NAME", name);
                }
            } else {
                mLogger.debug("Lightning: up to date " + obmname, false, false);
            }
        }
    };
    launch();
};

UtilsOBM.isOnline = function() {
    try {
        if (
            Components.classes["@mozilla.org/network/io-service;1"].getService(
                Components.interfaces.nsIIOService2
            ).offline
        ) {
            return false;
        }
    } catch (e) {
        UO_logger.debug("error getting online status", false, false);
    }
    return true;
};

UtilsOBM.makeCalendarUri = function(id, username) {
    return "moz-storage-calendar://id=" + id + ";" + username;
};

UtilsOBM.calendarOwnerEmail = function(aItem) {
    return cal
        .getCalendarManager()
        .getCalendarPref_(aItem.calendar, "X-OBM-EMAIL");
};

UtilsOBM.calendarOwnMailTo = function(aItem) {
    return "mailto:" + UtilsOBM.calendarOwnerEmail(aItem);
};

UtilsOBM.isEventDelegation = function(aItem) {
    var calendarOwner = UtilsOBM.getCalendarOBMName(aItem.calendar);
    var tbirdUser = {},
        tbirdPasswd = {},
        obmLogin;
    var success = UtilsOBM.getOBMCredentials(tbirdUser, tbirdPasswd);
    UO_logger.debug("getOBMCredentials success ? " + success);
    if (success) {
        obmLogin = tbirdUser.value.split("@").shift();
        UO_logger.debug(
            "calendarOwner: " + calendarOwner + ", tbird User: " + obmLogin,
            false,
            false
        );
    }

    if (!success || obmLogin == calendarOwner) {
        return false;
    }
    return true;
};

UtilsOBM.domainUuid = {
    get: function() {
        return UtilsOBM.getPrefSafe("extensions.obm.server.uuid", null);
    },
    set: function(uuid) {
        if (!uuid || !uuid.length) {
            return false;
        }
        return UtilsOBM.setPrefSafe("extensions.obm.server.uuid", uuid);
    },
    delete: function() {
        return UtilsOBM.deletePreferenceBranch("extensions.obm.server.uuid");
    },
    getFromLoginResponse: function(loginXml) {
        try {
            var domain_uuid = OBMXPathFirst(loginXml, "domain/@uuid");
            UtilsOBM.domainUuid.set(domain_uuid);
        } catch (e) {
            UO_logger.debug("This Obm-sync server does not support domain uuid");
        }
    }
};

UtilsOBM.getRandomInt = function(min, max) {
    return Math.floor(Math.random() * (max - min + 1)) + min;
};

UtilsOBM.isUsersAddressBook = function(name) {
    return "-1" == name || "users" == name;
};

UtilsOBM.isDefaultEventTitle = function(event) {
    var stringBundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
        .getService(Components.interfaces.nsIStringBundleService)
        .createBundle("chrome://calendar/locale/calendar.properties");
    var newEventLabel = stringBundle.GetStringFromName("newEvent");
    return event.title == newEventLabel;
};

UtilsOBM.restartThunderbird = function() {
    let nsIAppStartup = Components.interfaces.nsIAppStartup;

    Components.classes["@mozilla.org/toolkit/app-startup;1"]
        .getService(nsIAppStartup)
        .quit(nsIAppStartup.eForceQuit | nsIAppStartup.eRestart);
};

UtilsOBM.getTimezoneOrUTC = function(tzId) {
    let timezone = cal.getTimezoneService().getTimezone(tzId);

    return timezone ? timezone : cal.dtz.UTC;
};

UtilsOBM.applyRemoteCalendarRights = function(
    localCalendar,
    remoteCalInfo,
    forceReSubscription
) {
    let remoteCalHasWriteRights = remoteCalInfo.write;
    let localCalHasWriteRights = !localCalendar.readOnly;
    let calendarManager = cal.getCalendarManager();
    let identityHash = {};

    cal.email.iterateIdentities(function(identity, account) {
        // This might overwrite identities if there are same emails in
        // different accounts, but I guess thats very uncommon
        identityHash[identity.email] = [identity, account];
        return true;
    });

    if (
        forceReSubscription ||
        remoteCalHasWriteRights != localCalHasWriteRights
    ) {
        // The rights to write events have been disabled remotely
        // Discard any changes done in this calendar since last time
        // Remove the calendar, and re-add it
        let oldColor = localCalendar.getProperty("color");

        calendarManager.unregisterCalendar(localCalendar);
        calendarManager.removeCalendar(localCalendar);

        let name =
            remoteCalInfo.name ||
            UtilsOBM.getPrefSafe("extensions.obm.autoconf.calname", null) ||
            remoteCalInfo.firstname + " " + remoteCalInfo.lastname;
        let obmCalMgr = UtilsOBM.getCalendarManager();
        var newCalendar = obmCalMgr.subscribeObmCalendar(
            remoteCalInfo.uid,
            name,
            remoteCalInfo.mail, !remoteCalHasWriteRights
        );

        // _subscribeTo has assigned a random color to the 'new'
        // calendar. We need to assign the old color and re-add it
        newCalendar.setProperty("color", oldColor);
        var wasWriteRightsRemoved = !remoteCalHasWriteRights;
        // Don't attempt to sync events, since we don't have the rights
        // to write on this calendar anymore
        if (wasWriteRightsRemoved) {
            UtilsOBM.removeAllCalendarSyncItems(localCalendar);
        }

        localCalendar = newCalendar;
    }

    // In any case, set up extra properties used by ACL bits for calendar
    localCalendar.setProperty("X-OBM-WRITABLE", remoteCalHasWriteRights);
    localCalendar.setProperty("X-OBM-UID", remoteCalInfo.uid);

    // Make sure the imip identity is set correctly, this will set
    // the organizerId correctly in cal.ProviderBase
    let username = {};
    let hashKeys = Object.keys(identityHash);
    if (remoteCalInfo.mail in identityHash) {
        let [identity, account] = identityHash[remoteCalInfo.mail];
        localCalendar.setProperty("imip.identity.key", identity.key);
    } else if (hashKeys.length == 1) {
        // Looks like there is just one, that must be it.
        localCalendar.setProperty(
            "imip.identity.key",
            identityHash[hashKeys[0]].key
        );
    } else {
        // Hmm, strange. The user seems to be using aliases. The best we can do
        // now is to find an identity with the same name (but a different
        // domain) or if that fails just any one with the same domain. If the
        // user has not logged in yet, the identity is reset.
        let foundByUser = null;
        let foundByDomain = null;
        if (UtilsOBM.getOBMCredentials(username, {})) {
            let userid = username.value.replace(/@.*/, "@");
            let userdomain = username.value.replace(/[^@]*@/, "@");
            for (let hash in identityHash) {
                let ident = identityHash[hash][0];

                if (!ident || !ident.email) {
                    continue;
                }

                if (ident.email.startsWith(userid)) {
                    foundByUser = ident.key;
                    break;
                } else if (ident.email.endsWith(userdomain)) {
                    foundByDomain = ident.key;
                    // We're not breaking here as we try to favor foundByUser
                }
            }
        }
        localCalendar.setProperty(
            "imip.identity.key",
            foundByUser || foundByDomain
        );
        if (!foundByUser && !foundByDomain) {
            UO_logger.warning(
                "Could not find Thunderbird identity for " + remoteCalInfo.mail,
                false,
                true
            );
        }
    }
};

UtilsOBM.getCalInfoForReadOnlyLocalCalendar = function(calendar) {
    return UtilsOBM.getCalInfoForLocalCalendar(calendar, true);
};

UtilsOBM.getCalInfoForLocalCalendar = function(calendar, readOnly) {
    var calInfo = {};

    calInfo.uid = calendar.getProperty("X-OBM-NAME");
    calInfo.name = calendar.name;
    calInfo.mail = calendar.getProperty("X-OBM-EMAIL");
    calInfo.read = true;
    calInfo.write = !readOnly;

    return calInfo;
};

/**
 * Calculate differences between two sets, i.e arrays.
 *
 * @param set1          The set to use as a basis (i.e old state)
 * @param set2          The set to compare against (i.e new state)
 * @param keyfunc       (optional) The function to use to calculate a key
 *                        suitable for an object key (i.e string)
 * @param onadd         (optional) A function to call for added item (only in set2)
 * @param onchange      (optional) A function to call for changed item (in both sets)
 * @param ondelete      (optional) A function to call for removed item (only in set1)
 * @return              An array [added,changed,removed], where each element is also an array.
 */
UtilsOBM.calculateChanges = function calculateChanges(
    set1,
    set2,
    keyfunc,
    onadd,
    onchange,
    ondelete
) {
    if (!keyfunc)
        keyfunc = function(x) {
            x;
        };

    let map = [];
    let added = [];
    let changed = [];
    let removed = [];

    for (let keyobj of set1) {
        map[keyfunc(keyobj)] = keyobj;
    }

    for (let keyobj of set2) {
        let key = keyfunc(keyobj);
        if (key in map) {
            changed.push(keyobj);
            if (onchange) onchange(keyobj);
            delete map[key];
        } else {
            added.push(keyobj);
            if (onadd) onadd(keyobj);
        }
    }

    for (let keyobj of map) {
        removed.push(keyobj);
    }

    if (ondelete) removed.forEach(ondelete);

    return [added, changed, removed];
};

/**
 * Change the item using |func| without changing LAST-MODIFIED and DTSTAMP.
 * The function is called with the item as its argument.
 *
 * @param item      The item to change
 * @param func      The function to execute without changing the stamps.
 * @return          The result of the function call.
 */
UtilsOBM.retainStamps = function(item, func) {
    let lastmod = item.getProperty("LAST-MODIFIED");
    let stamp = item.getProperty("DTSTAMP");
    let res = func.call(null, item);

    item.setProperty("DTSTAMP", stamp ? stamp.clone() : null);
    item.setProperty("LAST-MODIFIED", lastmod ? lastmod.clone() : null);

    return res;
};

UtilsOBM.getSyncAfterFromPref = function() {
    let fullSync = UtilsOBM.getPrefSafe(
        "extensions.obm.calendar.syncrange.full",
        false
    );
    if (fullSync) {
        return 0;
    } else {
        let maxMonths = UtilsOBM.getPrefSafe(
            "extensions.obm.calendar.syncrange.maxMonths",
            3
        );
        let syncAfter = new Date();

        syncAfter.setMonth(syncAfter.getMonth() - maxMonths);

        return syncAfter.getTime();
    }
};

UtilsOBM.createResourceUri = function(email) {
    let prepareUrl = UtilsOBM.getPrefSafe("extensions.obm.server");

    prepareUrl = prepareUrl.replace("/obm-sync/services", "/obm-sync/resources");
    prepareUrl += "/" + email.toString();
    prepareUrl += "?syncRangeAfter=" + UtilsOBM.getSyncAfterFromPref();

    return UtilsOBM.makeURL(prepareUrl);
};

UtilsOBM.resubscribeToResource = function(resource) {
    let calendarManager = cal.getCalendarManager();
    let email = calendarManager.getCalendarPref_(resource, "X-OBM-EMAIL");
    let name = resource.name;
    let color = resource.getProperty("color");
    let readOnly = resource.readOnly;

    let newUri = UtilsOBM.createResourceUri(
        resource.getProperty("X-OBM-RESOURCE-EMAIL")
    );

    calendarManager.unregisterCalendar(resource);
    calendarManager.removeCalendar(resource);

    var newCalendar = calendarManager.createCalendar("ics", newUri);

    // Components.utils.import("chrome://obm-extension/content/components/calOBMCalendar.js")
    // var newCalendar = new calOBMCalendar();
    // newCalendar.uri = newUri;

    calendarManager.registerCalendar(newCalendar);
    newCalendar.name = name;
    calendarManager.setCalendarPref_(
        newCalendar,
        "X-OBM-EMAIL",
        email.toString()
    );
    newCalendar.setProperty("color", color.toString());
    newCalendar.setProperty("X-OBM-DEFAULT-NAME", name.toString());
    calendarManager.setCalendarPref_(
        newCalendar,
        "X-OBM-RESOURCE-EMAIL",
        email.toString()
    );
};

UtilsOBM.makeURL = function(aUriString) {
    var ioSvc = Components.classes[
        "@mozilla.org/network/io-service;1"
    ].getService(Components.interfaces.nsIIOService);
    return ioSvc.newURI(aUriString, null, null);
};

UtilsOBM.parseIntWithDefault = function(aNumberString, aRadix, aDefaultValue) {
    return parseInt(aNumberString, aRadix) || aDefaultValue;
};
var { ConsoleAPI } = ChromeUtils.import("resource://gre/modules/Console.jsm");
var obmConsole = new ConsoleAPI({
    prefix: "OBM Plugin",
    consoleID: "obm",
    maxLogLevel: Services.prefs.getBoolPref("OBM.debug.log", true) ? "all" : "warn",
});

UtilsOBM.console = obmConsole;
UtilsOBM.log = obmConsole.log;
UtilsOBM.warn = obmConsole.warn;
UtilsOBM.error = obmConsole.error;

var UO_logger = new ObmLogger();
UO_logger.setHeader('UtilsOBM --- Logs : ')