/******************************************************************************
Copyright (C) 2012-2013 Linagora

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version, provided you comply with the Additional Terms applicable for OBM
connector by Linagora pursuant to Section 7 of the GNU Affero General Public
License, subsections (b), (c), and (e), pursuant to which you must notably (i)
retain the “Message sent thanks to OBM, Free Communication by Linagora”
signature notice appended to any and all outbound messages (notably e-mail and
meeting requests), (ii) retain all hypertext links between OBM and obm.org, as
well as between Linagora and linagora.com, and (iii) refrain from infringing
Linagora intellectual property rights over its trademarks and commercial
brands. Other Additional Terms apply, see <http://www.linagora.com/licenses/>
for more details.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License and
its applicable Additional Terms for OBM along with this program. If not, see
<http://www.gnu.org/licenses/> for the GNU Affero General   Public License
version 3 and <http://www.linagora.com/licenses/> for the Additional Terms
applicable to OBM connectors.
******************************************************************************/

var EXPORTED_SYMBOLS = ["cal"];

// NOTE: This file doesn't really export any symbols, it just overlays
// Lightning's calItipUtils.jsm

Components.utils.import("resource://calendar/modules/calItipUtils.jsm");
Components.utils.import("resource://obm-extension/modules/UtilsOBM.jsm");

/**
 * Check if the first attendee on the itipItemItem has a different
 * participation status then the one on the item
 *
 * @param itipItemItem  The incoming itip item.
 * @param item          The item to check on.
 * @return              False if the pariticpation status is the same,
 *                        otherwise true.
 */
function checkAttendeeUpdated(itipItemItem, item) {
    let attendee = itipItemItem.getAttendees({})[0];
    if (!attendee) {
        return false;
    }

    let attendeeIsNew = false;
    let targetAttendee = item.getAttendeeById(attendee.id);
    if (!targetAttendee) {
        targetAttendee = attendee;
        attendeeIsNew = true;
    }
    let partStatMatches = (targetAttendee.participationStatus !=
        attendee.participationStatus);
    return (attendeeIsNew || partStatMatches);
}

/**
 * Checks the master item and all exceptions if the participation status
 * matches.
 *
 * @param itipItemItem  The incoming itip item.
 * @param item          The item to check on.
 * @return              False if the pariticpation status is the same,
 *                        otherwise true.
 */
function checkAttendeeUpdatedRecurring(itipItemItem, item) {
    if (checkAttendeeUpdated(itipItemItem, item)) {
        return true;
    }

    let itemIsUpdated = false;
    if (itipItemItem.recurrenceInfo) {
        for (let recId of itipItemItem.recurrenceInfo.getExceptionIds({})) {
            let itemExc = item.recurrenceInfo.getExceptionFor(recId, true);
            let itipItemItemExc = itipItemItem.recurrenceInfo.getExceptionFor(recId, false);
            if (itemExc && itipItemItemExc && checkAttendeeUpdated(itipItemItemExc, itemExc)) {
                itemIsUpdated = true;
            }
        }
    }

    return itemIsUpdated;
}

/**
 * Iterator for all combinations of items in the itipItem and foundItems. The
 * iterator will yield [itipItemitem, foundItem].
 *
 * @param itipItem          The itip item to get items from
 * @param foundItems        The calendar items to iterate over
 * @param occurrencesOnly   If true, only itipItems that are occurrences are
 *                            taken into account.
 * @return                  This is an iterator, use it in a for() loop.
 */
function* iterateFoundItems(itipItem, foundItems, occurrencesOnly) {
    for (let itipItemItem of itipItem.getItemList({})) {
        for (let item of foundItems) {
            let rid = itipItemItem.recurrenceId;
            if (occurrencesOnly && (!rid || !item.recurrenceInfo)) {
                continue;
            }

            if (rid) {
                if (item.recurrenceInfo) {
                    item = item.recurrenceInfo.getOccurrenceFor(rid);
                    if (!item) {
                        continue;
                    }
                } else {
                    if (itipItemItem.parentItem.hasProperty("X-MOZ-FAKED-MASTER")) {
                        // Faked master items don't behave well w.r.t. stamp
                        // times. Make sure to take the times from the
                        // occurrence.
                        let occstamp = itipItemItem.getProperty("DTSTAMP");
                        let occlastmod = itipItemItem.getProperty("LAST-MODIFIED");
                        itipItemItem.parentItem.setProperty("DTSTAMP", occstamp);
                        itipItemItem.parentItem.setProperty("LAST-MODIFIED", occlastmod);
                    }
                    itipItemItem = itipItemItem.parentItem;
                }
            }

            yield [itipItemItem, item];
        }
    }
}

function isObmUserAttendee(aItem, obmCalendarEmail) {
    let attendees = aItem.getAttendees({});
    let userIsAttendee = false;
    for (let attendee of attendees) {
        let attendeeEmail = attendee.id.replace(/^mailto:/i, '').toLowerCase();
        if (attendeeEmail == obmCalendarEmail) {
            userIsAttendee = true;
            break;
        }
    }
    return userIsAttendee;
}

function isObmUserOrganizer(aItem, obmCalendarEmail) {
    return obmCalendarEmail == aItem.organizer.id.replace(/^mailto:/, '');
}

function getWritableObmCalendars() {
    let writableCalendars = cal.getCalendarManager().getCalendars({}).
    filter(function(calendar) {
        let isObmWritable = (cal.itip.isSchedulingCalendar(calendar) &&
            (calendar.getProperty("X-OBM-EMAIL") != null));
        return isObmWritable;
    });
    return writableCalendars;
}

function isObmInternalEvent(aItem) {
    let eventIsInternal = false;
    let uuid = UtilsOBM.getPrefSafe("extensions.obm.server.uuid");
    let authDomain = UtilsOBM.getPrefSafe("extensions.obm.server.domain");
    if (uuid && aItem.hasProperty("X-OBM-DOMAIN-UUID")) {
        let icsObmUuid = aItem.getProperty('X-OBM-DOMAIN-UUID');
        eventIsInternal = (uuid == icsObmUuid);
    } else if (authDomain && aItem.hasProperty("X-OBM-DOMAIN")) {
        let obmDomain = aItem.getProperty('X-OBM-DOMAIN');
        eventIsInternal = (authDomain == obmDomain);
    }
    return eventIsInternal;
}

function getObmIdSet(calendars) {
    let obmIdSet = {};
    for (let calendar of calendars) {
        let email = calendar.getProperty("X-OBM-EMAIL");
        let obmId = email.toLowerCase();
        obmIdSet[obmId] = true;
    }
    return obmIdSet;
}

function isOneOfObmUsersAttendee(aItem, obmCalendarEmails) {
    let attendees = aItem.getAttendees({});
    for (let attendee of attendees) {
        let attendeeEmail = attendee.id.replace(/^mailto:/, '').toLowerCase();
        if (attendeeEmail in obmCalendarEmails) {
            return true;
        }
    }
    return false;
}

/**
 * Check if the itipitem is sent by an attendee (so it's a reply)
 */
function isItipSentByAttendee(method) {
    return (["REFRESH", "REPLY"].indexOf(method) >= 0);
}

/**
 * Find the calendars whose owner is in the attendees of the itipItem
 */
function getProspectiveTargetCalendars(item) {
    let writableCalendars;
    let isInternal = isObmInternalEvent(item);
    if (isInternal) {
        function filterObmCalendars(calendar) {
            let obmCalendarEmail = calendar.getProperty("X-OBM-EMAIL");
            return obmCalendarEmail &&
                (isObmUserAttendee(item, obmCalendarEmail) ||
                    isObmUserOrganizer(item, obmCalendarEmail));
        }
        writableCalendars = getWritableObmCalendars().filter(filterObmCalendars);
    } else {
        function isWritableCalendar(aCalendar) {
            /* TODO: missing ACL check for existing items (require callback API) */
            return (cal.itip.isSchedulingCalendar(aCalendar) &&
                cal.userCanAddItemsToCalendar(aCalendar));
        }
        writableCalendars = cal.getCalendarManager().getCalendars({}).filter(isWritableCalendar);
    }
    return writableCalendars;
}

/**
 * Remove from the calendars list the calendar of the organizer of the itipItem
 */
function filterOutOrganizersCalendars(calendars, item) {
    let filteredCalendars = [];
    let organizerId = item.organizer.id;
    let organizerEmail = organizerId.replace(/^mailto:/, '');
    for (let calendar of calendars) {
        let obmEmail = calendar.getProperty("X-OBM-EMAIL");
        if (obmEmail != organizerEmail) {
            filteredCalendars.push(calendar);
        }
    }
    return filteredCalendars;
}

function setTargetCalendars(itipItem) {
    let itemList = itipItem.getItemList({});
    if (!itemList.length) {
        return;
    }
    let item = itemList[0];
    let writableCalendars = getProspectiveTargetCalendars(item);
    if (writableCalendars.length > 0) {
        let itipSentByAttendee = isItipSentByAttendee(itipItem.receivedMethod);
        let targetCalendars = null;
        if (itipSentByAttendee) {
            targetCalendars = writableCalendars;
        } else {
            targetCalendars = filterOutOrganizersCalendars(writableCalendars, item);
        }
        let compCal = Components.classes["@mozilla.org/calendar/calendar;1?type=composite"]
            .createInstance(Components.interfaces.calICompositeCalendar);
        targetCalendars.forEach(compCal.addCalendar, compCal);
        itipItem.targetCalendar = compCal;
    }
}

// Monkeypatch Lightning to change the set of target calendars
let calInitItemFromMsgData = cal.itip.initItemFromMsgData;
cal.itip.initItemFromMsgData = function(itipItem, imipMethod, aMsgHdr) {
    let rv = calInitItemFromMsgData.apply(cal.itip, arguments);
    setTargetCalendars(itipItem);
    return rv;
};

// Monkeypatch Lightning to disable buttons if no OBM calendars match to event
// attendees
let calGetOptionsText = cal.itip.getOptionsText;
cal.itip.getOptionsText = function(itipItem, rc, actionFunc) {
    let data = calGetOptionsText.apply(cal.itip, arguments);

    if (actionFunc && actionFunc.method == "REQUEST") {
        let items = itipItem.getItemList({});
        let disabled = false;
        if (items.length > 0) {
            let item = items[0];
            let writableObmCalendars = getWritableObmCalendars();
            if (writableObmCalendars.length > 0) {
                let obmIdSet = getObmIdSet(writableObmCalendars);
                disabled = (isObmInternalEvent(item) &&
                    !isOneOfObmUsersAttendee(item, obmIdSet));
            }
        }
        if (disabled) {
            data.button1.disabled = true;
            data.button2.disabled = true;
            data.button3.disabled = true;
        }
    }

    return data;
};

let calProcessItipItem = cal.itip.processItipItem;
cal.itip.processItipItem = function(itipItem, ltnOptionsFunc) {
    function obmOptionsFunc(itipItem, rc, actionFunc, foundItems) {
        // Ignore updates if the attendee status didn't change.
        let method = itipItem.receivedMethod.toUpperCase();
        if (actionFunc && foundItems.length > 0 && method == "REPLY") {
            for (let [itipItemItem, item] of iterateFoundItems(itipItem, foundItems, true)) {
                if (!checkAttendeeUpdatedRecurring(itipItemItem, item)) {
                    actionFunc = null;
                    break;
                }
            }
        }

        // If a single occurrence was found, then we don't want to show the
        // master item, but instead the occurrence this affects.
        let itipItemItems = itipItem.getItemList({});
        if (itipItemItems.length == 1 && itipItemItems[0].recurrenceId) {
            let foundOccurrences = foundItems.map(function(x) {
                return x.recurrenceInfo ? x.recurrenceInfo.getExceptionFor(itipItemItems[0].recurrenceId) : x;
            });

            if (foundOccurrences.length) {
                foundItems = foundOccurrences;
            }
        }

        return ltnOptionsFunc(itipItem, rc, actionFunc, foundItems);
    }
    return calProcessItipItem.call(cal.itip, itipItem, obmOptionsFunc);
};

function obmItipItemCompare(itipItemItem, item) {
    return cal.itip.compare(itipItemItem, item) <= 0 && cal.itip.getSequence(itipItemItem) == cal.itip.getSequence(item);
}

// Monkeypatch Lightning to fix bug 918493. This hack can be removed once the
// minimum version exceeds the target milestone set in that bug.
(function patchItipItemFinder() {
    let itipModule = Components.utils.getGlobalForObject(cal.itip);
    let calProcessFoundItems = itipModule.ItipItemFinder.prototype.processFoundItems;

    let createOrganizer = itipModule.createOrganizer;
    let ItipOpListener = itipModule.ItipOpListener;
    let updateItem = itipModule.updateItem;
    let addScheduleAgentClient = itipModule.addScheduleAgentClient;

    // This function is run with the this-scope of ItipItemFinder.
    function obmProcessFoundItems() {
        // This function is code directly from Lightning. Do not change.
        let calWrapupFoundItems = function() {
            cal.LOG("iTIP operations: " + operations.length);
            let actionFunc = null;
            if (operations.length > 0) {
                actionFunc = function execOperations(opListener, partStat) {
                    for (let op of operations) {
                        try {
                            op(opListener, partStat);
                        } catch (exc) {
                            cal.ERROR(exc);
                        }
                    }
                };
                actionFunc.method = actionMethod;
            }
            this.mOptionsFunc(this.mItipItem, rc, actionFunc, this.mFoundItems);
        }.bind(this);

        // This function is code mostly from Lightning. Do not change.
        let calProcessRequest = function(itipItemItem, item) {
            let newItem = updateItem(item, itipItemItem);
            let att = cal.getInvitedAttendee(newItem);
            if (!att) { // fall back to using configured organizer
                att = createOrganizer(newItem.calendar);
                if (att) {
                    att.isOrganizer = false;
                }
            }
            if (att) {
                let foundAttendee = item.getAttendeeById(att.id) || att;
                let self = this;

                // This code is a variation from the original Lightning code.
                //  1. Lightning checks for strict equality of the new itipItem and the existing calendar item
                //  2. We're checking for strict sequence equality, but are tolerant on the stamp, so that if
                //     a minor update occurred since the last received invitation email (thus in this case
                //     itipItemItem.stamp < item.stamp), the user can still see the accept/decline buttons
                if (foundAttendee.participationStatus == "NEEDS-ACTION" &&
                    (item.calendar.getProperty("itip.disableRevisionChecks") ||
                        obmItipItemCompare(itipItemItem, item))) {
                    actionMethod = "REQUEST:NEEDS-ACTION";
                    operations.push(function(opListener, partStat) {
                        let changedItem = item.clone();
                        changedItem.removeAttendee(foundAttendee);
                        foundAttendee = foundAttendee.clone();
                        if (partStat) {
                            foundAttendee.participationStatus = partStat;
                        }
                        changedItem.addAttendee(foundAttendee);

                        itipModule.copyProviderProperties(self.mItipItem, itipItemItem, changedItem);

                        return changedItem.calendar.modifyItem(
                            changedItem, item, new ItipOpListener(opListener, item));
                    });
                } else if (item.calendar.getProperty("itip.disableRevisionChecks") ||
                    cal.itip.compare(itipItemItem, item) > 0) {
                    addScheduleAgentClient(newItem, item.calendar);

                    let isMinorUpdate = (cal.itip.getSequence(newItem) ==
                        cal.itip.getSequence(item));
                    actionMethod = (isMinorUpdate ? method + ":UPDATE-MINOR" :
                        method + ":UPDATE");

                    itipModule.copyProviderProperties(this.mItipItem, itipItemItem, newItem);

                    operations.push(function(opListener, partStat) {
                        if (!partStat) { // keep PARTSTAT
                            let att_ = cal.getInvitedAttendee(item);
                            partStat = (att_ ? att_.participationStatus : "NEEDS-ACTION");
                        }
                        newItem.removeAttendee(att);
                        att = att.clone();
                        att.participationStatus = partStat;
                        newItem.addAttendee(att);
                        return newItem.calendar.modifyItem(
                            newItem, item, new ItipOpListener(opListener, item));
                    });
                }
            }
        }.bind(this);

        // Now comes the code to mock the lightning code. Use our replacement
        // in the case we are covering, otherwise use the original function
        let method = this.mItipItem.receivedMethod.toUpperCase();
        let operations = [];
        let rc = Components.results.NS_OK;
        let actionMethod = method;

        if (method == "REQUEST" && this.mFoundItems.length) {
            let hasUndecided = false;
            let calendars = cal.getCalendarManager({}).getCalendars({});
            let username = { value: null };
            let domain = UtilsOBM.getPrefSafe("extensions.obm.server.domain");
            UtilsOBM.getOBMCredentials(username, {});

            for (let [itipItemItem, item] of iterateFoundItems(this.mItipItem, this.mFoundItems)) {
                for (let calendar of calendars) {
                    let isObmUser = calendar.getProperty("X-OBM-NAME") == username.value ||
                        calendar.getProperty("X-OBM-EMAIL") == username.value + '@' + domain;
                    if (calendar.type != "obm" || calendar.readOnly || isObmUser) {
                        continue;
                    }
                    let organizer = calendar.getProperty("organizerId");
                    let att = item.getAttendeeById(organizer);

                    // If we are subscribed to an obm calendar that is owned by
                    // one of the attendees, and that attendee has not decided
                    // yet, we need to clear the found items to make sure the
                    // user can decide for that attendee. This is used for the
                    // secretary case.
                    if (att && att.participationStatus == "NEEDS-ACTION") {
                        this.mFoundItems = [];
                        return calProcessFoundItems.apply(this, arguments);
                    }
                }

            }

            for (let [itipItemItem, item] of iterateFoundItems(this.mItipItem, this.mFoundItems)) {
                calProcessRequest(itipItemItem, item);
            }
            return calWrapupFoundItems();
        } else {
            // we can let the original function take care
            return calProcessFoundItems.apply(this, arguments);
        }
    }

    // Override the lightning function.
    itipModule.ItipItemFinder.prototype.processFoundItems = obmProcessFoundItems;
})();